@@ -1,5 +1,5 @@
 #if	!defined(lint) && !defined(DOS)
-static char rcsid[] = "$Id: browse.c,v 4.62 1996/05/29 00:05:04 mikes Exp $";
+static char rcsid[] = "$Id: browse.c,v 4.82 1998/09/16 00:18:45 hubert Exp $";
 #endif
 /*
  * Program:	Routines to support file browser in pico and Pine composer
@@ -21,7 +21,7 @@ static char rcsid[] = "$Id: browse.c,v 4.62 1996/05/29 00:05:04 mikes Exp $";
  * permission of the University of Washington.
  * 
  * Pine, Pico, and Pilot software and its included text are Copyright
- * 1989-1996 by the University of Washington.
+ * 1989-1998 by the University of Washington.
  * 
  * The full text of our legal notices is contained in the file called
  * CPYRIGHT, included with this distribution.
@@ -49,14 +49,7 @@ static char rcsid[] = "$Id: browse.c,v 4.62 1996/05/29 00:05:04 mikes Exp $";
  *
  *
  */
-#include <stdio.h>
-#include <ctype.h>
-#include <errno.h>
-#include "osdep.h"
-#include "pico.h"
-#include "estruct.h"
-#include "edef.h"
-#include "efunc.h"
+#include "headers.h"
 
 #ifndef	_WINDOWS
 
@@ -112,7 +105,7 @@ static	char	*browser_title = NULL;
     int    sisin(char *, char *);
     int    BrowserAnchor(char *);
     void   ClearBrowserScreen(void);
-    void   BrowserRunChild(char *);
+    void   BrowserRunChild(char *, char *);
     int    LikelyASCII(char *);
 #else
     struct bmaster *getfcells();
@@ -179,7 +172,7 @@ static int  bfmappings[2][12][2] = { { { F1,  '?'},	/* stand alone */
 				       { F9,  'd'},
 				       { F10, 'r'},
 				       { F11, 'c'},
-				       { F12, NODATA } } };
+				       { F12, 'a'} } };
 
 
 /*
@@ -207,23 +200,55 @@ static char *BrowseHelpText[] = {
 NULL
 };
 
+/*
+ * Help for standalone browser (pilot)
+ */
 static char *sa_BrowseHelpText[] = {
-"Help for File Browser",
-"  ",
-"\tThe File Browser is used to display and manipulate files.",
-"  ",
-"~\tBoth directories and files are displayed.  Press ~V",
-"~\tor ~R~e~t~u~r~n to display the selected directory or view a",
-"~\ttext file's contents.  Other commands are available to",
-"~\tEdit, Copy, Rename, and Delete the selected (highlighted)",
-"~\tfile.",
-"  ",
-"\tThe file named \"..\" is special, and means the \"parent\"",
-"\tof the directory being displayed.  Select this directory",
-"\tto move upward in the directory tree.",
-"  ",
-"End of File Browser Help.",
-"  ",
+"Help for Pilot (PIne's Looker-upper Of Things",
+" ",
+"\tPilot is a simple, display-oriented file system browser based on the",
+"~\tPine message system composer.  As with Pine, commands are displayed at",
+"~\tthe bottom of the screen, and context-sensitive help is provided.",
+" ",
+"~\tPilot displays the current working directory at the top of the screen.",
+"~\tThe directory's contents are displayed in columns of file name, file",
+"~\tsize pairs.  Names that are directories are indicated by the name",
+"~\t~(~d~i~r~) in place of the file size.  The parent of the current working",
+"~\tdirectory is indicated by the file name ~.~. and size of ~(~p~a~r~e~n~t~ ~d~i~r~).",
+"~\tFile names that are symbolic links to other files are displayed with a",
+"~\tfile size of ~-~-.",
+" ",
+"\tThe following function keys are available in Pilot:",
+" ",
+"~\t~?\tDisplay this help text.",
+"~\t~Q\tQuit Pilot.",
+" ",
+"~\t~V\tView the currently selected file or open the selected directory.",
+"~\t\tNote: Pilot invokes ~p~i~n~e ~-~F, or the program defined by the ~P~A~G~E~R",
+"~\t\t      environment variable, to view the file.",
+"~\t~L\tLaunch an external application program.",
+" ",
+"~\t~W\tSearch for a file by name.",
+"~\t~-\tScroll up one page.",
+"~\t~S~p~a~c~e\tScroll down one page.",
+"~\t~N,~^~F\tMove forward (right) one column.",
+"~\t~P,~^~B\tMove back (left) one column.",
+"~\t~^~N\tMove down one row.",
+"~\t~^~P\tMove up one row.",
+" ",
+"~\t~D\tDelete the selected file.",
+"~\t~R\tRename the selected file or directory.",
+"~\t~C\tCopy the selected file.",
+"~\t~E\tEdit the selected file.",
+"~\t\tNote: Pilot invokes ~p~i~c~o, or the program defined by the ~E~D~I~T~O~R",
+"~\t\t      environment variable, to edit the file.",
+" ",
+" ",
+"\tPine and Pico are trademarks of the University of Washington.",
+"\tNo commercial use of these trademarks may be made without prior",
+"\twritten permission of the University of Washington.",
+" ",
+"End of Pilot Help.",
 NULL
 };
 
@@ -232,10 +257,10 @@ NULL
 /*
  * pico_file_browse - Exported version of FileBrowse below.
  */
-pico_file_browse(pdata, dir, fn, sz, flags)
+pico_file_browse(pdata, dir, dirlen, fn, fnlen, sz, flags)
     PICO *pdata;
     char *dir, *fn, *sz;
-    int   flags;
+    int   dirlen, fnlen, flags;
 {
     int  rv;
     char title_buf[64];
@@ -262,7 +287,7 @@ pico_file_browse(pdata, dir, fn, sz, flags)
 
     sprintf(title_buf, "%s FILE", pdata->pine_anchor);
     set_browser_title(title_buf);
-    rv = FileBrowse(dir, fn, sz, flags);
+    rv = FileBrowse(dir, dirlen, fn, fnlen, sz, flags);
     set_browser_title(NULL);
     vttidy();			/* clean up tty handling */
     zotdisplay();		/* and display structures */
@@ -277,7 +302,9 @@ pico_file_browse(pdata, dir, fn, sz, flags)
  *
  *	    intput:  
  *		     dir points to initial dir to browse.
- *		     fn  initial file name. 
+ *                   dirlen- buffer size of dir
+ *		     fn  initial file name.
+ *                   fnlen- buffer size of fn 
  *
  *         returns:
  *                   dir points to currently selected directory (without
@@ -290,14 +317,14 @@ pico_file_browse(pdata, dir, fn, sz, flags)
  *                   0 if no files seleted
  *                  -1 if there where problems
  */
-FileBrowse(dir, fn, sz, flags)
+FileBrowse(dir, dirlen, fn, fnlen, sz, flags)
 char *dir, *fn, *sz;			/* dir, name and optional size */
-int   flags;
+int   dirlen, fnlen, flags;
 {
     int status, i, j, c, new_c;
     int row, col, crow, ccol;
-    char *p, child[NLINE], tmp[NLINE];
-    struct bmaster *mp, *getfcells();
+    char *p, *envp, child[NLINE], tmp[NLINE];
+    struct bmaster *mp;
     struct fcell *tp;
 #ifdef MOUSE
     MOUSEPRESS mousep;
@@ -319,6 +346,14 @@ int   flags;
     /* build contents of cell structures */
     if((gmp = getfcells(dir)) == NULL)
       return(-1);
+    
+    tp = NULL;
+    if(fn && *fn){
+	if((tp = FindCell(gmp, fn)) != NULL){
+	    gmp->current = tp;
+	    PlaceCell(gmp, gmp->current, &row, &col);
+	}
+    }
 
     /* paint screen */
     PaintBrowser(gmp, 0, &crow, &ccol);
@@ -349,7 +384,7 @@ int   flags;
 	(*term.t_flush)();
 
 #ifdef MOUSE
-	mouse_in_content(K_MOUSE, -1, -1, 0, 0);
+	mouse_in_content(KEY_MOUSE, -1, -1, 0, 0);
 	register_mfunc(mouse_in_content,2,0,term.t_nrow-(term.t_mrow+1),
 	    term.t_ncol);
 #endif
@@ -361,6 +396,8 @@ int   flags;
 	if(Pmaster){
 	    if(Pmaster->newmail && (c == NODATA || time_to_check())){
 		if((*Pmaster->newmail)(c == NODATA ? 0 : 2, 1) >= 0){
+		    int rv;
+		    
 		    if(km_popped){		/* restore display */
 			km_popped = 0;
 			PaintBrowser(gmp, 0, &crow, &ccol);
@@ -368,7 +405,12 @@ int   flags;
 
 		    clearcursor();
 		    mlerase();
-		    (*Pmaster->showmsg)(c);
+		    rv = (*Pmaster->showmsg)(c);
+		    ttresize();
+		    picosigs();
+		    if(rv)	/* Did showmsg corrupt the display? */
+		      PaintBrowser(gmp, 0, &crow, &ccol); /* Yes, repaint */
+
 		    mpresf = 1;
 		}
 
@@ -377,7 +419,7 @@ int   flags;
 	    }
 	}
 	else{
-	    if(timeout && (c == NODATA || time_to_check()))
+	    if(timeo && (c == NODATA || time_to_check()))
 	      if(pico_new_mail())
 		emlwrite("You may possibly have new mail.", NULL);
 	}
@@ -402,6 +444,9 @@ int   flags;
 
 	if(c == NODATA)				/* GetKey timed out */
 	  continue;
+	else if(Pmaster)
+	  (*Pmaster->keybinput)();
+	  
 
 	if(mpresf){				/* blast old messages */
 	    if(mpresf++ > MESSDELAY){		/* every few keystrokes */
@@ -412,7 +457,7 @@ int   flags;
 						/* process commands */
 	switch(new_c = normalize_cmd(c,bfmappings[(gmode&MDBRONLY)?0:1],2)){
 
-	  case K_PAD_RIGHT:			/* move right */
+	  case KEY_RIGHT:			/* move right */
 	  case (CTRL|'@'):
 	  case (CTRL|'F'):			/* forward  */
 	  case 'n' :
@@ -435,7 +480,7 @@ int   flags;
 	    }
 	    break;
 
-	  case K_PAD_LEFT:				/* move left */
+	  case KEY_LEFT:				/* move left */
 	  case (CTRL|'B'):				/* back */
 	  case 'p' :
 	  case 'P' :
@@ -502,8 +547,8 @@ int   flags;
 
 	  case (CTRL|'V'):				/* page forward */
 	  case ' ':
-	  case K_PAD_NEXTPAGE :
-	  case K_PAD_END :
+	  case KEY_PGDN :
+	  case KEY_END :
 	    tp = gmp->top;
 	    i = term.t_nrow - term.t_mrow - 2;
 
@@ -531,8 +576,8 @@ int   flags;
 
 	  case '-' :
 	  case (CTRL|'Y'):				/* page backward */
-	  case K_PAD_PREVPAGE :
-	  case K_PAD_HOME :
+	  case KEY_PGUP :
+	  case KEY_HOME :
 	    tp = gmp->top;
 	    i = term.t_nrow - term.t_mrow - 4;
 	    while(i-- && tp != NULL){
@@ -564,7 +609,7 @@ int   flags;
 
 	    break;
 
-	  case K_PAD_DOWN :
+	  case KEY_DOWN :
 	  case (CTRL|'N'):				/* next */
 	    tp = gmp->current;
 	    i = gmp->fpl;
@@ -592,7 +637,7 @@ int   flags;
 	    }
 	    break;
 
-	  case K_PAD_UP :
+	  case KEY_UP :
 	  case (CTRL|'P'):				/* previous */
 	    tp = gmp->current;
 	    i = gmp->fpl;
@@ -616,7 +661,7 @@ int   flags;
 	    break;
 
 #ifdef MOUSE	    
-	  case K_MOUSE:
+	  case KEY_MOUSE:
 	    mouse_get_last (NULL, &mousep);
 	    if (mousep.doubleclick) {
 		goto Selected;
@@ -656,12 +701,16 @@ int   flags;
 		    break;
 		}
 
-		sprintf(tmp, "pico%s%s%s %s",
-			(gmode & MDFKEY) ? " -f" : "",
-			(gmode & MDSHOCUR) ? " -g" : "",
-			(gmode & MDMOUSE) ? " -m" : "",
-			child);
-		BrowserRunChild(tmp);		/* spawn pico */
+		if(envp = (char *) getenv("EDITOR"))
+		  sprintf(tmp, "%s \'%s\'", envp, child);
+		else
+		  sprintf(tmp, "pico%s%s%s \'%s\'",
+			  (gmode & MDFKEY) ? " -f" : "",
+			  (gmode & MDSHOCUR) ? " -g" : "",
+			  (gmode & MDMOUSE) ? " -m" : "",
+			  child);
+
+		BrowserRunChild(tmp, gmp->dname);	/* spawn pico */
 		PaintBrowser(gmp, 0, &crow, &ccol);	/* redraw browser */
 	    }
 	    else{
@@ -728,7 +777,7 @@ int   flags;
 			break;
 		    }
 
-		    BrowserRunChild(tmp);
+		    BrowserRunChild(tmp, gmp->dname);
 		    PaintBrowser(gmp, 0, &crow, &ccol);
 		    break;
 		  default:
@@ -758,7 +807,7 @@ int   flags;
 	    i = 0;
 	    while(i++ < 2){		/* verify twice!! */
 		if(i == 1){
-		    if(fexist(child, "w", (long *)NULL) != FIOSUC)
+		    if(fexist(child, "w", (off_t *)NULL) != FIOSUC)
 		      strcpy(tmp, "File is write protected! OVERRIDE");
 		    else
 		      sprintf(tmp, "Delete file \"%.*s\"", NLINE - 20, child);
@@ -822,9 +871,17 @@ int   flags;
 		}
 	    }
 
-	    if(Pmaster)
-	      (*Pmaster->helper)(Pmaster->browse_help,
-				 "Help for Browsing", 1);
+	    if(Pmaster){
+		VARS_TO_SAVE *saved_state;
+
+		saved_state = save_pico_state();
+		(*Pmaster->helper)(Pmaster->browse_help,
+				   "Help for Browsing", 1);
+		if(saved_state){
+		    restore_pico_state(saved_state);
+		    free_pico_state(saved_state);
+		}
+	    }
 	    else if(gmode&MDBRONLY)
 	      pico_help(sa_BrowseHelpText, "Browser Help", 1);
 	    else
@@ -880,11 +937,12 @@ int   flags;
 		    }
  
 		    if((gmode&MDTREE) && !in_oper_tree(child)){
-			emlwrite("Attempt to Goto directory denied", NULL);
+		      emlwrite("\007 Can't go outside of %s in restricted mode",
+			       opertree);
 			break;
 		    }
  
-		    if(isdir(child, (long *) NULL)){
+		    if(isdir(child, (long *) NULL, NULL)){
 			if((mp = getfcells(child)) == NULL){
 			    /* getfcells should explain what happened */
 			    i++;
@@ -905,7 +963,118 @@ int   flags;
 	    }
 	    BrowserKeys();
 	    break;
- 
+
+	  case 'a':					/* Add */
+	  case 'A':
+	    if(gmode&MDSCUR){				/* not allowed! */
+		emlwrite("Add not allowed in restricted mode",NULL);
+		break;
+	    }
+
+	    i = 0;
+	    child[0] = '\0';
+
+	    while(!i){
+
+		switch(status=mlreply("Name of file to add: ", child, NLINE,
+				      QFFILE, NULL)){
+		  case HELPCH:
+		    emlwrite("\007No help yet!", NULL);
+/* remove break and sleep after help text is installed */
+		    sleep(3);
+		    break;
+		  case (CTRL|'L'):
+		    PaintBrowser(gmp, 0, &crow, &ccol);
+		    break;
+		  case ABORT:
+		    emlwrite("Add File Cancelled", NULL);
+		    i++;
+		    break;
+		  case FALSE:
+		    i++;
+		    mlerase();
+		    break;
+		  case TRUE:
+		    i++;
+
+		    if(child[0] == '\0'){
+			emlwrite("No file named.  Add Cancelled.", NULL);
+			break;
+		    }
+
+		    if(!compresspath(gmp->dname, child, NLINE)){
+			emlwrite("Too many ..'s in name", NULL);
+			break;
+		    }
+
+		    if((gmode&MDTREE) && !in_oper_tree(child)){
+		       emlwrite("\007Restricted mode allows Add in %s only",
+				opertree);
+			break;
+		    }
+
+		    if((status = fexist(child, "w", (off_t *)NULL)) == FIOSUC){
+			sprintf(tmp,"File \"%.*s\" already exists!",
+				NLINE - 20, child);
+			emlwrite(tmp, NULL);
+			break;
+		    }
+		    else if(status != FIOFNF){
+			fioperr(status, child);
+			break;
+		    }
+
+		    if(ffwopen(child, FALSE) != FIOSUC){
+			/* ffwopen should've complained */
+			break;
+		    }
+		    else{			/* highlight new file */
+			ffclose();
+			emlwrite("Added File \"%s\"", child);
+
+			if((p = strrchr(child, C_FILESEP)) == NULL){
+			    emlwrite("Problems refiguring browser", NULL);
+			    break;
+			}
+
+			*p = '\0';
+			if(p != child){
+			    strcpy(tmp, child);
+			    j = 0;
+			    while(child[j++] = *++p)
+			      ;
+			}
+			else
+			  strcpy(tmp, S_FILESEP);
+
+			/*
+			 * new file in same dir? if so, refigure files
+			 * and redraw...
+			 */
+			if(!strcmp(tmp, gmp->dname)){ 
+			    if((mp = getfcells(gmp->dname)) == NULL)
+			      /* getfcells should explain what happened */
+			      break;
+
+			    zotmaster(&gmp);
+			    gmp = mp;
+			    if((tp = FindCell(gmp, child)) != NULL){
+				gmp->current = tp;
+				PlaceCell(gmp, gmp->current, &row, &col);
+			    }
+
+			    PaintBrowser(gmp, 1, &crow, &ccol);
+			}
+		    }
+		    break;
+		  default:
+		    break;
+		}
+	    }
+
+	    BrowserKeys();
+	    break;
+
 	  case 'c':					/* copy */
 	  case 'C':
 	    if(gmp->current->mode == FIODIR){
@@ -954,10 +1123,18 @@ int   flags;
 			break;
 		    }
 
-		    strcpy(tmp, child); 		/* add full path! */
-		    sprintf(child, "%s%c%s", gmp->dname, C_FILESEP, tmp);
+		    if(!compresspath(gmp->dname, child, NLINE)){
+			emlwrite("Too many ..'s in name", NULL);
+			break;
+		    }
 
-		    if((status = fexist(child, "w", (long *)NULL)) == FIOSUC){
+		    if((gmode&MDTREE) && !in_oper_tree(child)){
+		       emlwrite("\007Restricted mode allows Copy in %s only",
+				opertree);
+			break;
+		    }
+
+		    if((status = fexist(child, "w", (off_t *)NULL)) == FIOSUC){
 			sprintf(tmp,"File \"%.*s\" exists! OVERWRITE",
 				NLINE - 20, child);
 			if((status = mlyesno(tmp, 0)) != TRUE){
@@ -1022,7 +1199,6 @@ int   flags;
 	  case 'r':					/* rename */
 	  case 'R':
 	    i = 0;
-	    child[0] = '\0';
 
 	    if(!strcmp(gmp->current->fname, "..")){
 		emlwrite("\007Can't rename \"..\"", NULL);
@@ -1034,6 +1210,8 @@ int   flags;
 		break;
 	    }
 
+	    strcpy(child, gmp->current->fname);
+
 	    while(!i){
 
 		switch(status=mlreply("Rename file to: ", child, NLINE, QFFILE,
@@ -1059,10 +1237,18 @@ int   flags;
 			break;
 		    }
 
-		    strcpy(tmp, child);
-		    sprintf(child, "%s%c%s", gmp->dname, C_FILESEP, tmp);
+		    if(!compresspath(gmp->dname, child, NLINE)){
+			emlwrite("Too many ..'s in name", NULL);
+			break;
+		    }
 
-		    status = fexist(child, "w", (long *)NULL);
+		    if((gmode&MDTREE) && !in_oper_tree(child)){
+		       emlwrite("\007Restricted mode allows Rename in %s only",
+				opertree);
+			break;
+		    }
+
+		    status = fexist(child, "w", (off_t *)NULL);
 		    if(status == FIOSUC || status == FIOFNF){
 			if(status == FIOSUC){
 			    sprintf(tmp,"File \"%.*s\" exists! OVERWRITE",
@@ -1203,23 +1389,30 @@ int   flags;
 			gmp->current->fname);
 
 		if(LikelyASCII(child)){
-		    char *pg = (char *) getenv("PAGER");
-
-		    sprintf(tmp, "%s %s", pg ? pg : BROWSER_PAGER, child);
-		    BrowserRunChild(tmp);
+		    sprintf(tmp, "%s \'%s\'",
+			    (envp = (char *) getenv("PAGER"))
+			      ? envp : BROWSER_PAGER, child);
+		    BrowserRunChild(tmp, gmp->dname);
 		    PaintBrowser(gmp, 0, &crow, &ccol);
 		}
 
 		break;
 	    }
 	    else{				/* just return */
-		strcpy(dir, gmp->dname);
-		strcpy(fn, gmp->current->fname);
-		if(sz != NULL)			/* size uninteresting */
-		  strcpy(sz, gmp->current->size);
-
-		zotmaster (&gmp);
-		return (1);
+	      if ((strlen(gmp->dname) < dirlen) && 
+		  (strlen(gmp->current->fname) < fnlen)){
+		  strcpy(dir, gmp->dname);
+		  strcpy(fn, gmp->current->fname);
+	      }
+	      else {
+		zotmaster(&gmp);
+		return(-1);
+	      }
+	      if(sz != NULL)			/* size uninteresting */
+		strcpy(sz, gmp->current->size);
+	      
+	      zotmaster (&gmp);
+	      return (1);
 	    }
 	    break;
 
@@ -1353,10 +1546,12 @@ getfcells(dname)
     char *dname;
 {
     int  i, 					/* various return codes */
+         flength,
          nentries = 0;				/* number of dir ents */
-    long l;
+    off_t attsz;
     char *np,					/* names of files in dir */
-         *dcp,					/* to add file to path */
+         *dcp,                                  /* to add file to path */
+         *tmpstr,
          **filtnames,				/* array filtered names */
 	 errbuf[NLINE];
     struct fcell *ncp,				/* new cell pointer */
@@ -1416,6 +1611,8 @@ getfcells(dname)
     i = 0;					/* index of filt'd array */
     np = mp->names;
     while(nentries--){
+	int ii;
+
 	/*
 	 * Filter dot files?  Always filter ".", never filter "..",
 	 * and sometimes fitler ".*"...
@@ -1428,9 +1625,9 @@ getfcells(dname)
 
 	filtnames[i++] = np;
 
-	if((l = strlen(np)) > mp->longest)	/* cast l ? */
-	  mp->longest = l;			/* remember longest */
-	np += l + 1;				/* advance name pointer */
+	if((ii = (int)strlen(np)) > mp->longest)
+	  mp->longest = ii;			/* remember longest */
+	np += ii + 1;				/* advance name pointer */
 
     }
     nentries = i;				/* new # of entries */
@@ -1477,10 +1674,24 @@ getfcells(dname)
 	/* fill in the new cell */
 	ncp->fname = filtnames[i++];
 
-	strcpy(&dcp[1], ncp->fname);		/* use abolute path! */
-
 	/* fill in file's mode */
-	switch(fexist(mp->dname, "t", &l)){
+	if ((flength = strlen(ncp->fname) + 1 + strlen(dname)) < NLINE){
+	  strcpy(&dcp[1], ncp->fname);     /* use absolute path! */
+	  tmpstr = mp->dname;
+	}
+	else{
+	  if((tmpstr = (char *)malloc((flength+1)*sizeof(char))) == NULL){
+	    emlwrite("\007Can't malloc cells for temp buffer!", NULL);
+            zotfcells(mp->head);                /* clean up cells */
+            free((char *) filtnames);
+            free((char *) mp);
+            return(NULL);                       /* bummer. */
+	  }
+	  strcpy(tmpstr, dname);
+	  tmpstr = strcat(tmpstr, S_FILESEP);
+	  tmpstr = strcat(tmpstr, ncp->fname);
+	}
+	switch(fexist(tmpstr, "t", &attsz)){
 	  case FIODIR :
 	    ncp->mode = FIODIR;
 	    sprintf(ncp->size, "(%sdir)",
@@ -1493,9 +1704,11 @@ getfcells(dname)
 	    break;
 	  default :
 	    ncp->mode = FIOSUC;			/* regular file */
-	    strcpy(ncp->size,  prettysz(l));
+	    strcpy(ncp->size,  prettysz(attsz));
 	    break;
 	}
+	if (flength >= NLINE)
+	  free((char *) tmpstr);
     }
 
     dcp[(dcp == mp->dname) ? 1 : 0] = '\0';	/* remember to cap dname */
@@ -1641,6 +1854,8 @@ BrowserKeys()
     else{
 	menu_browse[SELECT_KEY].name  = "S";
 	menu_browse[SELECT_KEY].label = "[Select]";
+	menu_browse[PICO_KEY].name  = "A";
+	menu_browse[PICO_KEY].label = "Add";
     }
 
     wkeyhelp(menu_browse);
@@ -1666,6 +1881,9 @@ struct bmaster *mp;
 
 	mp->fpl = i - 1;			/* files per line */
     }
+
+    if(mp->fpl == 0)
+      mp->fpl = 1;
 }
 
 
@@ -1903,7 +2121,7 @@ char *dir;
 	while((p = strchr(p, C_FILESEP)) && (l-(p-dir)) > term.t_ncol-i-21)
 	  p++;
 
-	if(!*p)					/* no suitable length! */
+	if(!p)					/* no suitable length! */
 	  p = &dir[l-(term.t_ncol-i-19)];
 
 	sprintf(buf, "%s Dir ...%s", (gmode&MDBRONLY) ? "" : " BROWSER  ", p);
@@ -1958,18 +2176,33 @@ ClearBrowserScreen()
 
 
 void
-BrowserRunChild(child)
+BrowserRunChild(child, dir)
     char *child;
+    char *dir;
 {
-    int  status;
-    char tmp[NLINE];
+    int    status;
+    char   tmp[NLINE];
+    time_t t_in, t_out;
 
     ClearBrowserScreen();
     movecursor(0, 0);
     (*term.t_close)();
+    if(!isdir(dir, NULL, &t_in))
+      t_in = 0;
+
     fflush(stdout);
     status = system(child);
     (*term.t_open)();
+    if(t_in && isdir(dir, NULL, &t_out) && t_in < t_out){
+	struct bmaster *mp;
+
+	if(mp = getfcells(dir)){
+	    zotmaster(&gmp);
+	    gmp = mp;
+	}
+	/* else getfcells should explain what happened */
+    }
+
     /* complain about non-zero exit status */
     if((status >> 8) & 0xff){
 
