@@ -1,5 +1,5 @@
 #if	!defined(lint) && !defined(DOS)
-static char rcsid[] = "$Id: display.c,v 4.27 1994/10/10 21:56:33 mikes Exp $";
+static char rcsid[] = "$Id: display.c,v 4.88 1998/12/29 18:20:20 hubert Exp $";
 #endif
 /*
  * Program:	Display functions
@@ -15,28 +15,16 @@ static char rcsid[] = "$Id: display.c,v 4.27 1994/10/10 21:56:33 mikes Exp $";
  *
  * Please address all bugs and comments to "pine-bugs@cac.washington.edu"
  *
- * Copyright 1991-1994  University of Washington
  *
- *  Permission to use, copy, modify, and distribute this software and its
- * documentation for any purpose and without fee to the University of
- * Washington is hereby granted, provided that the above copyright notice
- * appears in all copies and that both the above copyright notice and this
- * permission notice appear in supporting documentation, and that the name
- * of the University of Washington not be used in advertising or publicity
- * pertaining to distribution of the software without specific, written
- * prior permission.  This software is made available "as is", and
- * THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
- * WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT LIMITATION ALL IMPLIED
- * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND IN
- * NO EVENT SHALL THE UNIVERSITY OF WASHINGTON BE LIABLE FOR ANY SPECIAL,
- * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
- * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT
- * (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION
- * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
- *
- * Pine and Pico are trademarks of the University of Washington.
- * No commercial use of these trademarks may be made without prior
- * written permission of the University of Washington.
+ * Pine and Pico are registered trademarks of the University of Washington.
+ * No commercial use of these trademarks may be made without prior written
+ * permission of the University of Washington.
+ * 
+ * Pine, Pico, and Pilot software and its included text are Copyright
+ * 1989-1998 by the University of Washington.
+ * 
+ * The full text of our legal notices is contained in the file called
+ * CPYRIGHT, included with this distribution.
  *
  */
 /*
@@ -47,12 +35,7 @@ static char rcsid[] = "$Id: display.c,v 4.27 1994/10/10 21:56:33 mikes Exp $";
  *
  */
 
-#include        <stdio.h>
-#include	"osdep.h"
-#include        "pico.h"
-#include	"estruct.h"
-#include        "edef.h"
-#include        "efunc.h"
+#include	"headers.h"
 
 
 #ifdef	ANSI
@@ -88,39 +71,48 @@ static char rcsid[] = "$Id: display.c,v 4.27 1994/10/10 21:56:33 mikes Exp $";
  * Standard pico keymenus...
  */
 static KEYMENU menu_pico[] = {
-    {"^G", "Get Help"},	{"^O", "WriteOut"}, 	{"^R", "Read File"},
-    {"^Y", "Prev Pg"},	{"^K", "Cut Text"},	{"^C", "Cur Pos"},
-    {"^X", "Exit"},		{"^J", "Justify"},	{"^W", "Where is"},
+    {"^G", "Get Help", KS_SCREENHELP},	{"^O", "WriteOut", KS_SAVEFILE},
+    {"^R", "Read File", KS_READFILE},	{"^Y", "Prev Pg", KS_PREVPAGE},
+    {"^K", "Cut Text", KS_NONE},	{"^C", "Cur Pos", KS_CURPOSITION},
+    {"^X", "Exit", KS_EXIT},		{"^J", "Justify", KS_JUSTIFY},
+    {"^W", "Where is", KS_WHEREIS},	{"^V", "Next Pg", KS_NEXTPAGE},
+    {"^U", NULL, KS_NONE},
 #ifdef	SPELLER
-    {"^V", "Next Pg"},	{"^U", NULL},	{"^T", "To Spell"}
+    {"^T", "To Spell", KS_SPELLCHK}
 #else
-    {"^V", "Next Pg"},	{"^U", NULL},	{"^D", "Del Char"}
+    {"^D", "Del Char", KS_NONE}
 #endif
 };
 #define	UNCUT_KEY	10
 
 
 static KEYMENU menu_compose[] = {
-    {"^G", "Get Help"},	{"^X", "Send"}, 	{"^R", "Read File"},
-    {"^Y", "Prev Pg"},	{"^K", "Cut Text"},	{"^O", "Postpone"},
-    {"^C", "Cancel"},	{"^J", "Justify"},	{NULL, NULL},
+    {"^G", "Get Help", KS_SCREENHELP},	{"^X", NULL, KS_SEND},
+    {"^R", "Read File", KS_READFILE},	{"^Y", "Prev Pg", KS_PREVPAGE},
+    {"^K", "Cut Text", KS_NONE},	{"^O", "Postpone", KS_POSTPONE},
+    {"^C", "Cancel", KS_CANCEL},	{"^J", "Justify", KS_JUSTIFY},
+    {NULL, NULL, KS_NONE},		{"^V", "Next Pg", KS_NEXTPAGE},
+    {"^U", NULL, KS_NONE},
 #ifdef	SPELLER
-    {"^V", "Next Pg"},	{"^U", NULL},		{"^T", "To Spell"}
+    {"^T", "To Spell", KS_SPELLCHK}
 #else
-    {"^V", "Next Pg"},	{"^U", NULL},		{"^D", "Del Char"}
+    {"^D", "Del Char", KS_NONE}
 #endif
 };
+#define	EXIT_KEY	1
+#define	PSTPN_KEY	5
 #define	WHERE_KEY	8
 
 
+/*
+ * Definition's for pico's modeline
+ */
+#define	PICO_TITLE	"  UW PICO(tm) %s  "
+#define	PICO_MOD_MSG	"Modified  "
+#define	PICO_NEWBUF_MSG	" New Buffer "
 
 #define WFDEBUG 0                       /* Window flag debug. */
 
-typedef struct  VIDEO {
-        short   v_flag;                 /* Flags */
-        CELL    v_text[1];              /* Screen data. */
-}       VIDEO;
-
 #define VFCHG   0x0001                  /* Changed flag			*/
 #define	VFEXT	0x0002			/* extended (beyond column 80)	*/
 #define	VFREV	0x0004			/* reverse video status		*/
@@ -128,14 +120,17 @@ typedef struct  VIDEO {
 
 int     vtrow   = 0;                    /* Row location of SW cursor */
 int     vtcol   = 0;                    /* Column location of SW cursor */
-int     ttrow   = HUGE;                 /* Row location of HW cursor */
-int     ttcol   = HUGE;                 /* Column location of HW cursor */
+int     ttrow   = FARAWAY;              /* Row location of HW cursor */
+int     ttcol   = FARAWAY;              /* Column location of HW cursor */
 int	lbound	= 0;			/* leftmost column of current line
 					   being displayed */
 
 VIDEO   **vscreen;                      /* Virtual screen. */
 VIDEO   **pscreen;                      /* Physical screen. */
 
+#define	ISCONTROL(C)	((C) < 0x20 || (C) == 0x7F \
+			 || ((gmode & P_HICTRL) && ((C) > 0x7F && (C) < 0xA0)))
+
 
 /*
  * Initialize the data structures used by the display code. The edge vectors
@@ -149,6 +144,9 @@ vtinit()
     register int i;
     register VIDEO *vp;
 
+    if(Pmaster == NULL)
+      vtterminalinfo(gmode & MDTCAPWINS);
+
     (*term.t_open)();
 
     (*term.t_rev)(FALSE);
@@ -177,6 +175,8 @@ vtinit()
 		     NULL);
             return(FALSE);
 	}
+	else
+	  memset(vp, ' ', sizeof(VIDEO)+(term.t_ncol*sizeof(CELL)));
 
 	vp->v_flag = 0;
         vscreen[i] = vp;
@@ -195,6 +195,8 @@ vtinit()
 		     NULL);
             return(FALSE);
 	}
+	else
+	  memset(vp, ' ', sizeof(VIDEO)+(term.t_ncol*sizeof(CELL)));
 
 	vp->v_flag = 0;
         pscreen[i] = vp;
@@ -203,6 +205,13 @@ vtinit()
     return(TRUE);
 }
 
+vtterminalinfo(termcap_wins)
+    int termcap_wins;
+{
+    return((term.t_terminalinfo) ? (*term.t_terminalinfo)(termcap_wins)
+				 : (Pmaster ? 0 : TRUE));
+}
+
 
 /*
  * Clean up the virtual terminal system, in anticipation for a return to the
@@ -260,10 +269,10 @@ CELL c;
 	}
         while ((vtcol&0x07) != 0);
     }
-    else if (c.c < 0x20 || c.c == 0x7F) {
+    else if (ISCONTROL(c.c)){
 	ac.c = '^';
         vtputc(ac);
-	ac.c = (c.c ^ 0x40);
+	ac.c = ((c.c & 0x7f) | 0x40);
         vtputc(ac);
     }
     else
@@ -296,10 +305,10 @@ CELL c;
         }
         while (((vtcol + lbound)&0x07) != 0 && vtcol < term.t_ncol);
     }
-    else if (c.c < 0x20 || c.c == 0x7F) {
+    else if (ISCONTROL(c.c)){
 	ac.c = '^';
         vtpute(ac);
-	ac.c = (c.c ^ 0x40);
+	ac.c = ((c.c & 0x7f) | 0x40);
         vtpute(ac);
     }
     else {
@@ -455,7 +464,7 @@ update()
 			 * to the top, and the composer is not being 
 			 * displayed...
 			 */
-		    if(Pmaster && !ComposerEditing 
+		    if(Pmaster && Pmaster->headents && !ComposerEditing 
 		       && (lback(lp) == wp->w_bufp->b_linep)
 		       && (ComposerTopLine == COMPOSER_TOP_LINE))
 		      j += entry_line(1000, TRUE); /* Never > 1000 headers */
@@ -493,10 +502,10 @@ out:
 	     * if the line at the top of the page is the top line
 	     * in the body, show the header...
 	     */
-	    if(Pmaster && !ComposerEditing){
+	    if(Pmaster && Pmaster->headents && !ComposerEditing){
 		if(lback(wp->w_linep) == wp->w_bufp->b_linep){
 		    if(ComposerTopLine == COMPOSER_TOP_LINE){
-			i = term.t_nrow - 4 - HeaderLen();
+			i = term.t_nrow - 2 - term.t_mrow - HeaderLen();
 			if(i > 0 && nlforw() >= i) {	/* room for header ? */
 			    if((i = nlforw()/2) == 0 && term.t_nrow&1)
 			      i = 1;
@@ -589,13 +598,13 @@ out:
 
         if (c.c == '\t')
             curcol |= 0x07;
-        else if (c.c < 0x20 || c.c == 0x7F)
+        else if (ISCONTROL(c.c))
             ++curcol;
 
         ++curcol;
     }
 
-    if (curcol >= term.t_ncol - 1) { 		/* extended line. */
+    if (curcol >= term.t_ncol) { 		/* extended line. */
 	/* flag we are extended and changed */
 	vscreen[currow]->v_flag |= VFEXT | VFCHG;
 	updext();				/* and output extended line */
@@ -617,7 +626,7 @@ out:
 		/* always flag extended lines as changed */
 		vscreen[i]->v_flag |= VFCHG;
 		if ((wp != curwp) || (lp != wp->w_dotp) ||
-		    (curcol < term.t_ncol - 1)) {
+		    (curcol < term.t_ncol)) {
 		    vtmove(i, 0);
 		    for (j = 0; j < llength(lp); ++j)
 		      vtputc(lgetc(lp, j));
@@ -641,10 +650,12 @@ out:
 
     if (sgarbf != FALSE){
 	if(Pmaster){
+	    int rv;
+       
 	    showCompTitle();
 
 	    if(ComposerTopLine != COMPOSER_TOP_LINE){
-		UpdateHeader();			/* arrange things */
+		UpdateHeader(0);		/* arrange things */
 		PaintHeader(COMPOSER_TOP_LINE, TRUE);
 	    }
 
@@ -658,10 +669,15 @@ out:
 		peeol();
 		vscreen[i]->v_flag |= VFCHG;
 	    }
+	    rv = (*Pmaster->showmsg)('X' & 0x1f);	/* ctrl-L */
+	    ttresize();
+	    picosigs();		/* restore altered handlers */
+	    if(rv)		/* Did showmsg corrupt the display? */
+	      PaintBody(0);	/* Yes, repaint */
 	    movecursor(wheadp->w_toprow, 0);
 	}
 	else{
-	    for (i = 0; i < term.t_nrow - 2; ++i){
+	    for (i = 0; i < term.t_nrow-term.t_mrow; ++i){
 		vscreen[i]->v_flag |= VFCHG;
 		vp1 = pscreen[i];
 		c.c = ' ';
@@ -688,14 +704,15 @@ out:
      * the "updateline" code is only called with a line that has been updated
      * for sure.
      */
-    if(Pmaster){
-	i = curwp->w_toprow;
-	c.c = term.t_nrow-2;
-    }
-    else{
-	i = 0;
-	c.c = term.t_nrow;
-    }
+    if(Pmaster)
+      i = curwp->w_toprow;
+    else
+      i = 0;
+
+    if (term.t_nrow > term.t_mrow)
+       c.c = term.t_nrow - term.t_mrow;
+    else
+       c.c = 0;
 
     for (; i < (int)c.c; ++i){
 
@@ -724,14 +741,17 @@ out:
     if(Pmaster == NULL){
 
 	if(sgarbk != FALSE){
-	    movecursor(term.t_nrow-1, 0);
-	    peeol();
-	    movecursor(term.t_nrow, 0);
-	    peeol();
+	    if(term.t_mrow > 0){
+		movecursor(term.t_nrow-1, 0);
+		peeol();
+		movecursor(term.t_nrow, 0);
+		peeol();
+	    }
+
 	    if(lastflag&CFFILL){
 		menu_pico[UNCUT_KEY].label = "UnJustify";
 		emlwrite("Can now UnJustify!", NULL);
-		mpresf = HUGE;	/* remove this after next keystroke! */
+		mpresf = FARAWAY;	/* remove this after next keystroke! */
 	    }
 	    else
 	      menu_pico[UNCUT_KEY].label = "UnCut Text";
@@ -746,10 +766,17 @@ out:
     movecursor(currow, curcol - lbound);
 #ifdef _WINDOWS
     mswin_endupdate ();
+
+    /* 
+     * Update the scroll bars.  This function is where curbp->b_linecnt
+     * is really managed.  See update_scroll.
+     */
+    update_scroll ();
 #endif
     (*term.t_flush)();
 }
 
+
 /* updext - update the extended line which the cursor is currently
  *	    on at a column greater than the terminal width. The line
  *	    will be scrolled right or left to let the user see where
@@ -863,7 +890,7 @@ short *flags;				/* and how we want it that way  */
 	    }
 
 	    if (cp7==cp2 && cp4-cp2 > 3){
-		o_insert((char)cp1->c);     /* insert the char */
+		o_insert((int)cp1->c);  /* insert the char */
 		display = FALSE;        /* only do it once!! */
 	    }
 	}
@@ -916,12 +943,27 @@ WINDOW *wp;
         if(ComposerEditing)
 	  ShowPrompt();
 	else{
+	    menu_compose[EXIT_KEY].label  = (Pmaster->headents)
+					      ? "Send" :"Exit";
+	    menu_compose[PSTPN_KEY].name  = (Pmaster->headents)
+					      ? "^O" : NULL;
+	    menu_compose[PSTPN_KEY].label = (Pmaster->headents)
+					      ? "Postpone" : NULL;
 	    menu_compose[WHERE_KEY].name  = (Pmaster->alt_ed) ? "^_" : "^W";
 	    menu_compose[WHERE_KEY].label = (Pmaster->alt_ed) ? "Alt Edit" 
 							      : "Where is";
-	    menu_compose[UNCUT_KEY].label = (thisflag&CFFILL) ? "UnFill"
+	    KS_OSDATASET(&menu_compose[WHERE_KEY],
+			 (Pmaster->alt_ed) ? KS_ALTEDITOR : KS_WHEREIS);
+	    menu_compose[UNCUT_KEY].label = (thisflag&CFFILL) ? "UnJustify"
 							      : "UnCut Text";
 	    wkeyhelp(menu_compose);
+#ifdef _WINDOWS
+	    /* When alt editor is available "Where is" is not on the menu
+	     * but the command is still available.  This call enables any
+	     * "Where is" menu items. */
+	    if (Pmaster->alt_ed)
+		mswin_menuitemadd (MENU|CTRL|'W', "", KS_WHEREIS, 0);
+#endif 
 	}
     }
     else{
@@ -935,10 +977,8 @@ WINDOW *wp;
 
 	n = 0;
 	c.a = 1;
-#if	defined(MSDOS) || defined(LINUX)
 	vtmove(1, 0);
 	vteeol();
-#endif
 	vscreen[n]->v_flag |= VFCHG; /* Redraw next time. */
 	vtmove(n, 0);		/* Seek to right line. */
 
@@ -955,50 +995,52 @@ WINDOW *wp;
 
 	n = 1;
 
-	sprintf((cp=tline), "  UW PICO(tm) %s  ", version);	/* VERSION */
-	
-	while ((c.c = *cp++) != 0) {
+	sprintf(cp = tline, PICO_TITLE, version);	/* write version */
+
+	while(c.c = *cp++){
 	    vtputc(c);
 	    ++n;
 	}
 
-	if (bp->b_fname[0] != 0){            /* File name. */
-	    sprintf(tline," File: ");
-	    cp = &bp->b_fname[0];
+	if(bp->b_fname[0]){				/* File name? */
+	    char *p, *endp, *prefix;
 
-	    if(strlen(cp) > (term.t_ncol-n-22)){   /* room for full path ? */
-		strcat(tline,".../");
-		while(strlen(cp) >= (term.t_ncol-n-22)){
-		    if(strchr(cp,'/') == NULL)
-		      cp = (char *)strchr(&bp->b_fname[0],'\0')-term.t_ncol-n-22;
-		    else
-		      cp = (char *)strchr(cp,'/') + 1;
+	    prefix = NULL;				/* for abreviation */
+	    endp   = strchr(cp = bp->b_fname, '\0');	/* find eol */
+	    i	   = term.t_ncol - n - 22;		/* space available */
+	    while(i > 0 && endp - cp > i){
+		if(!prefix){
+		    prefix = ".../";			/* path won't fit! */
+		    i -= 4;
 		}
+
+		if(!(p = strchr(cp, '/'))){		/* neat break? */
+		    cp = endp - i;			/* do best we can */
+		    break;
+		}
+		else
+		  cp = p + 1;
 	    }
 
-	    strcat(tline,cp);
+	    sprintf(tline, "%*.sFile: %s%s",
+		    ((i - (endp - cp)) > 0) ? (i - (endp - cp))/2 : 0, " ",
+		    prefix ? prefix : "", cp);
 
-	    c.c = ' ';
-	    for(i=((term.t_ncol-n-12-strlen(tline))/2); i > 0; n++, i--)
+	    for(cp = tline; c.c = *cp; cp++, n++)
 	      vtputc(c);
-
-	    cp = &tline[0];
-
-	    while ((c.c = *cp++) != 0){
-		vtputc(c);
-		++n;
-            }
-
         }
         else{
-	    cp = " New Buffer ";
-	    c.c = lchar;
-            for(i=(term.t_ncol-strlen(cp))/2; n < i; n++)
-	      vtputc(c);
-	    while ((c.c = *cp++) != 0) {
+	    cp = PICO_NEWBUF_MSG;
+	    if(sizeof(PICO_NEWBUF_MSG) < term.t_ncol){	/* enough room? */
+		c.c = lchar;
+		for(i = (term.t_ncol-sizeof(PICO_NEWBUF_MSG))/2; n < i; n++)
+		  vtputc(c);
+	    }
+
+	    while(c.c = *cp++){
 		vtputc(c);
 		++n;
-            }
+	    }
 	}
 
 #if WFDEBUG
@@ -1010,22 +1052,20 @@ WINDOW *wp;
 	vtputc((wp->w_flag&WFFORCE)!=0 ? 'F' : lchar);
 	n += 6;
 #endif
-
-	if ((bp->b_flag&BFCHG) != 0)                /* "MOD" if changed. */
-	  cp = "Modified  ";
-	else
-	  cp = "          ";
+	i = term.t_ncol - n;				/* space available */
+	cp = NULL;
+	if(bp->b_flag&BFCHG){				/* "MOD" if changed. */
+	    cp = PICO_MOD_MSG;
+	    i  = (i > sizeof(PICO_MOD_MSG)) ? i - sizeof(PICO_MOD_MSG) : 0;
+	}
 
 	c.c = lchar;
-	while (n < (term.t_ncol - strlen(cp))){	    /* Pad to full width. */
-	    vtputc(c);
-	    ++n;
-        }
+	while(i-- > 0)					/* Pad width */
+	  vtputc(c);
 
-	while ((c.c = *cp++) != 0){
+	if(cp)
+	  while(c.c = *cp++)
 	    vtputc(c);
-	    ++n;
-	}
     }
 }
 
@@ -1047,14 +1087,27 @@ int row, col;
 }
 
 
+/*
+ * Erase any sense we have of the cursor's HW location...
+ */
+clearcursor()
+{
+    ttrow = ttcol = FARAWAY;
+}
+
+
 /*
  * Erase the message line. This is a special routine because the message line
  * is not considered to be part of the virtual screen. It always works
  * immediately; the terminal buffer is flushed via a call to the flusher.
  */
+void
 mlerase()
 {
-    movecursor(term.t_nrow - 2, 0);
+    if (term.t_nrow < term.t_mrow)
+      return;
+
+    movecursor(term.t_nrow - term.t_mrow, 0);
     (*term.t_rev)(0);
     if (eolexist == TRUE)
       peeol();
@@ -1081,8 +1134,20 @@ int   dflt;
     char    buf[NLINE];
     KEYMENU menu_yesno[12];
 
-    for(rv = 0; rv < 12; rv++)
-      menu_yesno[rv].name = NULL;
+#ifdef _WINDOWS
+    if (mswin_usedialog ()) 
+      switch (mswin_yesno (prompt)) {
+	default:
+	case 0:		return (ABORT);
+	case 1:		return (TRUE);
+	case 2:		return (FALSE);
+      }
+#endif  
+
+    for(rv = 0; rv < 12; rv++){
+	menu_yesno[rv].name = NULL;
+	KS_OSDATASET(&menu_yesno[rv], KS_NONE);
+    }
 
     menu_yesno[1].name  = "Y";
     menu_yesno[1].label = (dflt == TRUE) ? "[Yes]" : "Yes";
@@ -1092,7 +1157,7 @@ int   dflt;
     menu_yesno[7].label = (dflt == FALSE) ? "[No]" : "No";
     wkeyhelp(menu_yesno);		/* paint generic menu */
     sgarbk = TRUE;			/* mark menu dirty */
-    if(Pmaster)
+    if(Pmaster && curwp)
       curwp->w_flag |= WFMODE;
 
     sprintf(buf, "%s ? ", prompt);
@@ -1131,6 +1196,21 @@ int   dflt;
 	    rv = FALSE;
 	    break;
 
+	  case (CTRL|'G') :
+	    if(term.t_mrow == 0 && km_popped == 0){
+		movecursor(term.t_nrow-2, 0);
+		peeol();
+		term.t_mrow = 2;
+		(*term.t_rev)(0);
+		wkeyhelp(menu_yesno);		/* paint generic menu */
+		mlwrite(buf, NULL);
+		(*term.t_rev)(1);
+		sgarbk = TRUE;			/* mark menu dirty */
+		km_popped++;
+		break;
+	    }
+	    /* else fall through */
+
 	  default:
 	    (*term.t_beep)();
 	  case NODATA :
@@ -1140,6 +1220,14 @@ int   dflt;
 	(*term.t_flush)();
 	if(rv != -1){
 	    (*term.t_rev)(0);
+	    if(km_popped){
+		term.t_mrow = 0;
+		movecursor(term.t_nrow, 0);
+		peeol();
+		sgarbf = 1;
+		km_popped = 0;
+	    }
+
 	    return(rv);
 	}
     }
@@ -1155,9 +1243,9 @@ int   dflt;
  * return. Handle erase, kill, and abort keys.
  */
 mlreply(prompt, buf, nbuf, flg, extras)
-char    *prompt, *buf;
-int      nbuf, flg;
-KEYMENU *extras;
+char	  *prompt, *buf;
+int	   nbuf, flg;
+EXTRAKEYS *extras;
 {
     return(mlreplyd(prompt, buf, nbuf, flg|QDEFLT, extras));
 }
@@ -1191,10 +1279,10 @@ static int rfkm[12][2] = {
  *            in file names.
  */
 mlreplyd(prompt, buf, nbuf, flg, extras)
-char    *prompt;
-char    *buf;
-int      nbuf, flg;
-KEYMENU *extras;
+char	  *prompt;
+char	  *buf;
+int	   nbuf, flg;
+EXTRAKEYS *extras;
 {
     register int    c;				/* current char       */
     register char   *b;				/* pointer in buf     */
@@ -1202,29 +1290,58 @@ KEYMENU *extras;
     register int    maxl;
     register int    plen;
     int      changed = FALSE;
+    int      return_val = 0;
     KEYMENU  menu_mlreply[12];
+    int	     extra_v[12];
+
+#ifdef _WINDOWS
+    if (mswin_usedialog ()) {
+	MDlgButton		btn_list[12];
+	int			i, j;
+
+	memset (&btn_list, 0, sizeof (MDlgButton) * 12);
+	j = 0;
+	for (i = 0; extras && extras[i].name != NULL; ++i) {
+	    if (extras[i].label[0] != '\0') {
+		if ((extras[i].key & CTRL) == CTRL) 
+		    btn_list[j].ch = (extras[i].key & ~CTRL) - '@';
+		else
+		    btn_list[j].ch = extras[i].key;
+		btn_list[j].rval = extras[i].key;
+		btn_list[j].name = extras[i].name;
+		btn_list[j++].label = extras[i].label;
+	    }
+	}
+	btn_list[j].ch = -1;
+
+	return (mswin_dialog (prompt, buf, nbuf, ((flg&QDEFLT) > 0), 
+			FALSE, btn_list, NULL, 0));
+    }
+#endif
 
     menu_mlreply[0].name = "^G";
     menu_mlreply[0].label = "Get Help";
+    KS_OSDATASET(&menu_mlreply[0], KS_SCREENHELP);
     for(j = 0, i = 1; i < 6; i++){	/* insert odd extras */
 	menu_mlreply[i].name = NULL;
+	KS_OSDATASET(&menu_mlreply[i], KS_NONE);
 	rfkm[2*i][1] = 0;
 	if(extras){
 	    for(; extras[j].name && j != 2*(i-1); j++)
 	      ;
 
 	    if(extras[j].name){
-		rfkm[2*i][1] = (extras[j].name[0] == '^')
-				 ? (CTRL | extras[j].name[1])
-				 : extras[j].name[1];
+		rfkm[2*i][1]	      = extras[j].key;
 		menu_mlreply[i].name  = extras[j].name;
 		menu_mlreply[i].label = extras[j].label;
+		KS_OSDATASET(&menu_mlreply[i], KS_OSDATAGET(&extras[j]));
 	    }
 	}
     }
 
     menu_mlreply[6].name = "^C";
     menu_mlreply[6].label = "Cancel";
+    KS_OSDATASET(&menu_mlreply[6], KS_NONE);
     for(j = 0, i = 7; i < 12; i++){	/* insert even extras */
 	menu_mlreply[i].name = NULL;
 	rfkm[2*(i-6)+1][1] = 0;
@@ -1233,17 +1350,20 @@ KEYMENU *extras;
 	      ;
 
 	    if(extras[j].name){
-		rfkm[2*(i-6)+1][1] = (extras[j].name[0] == '^')
-				       ? (CTRL | extras[j].name[1])
-				       : extras[j].name[0];
+		rfkm[2*(i-6)+1][1]    = extras[j].key;
 		menu_mlreply[i].name  = extras[j].name;
 		menu_mlreply[i].label = extras[j].label;
+		KS_OSDATASET(&menu_mlreply[i], KS_OSDATAGET(&extras[j]));
 	    }
 	}
     }
 
-    mlwrite(prompt, NULL);		/* paint prompt */
-    plen = strlen(prompt);
+    /* set up what to watch for and return values */
+    memset(extra_v, 0, sizeof(extra_v));
+    for(i = 0, j = 0; i < 12 && extras && extras[i].name; i++)
+      extra_v[j++] = extras[i].key;
+
+    plen = mlwrite(prompt, NULL);		/* paint prompt */
     if(!(flg&QDEFLT))
       *buf = '\0';
 
@@ -1252,7 +1372,7 @@ KEYMENU *extras;
     maxl = (nbuf < term.t_ncol - plen - 1) ? nbuf : term.t_ncol - plen - 1;
 
     pputs(buf, 1);
-    b = &buf[strlen(buf)];
+    b = &buf[(flg & QBOBUF) ? 0 : strlen(buf)];
     
     (*term.t_rev)(0);
     wkeyhelp(menu_mlreply);		/* paint generic menu */
@@ -1260,15 +1380,16 @@ KEYMENU *extras;
     (*term.t_rev)(1);
 
     for (;;) {
-	movecursor(term.t_nrow - 2, plen + b - buf);
+	movecursor(term.t_nrow - term.t_mrow, plen + b - buf);
 	(*term.t_flush)();
 
 
 	while((c = GetKey()) == NODATA)
 	  ;
 
-	switch(normal(c, rfkm, 2)){
+	switch(c = normalize_cmd(c, rfkm, 1)){
 	  case (CTRL|'A') :			/* CTRL-A beginning     */
+	  case KEY_HOME :
 	    b = buf;
 	    continue;
 
@@ -1278,13 +1399,15 @@ KEYMENU *extras;
 	    continue;
 
 	  case (CTRL|'C') :			/* CTRL-C abort		*/
-	    pputs("^C", 1);
+	    pputs("ABORT", 1);
 	    ctrlg(FALSE, 0);
 	    (*term.t_rev)(0);
 	    (*term.t_flush)();
-	    return(ABORT);
+	    return_val = ABORT;
+	    goto ret;
 
 	  case (CTRL|'E') :			/* CTRL-E end of line   */
+	  case KEY_END  :
 	    b = &buf[strlen(buf)];
 	    continue;
 
@@ -1294,10 +1417,25 @@ KEYMENU *extras;
 	    continue;
 
 	  case (CTRL|'G') :			/* CTRL-G help		*/
+	    if(term.t_mrow == 0 && km_popped == 0){
+		movecursor(term.t_nrow-2, 0);
+		peeol();
+		sgarbk = 1;			/* mark menu dirty */
+		km_popped++;
+		term.t_mrow = 2;
+		(*term.t_rev)(0);
+		wkeyhelp(menu_mlreply);		/* paint generic menu */
+		plen = mlwrite(prompt, NULL);		/* paint prompt */
+		(*term.t_rev)(1);
+		pputs(buf, 1);
+		break;
+	    }
+
 	    pputs("HELP", 1);
 	    (*term.t_rev)(0);
 	    (*term.t_flush)();
-	    return(HELPCH);
+	    return_val = HELPCH;
+	    goto ret;
 
 	  case (CTRL|'H') :			/* CTRL-H backspace	*/
 	  case 0x7f :				/*        rubout	*/
@@ -1308,6 +1446,7 @@ KEYMENU *extras;
 	    (*term.t_putchar)('\b');
 
 	  case (CTRL|'D') :			/* CTRL-D delete char   */
+	  case KEY_DEL :
 	    changed=TRUE;
 	    i = 0;
 	    do					/* blat out left char   */
@@ -1317,31 +1456,8 @@ KEYMENU *extras;
 
 	  case (CTRL|'L') :			/* CTRL-L redraw	*/
 	    (*term.t_rev)(0);
-	    return(CTRL|'L');
-
-	  case (CTRL|'T') :			/* CTRL-T special	*/
-	    (*term.t_rev)(0);
-	    return(CTRL|'T');
-
-	  case (CTRL|'V') :			/* CTRL-V special	*/
-	    if(flg & QPAGE){
-		(*term.t_rev)(0);
-		return(CTRL|'V');
-	    }
-	    else
-	      (*term.t_beep)();
-
-	    break;
-
-	  case (CTRL|'Y') :			/* CTRL-Y special	*/
-	    if(flg & QPAGE){
-		(*term.t_rev)(0);
-		return(CTRL|'Y');
-	    }
-	    else
-	      (*term.t_beep)();
-
-	    break;
+	    return_val = (CTRL|'L');
+	    goto ret;
 
 	  case (CTRL|'K') :			/* CTRL-K kill line	*/
 	    changed=TRUE;
@@ -1350,38 +1466,42 @@ KEYMENU *extras;
 	    movecursor(ttrow, plen);
 	    break;
 
-	  case K_PAD_LEFT:
+	  case KEY_LEFT:
 	    if(ttcol > plen)
 	      b--;
 	    continue;
 
-	  case K_PAD_RIGHT:
+	  case KEY_RIGHT:
 	    if(*b != '\0')
 	      b++;
 	    continue;
 
-	  case F3 :				/* abort */
-	    pputs("ABORT", 1);
-	    ctrlg(FALSE, 0);
-	    (*term.t_rev)(0);
-	    (*term.t_flush)();
-	    return(ABORT);
-
 	  case F1 :				/* sort of same thing */
 	    (*term.t_rev)(0);
 	    (*term.t_flush)();
-	    return(HELPCH);
+	    return_val = HELPCH;
+	    goto ret;
 
 	  case (CTRL|'M') :			/*        newline       */
 	    (*term.t_rev)(0);
 	    (*term.t_flush)();
-	    return(changed);
+	    return_val = changed;
+	    goto ret;
 
 	  default : 
 	    if(strlen(buf) >= maxl){		/* contain the text      */
 		(*term.t_beep)();
 		continue;
 	    }
+
+	    /* look for match in extra_v */
+	    for(i = 0; i < 12; i++)
+	      if(c && c == extra_v[i]){
+		  (*term.t_rev)(0);
+		  return_val = c;
+		  goto ret;
+	      }
+
 	    changed=TRUE;
 
 	    if(c&(~0xff)){			/* bag ctrl/special chars */
@@ -1395,11 +1515,19 @@ KEYMENU *extras;
 			continue;
 		    }
 		}
+		if(flg&QNODQT){	                /* reject double quotes? */
+		    if(c == '"'){
+			(*term.t_beep)();
+			continue;
+		    }
+		}
+		
+
 		do				/* blat out left char   */
 		  b[i+1] = b[i];
-		while(i-- >= 0);
-		*b++ = c;
-		pputc(c, 0);
+		while(i-- > 0);
+
+		pputc(*b++ = c, 0);
 	    }
 	}
 
@@ -1412,6 +1540,17 @@ KEYMENU *extras;
 	while(ttcol <= i)
 	  pputc(' ', 0);
     }
+
+ret:
+    if(km_popped){
+	term.t_mrow = 0;
+	movecursor(term.t_nrow, 0);
+	peeol();
+	sgarbf = 1;
+	km_popped = 0;
+    }
+
+    return(return_val);
 }
 
 
@@ -1432,8 +1571,8 @@ void	*arg;
 
     mlerase();
 
-    if((l = strlen(message)) == 0)		/* nothing to write, bag it */
-      return;
+    if((l = strlen(message)) == 0 || term.t_nrow < 2)	
+      return;    /* nothing to write or no space to write, bag it */
 
     /*
      * next, figure out where the to move the cursor so the message 
@@ -1464,10 +1603,10 @@ void	*arg;
 	}
     }
 
-    if(l-4 <= term.t_ncol)			/* this wouldn't be good */
-      movecursor(term.t_nrow-2, (term.t_ncol - (int)l - 4)/2);
+    if(l+4 <= term.t_ncol)
+      movecursor(term.t_nrow-term.t_mrow, (term.t_ncol - ((int)l + 4))/2);
     else
-      movecursor(term.t_nrow-2, 0);
+      movecursor(term.t_nrow-term.t_mrow, 0);
 
     (*term.t_rev)(1);
     pputs("[ ", 1);
@@ -1577,10 +1716,11 @@ void *arg;
     while(ttcol < term.t_ncol)
       pputc(' ', 0);
 
-    movecursor(term.t_nrow - 2, c);
+    movecursor(term.t_nrow - term.t_mrow, c);
     (*term.t_rev)(0);
     (*term.t_flush)();
     mpresf = TRUE;
+    return(c);
 }
 
 
@@ -1671,8 +1811,8 @@ int     r, n;
 	bcopy(vp1, vp2, term.t_ncol * sizeof(CELL));
     }
     pprints(r+n-1, r);
-    ttrow = HUGE;
-    ttcol = HUGE;
+    ttrow = FARAWAY;
+    ttcol = FARAWAY;
 #endif /* TERMCAP */
 }
 
@@ -1715,8 +1855,8 @@ int     r, n;
 	i++;
     }
     pprints(wp->w_toprow+wp->w_ntrows-n, wp->w_toprow+wp->w_ntrows-1);
-    ttrow = HUGE;
-    ttcol = HUGE;
+    ttrow = FARAWAY;
+    ttcol = FARAWAY;
 #endif /* TERMCAP */
 }
 
@@ -1777,7 +1917,7 @@ unsigned  *chs;
 	if(l >= 0)
 	  (*chs) += llength(lp);
     }
-    *r = i-l-2;
+    *r = i - l - term.t_mrow;
     return(l+curwp->w_toprow);
 }
 
@@ -1802,11 +1942,17 @@ int  row, col;
     if (old_ncol == term.t_ncol && old_nrow == term.t_nrow)
       return(TRUE);
 
-    curwp->w_toprow = 2;
-    curwp->w_ntrows = term.t_nrow-4; 	      /* "-4" for mode line and keys */
+    if(curwp){
+	curwp->w_toprow = 2;
+	curwp->w_ntrows = term.t_nrow - curwp->w_toprow - term.t_mrow;
+    }
 
-    if(Pmaster)
-      fillcol = (term.t_ncol > 80) ? 77 : term.t_ncol - 6;
+    if(Pmaster){
+	fillcol = Pmaster->fillcolumn;
+	(*Pmaster->resize)();
+    }
+    else if(userfillcol > 0)
+      fillcol = userfillcol;
     else
       fillcol = term.t_ncol - 6;	       /* we control the fill column */
 
@@ -1864,7 +2010,7 @@ int  row, col;
     }
 
     if(!ResizeBrowser()){
-	if(Pmaster){
+	if(Pmaster && Pmaster->headents){
 	    ResizeHeader();
 	}
 	else{
@@ -1878,6 +2024,13 @@ int  row, col;
     return(TRUE);
 }
 
+void
+redraw_pico_for_callback()
+{
+    refresh(0, 1);
+    update();
+}
+
 
 /*
  * showCompTitle - display the anchor line passed in from pine
@@ -1985,7 +2138,14 @@ peeol()
 
     cl.c = ' ';
     cl.a = 0;
-    (*term.t_eeol)();
+
+    /*
+     * Don't clear if we think we are sitting past the last column,
+     * that erases the last column if we just wrote it.
+     */
+    if(c < term.t_ncol)
+      (*term.t_eeol)();
+
     while(c < term.t_ncol && c >= 0 && r <= term.t_nrow && r >= 0)
       pscreen[r]->v_text[c++] = cl;
 }
@@ -2059,10 +2219,11 @@ CELL c;
     register int   i;
     register CELL *p;
 
-    if(o_insert((char)c.c)){		/* if we've got it, use it! */
+    if(o_insert((int) c.c)){		/* if we've got it, use it! */
 	p = pscreen[ttrow]->v_text;	/* then clean up physical screen */
 	for(i = term.t_ncol-1; i > ttcol; i--)
 	  p[i] = p[i-1];		/* shift right */
+
 	p[ttcol++] = c;			/* insert new char */
 	
 	return(1);
@@ -2125,9 +2286,12 @@ int      key;
 	if(pmt[j] == key)
 	  j++;
 
+	if (pscr(line, i) == NULL)
+	  return;
+	
 	if(pscr(line, i)->c != pmt[j]){
 	    if(j >= 1 && pmt[j-1] == key)
-	      j--;
+ 	      j--;
 	    break;
 	}
 
@@ -2165,7 +2329,17 @@ wkeyhelp(keymenu)
 KEYMENU *keymenu;
 {
     char *obufp, *p, fkey[4], linebuf[NLINE];
-    int   row, slot, tspace, nspace[6], index, n;
+    int   row, slot, tspace, adjusted_tspace, nspace[6], index, n;
+#ifdef	MOUSE
+    char  nbuf[NLINE];
+#endif
+
+#ifdef _WINDOWS
+    pico_config_menu_items (keymenu);
+#endif
+
+    if(term.t_mrow == 0)
+      return(FALSE);
 
     /*
      * Calculate amount of space for the names column by column...
@@ -2184,29 +2358,40 @@ KEYMENU *keymenu;
 	nspace[index] = (index < 4) ? 3 : 4;
 
     tspace = term.t_ncol/6;		/* total space for each item */
+    /*
+     * Avoid writing in bottom right corner so we won't scroll screens that
+     * scroll when you do that. The way this is setup, we won't do that
+     * unless the number of columns is evenly divisible by 6.
+     */
+    adjusted_tspace = (6 * tspace == term.t_ncol) ? tspace - 1 : tspace;
+
     index  = 0;
     for(row = 0; row <= 1; row++){
 	linebuf[0] = '\0';
 	obufp = &linebuf[0];
 	for(slot = 0; slot < 6; slot++){
 	    if(keymenu[index].name && keymenu[index].label){
-#if	defined(DOS) && defined(MOUSE)
-		register_key(index,
-			     (gmode&MDFKEY) ? F1 + (2 * slot) + row:
-			       (keymenu[index].name[0] == '^')
-				 ? (CTRL | keymenu[index].name[1])
-				 : keymenu[index].name[0],
-			     keymenu[index].label,
-			     term.t_nrow - 1 + row,
-			     (slot * tspace) + nspace[slot],
-			     strlen(keymenu[index].label));
-#endif
 		if(gmode&MDFKEY){
 		    p = fkey;
 		    sprintf(fkey, "F%d", (2 * slot) + row + 1);
 		}
 		else
 		  p = keymenu[index].name;
+#ifdef	MOUSE
+		sprintf(nbuf, "%.*s %s", nspace[slot], p,
+			keymenu[index].label);
+		register_key(index,
+			     (gmode&MDFKEY) ? F1 + (2 * slot) + row:
+			     (keymenu[index].name[0] == '^')
+			       ? (CTRL | keymenu[index].name[1])
+			       : (keymenu[index].name[0] == 'S'
+				  && !strcmp(keymenu[index].name, "Spc"))
+				   ? ' '
+				   : keymenu[index].name[0],
+			     nbuf, invert_label,
+			     term.t_nrow - 1 + row, (slot * tspace),
+			     strlen(nbuf));
+#endif
 
 		n = nspace[slot];
 		while(p && *p && n--){
@@ -2218,7 +2403,8 @@ KEYMENU *keymenu;
 		  *obufp++ = ' ';
 
 		p = keymenu[index].label;
-		n = tspace - nspace[slot];
+		n = ((slot == 5 && row == 1) ? adjusted_tspace
+					     : tspace) - nspace[slot];
 		while(p && *p && n-- > 0)
 		  *obufp++ = *p++;
 
@@ -2226,12 +2412,12 @@ KEYMENU *keymenu;
 		  *obufp++ = ' ';
 	    }
 	    else{
-		n = tspace;
+		n = (slot == 5 && row == 1) ? adjusted_tspace : tspace;
 		while(n--)
 		  *obufp++ = ' ';
 
-#if	defined(DOS) && defined(MOUSE)
-		register_key(index, NODATA, NULL, 0, 0, 0);
+#ifdef	MOUSE
+		register_key(index, NODATA, "", NULL, 0, 0, 0);
 #endif
 	    }
 
