@@ -1,5 +1,5 @@
 #if	!defined(lint) && !defined(DOS)
-static char rcsid[] = "$Id: file.c,v 4.45 1996/05/31 00:06:51 mikes Exp $";
+static char rcsid[] = "$Id: file.c,v 4.69 1999/01/29 19:59:05 hubert Exp $";
 #endif
 /*
  * Program:	High level file input and output routines
@@ -21,7 +21,7 @@ static char rcsid[] = "$Id: file.c,v 4.45 1996/05/31 00:06:51 mikes Exp $";
  * permission of the University of Washington.
  * 
  * Pine, Pico, and Pilot software and its included text are Copyright
- * 1989-1996 by the University of Washington.
+ * 1989-1998 by the University of Washington.
  * 
  * The full text of our legal notices is contained in the file called
  * CPYRIGHT, included with this distribution.
@@ -34,12 +34,7 @@ static char rcsid[] = "$Id: file.c,v 4.45 1996/05/31 00:06:51 mikes Exp $";
  * reading and writing of the disk are
  * in "fileio.c".
  */
-#include        <stdio.h>
-#include	"osdep.h"
-#include        "pico.h"
-#include	"estruct.h"
-#include        "edef.h"
-#include        "efunc.h"
+#include	"headers.h"
 
 
 #ifdef	ANSI
@@ -77,7 +72,7 @@ int f, n;
 	  return(0);
 	}
 
-	if((gmode&MDTREE) && !in_oper_tree(fname)){
+	if((gmode & MDTREE) && !in_oper_tree(fname)){
 	  emlwrite("Can't read file from outside of %s", opertree);
 	  return(0);
 	}
@@ -126,9 +121,9 @@ insfile(f, n)
 int f, n;
 {
     register int s;
-    char	 fname[NFILEN], dir[NFILEN];
+    char	 fname[NLINE], dir[NLINE];
     int		 retval, bye = 0, msg = 0;
-    char	 prompt[64];
+    char	 prompt[64], *infile;
     EXTRAKEYS    menu_ins[5];
     
     if (curbp->b_mode&MDVIEW) /* don't allow this command if */
@@ -173,11 +168,11 @@ int f, n;
 
 	sprintf(prompt, "%s to insert from %s %s: ",
 		msg ? "Number of message" : "File",
-		(msg || (gmode&MDCURDIR)) ? "current"
-					  : (gmode&MDTREE) ? opertree
-							   : "home",
+		(msg || (gmode&MDCURDIR))
+		  ? "current"
+		  : ((gmode & MDTREE) || opertree[0]) ? opertree : "home",
 		msg ? "folder" : "directory");
-	s = mlreplyd(prompt, fname, NFILEN, QDEFLT, msg ? NULL : menu_ins);
+	s = mlreplyd(prompt, fname, NLINE, QDEFLT, msg ? NULL : menu_ins);
 	/* something to read and it was edited or the default accepted */
         if(fname[0] && (s == TRUE || s == FALSE)){
 	    bye++;
@@ -191,15 +186,21 @@ int f, n;
 		    emlwrite("Can't insert file in restricted mode",NULL);
 		}
 		else{
-		    if(gmode&MDTREE)
-		      compresspath(opertree, fname, NFILEN);
+		    if((gmode & MDTREE)
+		       && !compresspath(opertree, fname, NLINE)){
+			emlwrite(
+			"Can't insert file from outside of %s: too many ..'s",
+			opertree);
+		    }
+		    else{
+			fixpath(fname, NLINE);
 
-		    fixpath(fname, NFILEN);
-		    if((gmode&MDTREE) && !in_oper_tree(fname))
-		      emlwrite("Can't insert file from outside of %s",
-			        opertree);
-		    else
-		      retval = ifile(fname);
+			if((gmode & MDTREE) && !in_oper_tree(fname))
+			  emlwrite("Can't insert file from outside of %s",
+				    opertree);
+			else
+			  retval = ifile(fname);
+		    }
 		}
 	    }
 	}
@@ -208,7 +209,7 @@ int f, n;
 	      case (CTRL|'I') :
 		{
 		    char *fn, *p;
-		    int   l = NFILEN;;
+		    int   l = NLINE;
 
 		    dir[0] = '\0';
 		    if(*fname && (p = strrchr(fname, C_FILESEP))){
@@ -228,16 +229,17 @@ int f, n;
 			}
 			else
 			  sprintf(dir, "%s%c%.*s", 
-				  (gmode&MDCURDIR) ? "."
-					  : (gmode&MDTREE) ? opertree
-							   : gethomedir(NULL),
+				  (gmode & MDCURDIR) ? "."
+					  : ((gmode & MDTREE) || opertree[0])
+					       ? opertree : gethomedir(NULL),
 				  C_FILESEP, p - fname, fname);
 		    }
 		    else{
 			fn = fname;
-			strcpy(dir, gmode&MDCURDIR ? "."
-					    : gmode&MDTREE ? opertree
-							   : gethomedir(NULL));
+			strcpy(dir, (gmode & MDCURDIR)
+				       ? "."
+				       : ((gmode & MDTREE) || opertree[0])
+					    ? opertree : gethomedir(NULL));
 		    }
 
 		    if(!pico_fncomplete(dir, fn, l - 1))
@@ -253,24 +255,37 @@ int f, n;
 		    emlwrite("Can't select messages yet!", NULL);
 		}
 		else{
-		    if(*fname && isdir(fname, NULL))
+		    if(*fname && isdir(fname, NULL, NULL))
 		      strcpy(dir, fname);
 		    else
-		      strcpy(dir, (gmode&MDCURDIR) ? "."
-					  : (gmode&MDTREE) ? opertree
-							   : gethomedir(NULL));
+		      strcpy(dir, (gmode&MDCURDIR)
+				     ? "."
+				     : ((gmode & MDTREE) || opertree[0])
+					  ? opertree : gethomedir(NULL));
 
 		    fname[0] = '\0';
-		    if((s = FileBrowse(dir, fname, NULL, FB_READ)) == 1){
+		    if((s = FileBrowse(dir, NLINE, fname, NLINE, 
+				       NULL, FB_READ)) == 1){
 			if(gmode&MDSCUR){
 			    emlwrite("Can't insert in restricted mode",
 				     NULL);
 			    sleep(2);
 			}
 			else{
-			    strcat(dir, S_FILESEP);
-			    strcat(dir, fname);
-			    retval = ifile(dir);
+			    if (infile = (char *)malloc((strlen(dir)+
+			                       strlen(S_FILESEP)+
+					       strlen(fname)+1)*sizeof(char))){
+			      strcpy(infile, dir);
+			      strcat(infile, S_FILESEP);
+			      strcat(infile, fname);
+			      retval = ifile(infile);
+			      free((char *) infile);
+			    }
+			    else {
+			      emlwrite("Trouble allocating space for insert!"
+				       ,NULL);
+			      sleep(3);
+			    }
 			}
 			bye++;
 		    }
@@ -289,7 +304,7 @@ int f, n;
 #if	!defined(DOS) && !defined(MAC)
 	      case (CTRL|'Y') :
 		if(Pmaster && Pmaster->upload){
-		    char tfname[NFILEN];
+		    char tfname[NLINE];
 
 		    if(gmode&MDSCUR){
 			emlwrite(
@@ -321,8 +336,16 @@ int f, n;
 #endif	/* !(DOS || MAC) */
 	      case HELPCH:
 		if(Pmaster){
-		    (*Pmaster->helper)(Pmaster->ins_help,
+		    VARS_TO_SAVE *saved_state;
+
+		    saved_state = save_pico_state();
+		    (*Pmaster->helper)(msg ? Pmaster->ins_m_help
+					   : Pmaster->ins_help,
 				       "Help for Insert File", 1);
+		    if(saved_state){
+			restore_pico_state(saved_state);
+			free_pico_state(saved_state);
+		    }
 		}
 		else
 		  pico_help(inshelptext, "Help for Insert File", 1);
@@ -374,92 +397,61 @@ readin(fname, lockfl)
 char    fname[];	/* name of file to read */
 int	lockfl;		/* check for file locks? */
 {
-        register LINE   *lp1;
-        register LINE   *lp2;
-        register int    i;
-        register WINDOW *wp;
-        register BUFFER *bp;
-        register int    s;
-        register int    nbytes;
-        register int    nline;
-	register char	*sptr;		/* pointer into filename string */
-	int		lflag;		/* any lines longer than allowed? */
-        char            line[NLINE];
-	CELL            ac;
+        char line[NLINE], *linep;
+	long nline;
+	int  s, done, newline;
 
-        bp = curbp;                             /* Cheap.               */
-	bp->b_linecnt = -1;			/* Must be recalculated */
-	ac.a = 0;
-        if ((s=bclear(bp)) != TRUE)             /* Might be old.        */
-                return (s);
-        bp->b_flag &= ~(BFTEMP|BFCHG);
-	/* removed 'C' mode detection */
-        strcpy(bp->b_fname, fname);
-	if((gmode&MDTREE) && !in_oper_tree(fname)){
-	    emlwrite("Can't read file from outside of %s", opertree);
-	    s = FIOERR;
-	    goto out;
-	}
+	curbp->b_linecnt = -1;			/* Must be recalculated */
+	if ((s = bclear(curbp)) != TRUE)	/* Might be old.        */
+	  return (s);
 
-        if ((s=ffropen(fname)) != FIOSUC){      /* Hard file open.      */
+	strcpy(curbp->b_fname, fname);
+	if ((s=ffropen(fname)) != FIOSUC){	/* Hard file open.      */
 	    if(s == FIOFNF)                     /* File not found.      */
 	      emlwrite("New file", NULL);
 	    else
 	      fioperr(s, fname);
+	}
+	else{
+	    emlwrite("Reading file", NULL);
+	    nline = 0L;
+	    done  = newline = 0;
+	    while(!done)
+	      if((s = ffgetline(line, NLINE, 1)) == FIOEOF){
+		  curbp->b_flag &= ~(BFTEMP|BFCHG);
+		  gotobob(FALSE, 1);
+		  sprintf(line,"Read %d line%s",
+			  nline, (nline > 1) ? "s" : "");
+		  emlwrite(line, NULL);
+		  break;
+	      }
+	      else{
+		  if(newline){
+		      lnewline();
+		      newline = 0;
+		  }
 
-	    goto out;
+		  switch(s){
+		    case FIOSUC :
+		      nline++;
+		      newline = 1;
+
+		    case FIOLNG :
+		      for(linep = line; *linep; linep++)
+			linsert(1, (unsigned char) *linep);
+
+		      break;
+
+		    default :
+		      done++;
+		      break;
+		  }
+	      }
+
+	    ffclose();                              /* Ignore errors.       */
 	}
 
-        emlwrite("Reading file", NULL);
-        nline = 0;
-	lflag = FALSE;
-        while ((s=ffgetline(line, NLINE)) == FIOSUC || s == FIOLNG) {
-		if (s == FIOLNG)
-			lflag = TRUE;
-                nbytes = strlen(line);
-                if ((lp1=lalloc(nbytes)) == NULL) {
-                        s = FIOERR;             /* Keep message on the  */
-                        break;                  /* display.             */
-                }
-                lp2 = lback(curbp->b_linep);
-                lp2->l_fp = lp1;
-                lp1->l_fp = curbp->b_linep;
-                lp1->l_bp = lp2;
-                curbp->b_linep->l_bp = lp1;
-                for (i=0; i<nbytes; ++i){
-		    ac.c = line[i];
-		    lputc(lp1, i, ac);
-		}
-                ++nline;
-        }
-        ffclose();                              /* Ignore errors.       */
-        if (s == FIOEOF) {                      /* Don't zap message!   */
-                sprintf(line,"Read %d line%s", nline, (nline > 1) ? "s" : "");
-                emlwrite(line, NULL);
-        }
-	if (lflag){
-		sprintf(line,"Read %d line%s, Long lines wrapped",
-			nline, (nline > 1) ? "s" : "");
-                emlwrite(line, NULL);
-        }
-out:
-        for (wp=wheadp; wp!=NULL; wp=wp->w_wndp) {
-                if (wp->w_bufp == curbp) {
-                        wp->w_linep = lforw(curbp->b_linep);
-                        wp->w_dotp  = lforw(curbp->b_linep);
-                        wp->w_doto  = 0;
-                        wp->w_imarkp = NULL;
-                        wp->w_imarko = 0;
-
-			if(Pmaster)
-			  wp->w_flag |= WFHARD;
-			else
-			  wp->w_flag |= WFMODE|WFHARD;
-                }
-        }
-        if (s == FIOERR || s == FIOFNF)		/* False if error.      */
-                return(FALSE);
-        return (TRUE);
+	return(s != FIOERR && s != FIOFNF);		/* true if success */
 }
 
 
@@ -478,8 +470,7 @@ int f, n;
         register WINDOW *wp;
         register int    s;
         char            fname[NFILEN];
-	char		shows[128], *bufp;
-	long		l;		/* length returned from fexist() */
+	char		shows[NLINE], *bufp;
 	EXTRAKEYS	menu_write[3];
 
 	if(curbp->b_fname[0] != 0)
@@ -505,17 +496,23 @@ int f, n;
 		    return(s);
 		}
 	      case TRUE:
-		if(gmode&MDTREE)
-		  compresspath(opertree, fname, NFILEN);
-
-		fixpath(fname, NFILEN);		/*  fixup ~ in file name  */
-		if((gmode&MDTREE) && !in_oper_tree(fname)){
-		    emlwrite("Can't write outside of %s", opertree);
+		if((gmode & MDTREE) && !compresspath(opertree, fname, NFILEN)){
+		    emlwrite("Can't write outside of %s: too many ..'s",
+			     opertree);
 		    sleep(2);
 		    continue;
 		}
-		else
-		  break;
+		else{
+		    fixpath(fname, NFILEN);	/*  fixup ~ in file name  */
+		    if((gmode & MDTREE) && !in_oper_tree(fname)){
+			emlwrite("Can't write outside of %s", opertree);
+			sleep(2);
+			continue;
+		    }
+		}
+
+		break;
+
 	      case (CTRL|'I'):
 		{
 		    char *fn, *p, dir[NFILEN];
@@ -539,16 +536,18 @@ int f, n;
 			}
 			else
 			  sprintf(dir, "%s%c%.*s", 
-				  (gmode&MDCURDIR) ? "."
-					  : gmode&MDTREE ? opertree
-							 : gethomedir(NULL),
+				  (gmode & MDCURDIR)
+				     ? "."
+				     : ((gmode & MDTREE) || opertree[0])
+					  ? opertree : gethomedir(NULL),
 				  C_FILESEP, p - fname, fname);
 		    }
 		    else{
 			fn = fname;
-			strcpy(dir, gmode&MDCURDIR ? "."
-					: gmode&MDTREE ? opertree
-						       : gethomedir(NULL));
+			strcpy(dir, (gmode & MDCURDIR)
+				       ? "."
+				       : ((gmode & MDTREE) || opertree[0])
+					    ? opertree : gethomedir(NULL));
 		    }
 
 		    if(!pico_fncomplete(dir, fn, l - 1))
@@ -560,7 +559,7 @@ int f, n;
 		/* If we have a file name, break up into path and file name.*/
 		*shows = 0;
 		if(*fname) {
-		    if (isdir (fname, NULL)) {
+		    if (isdir (fname, NULL, NULL)) {
 			/* fname is a directory. */
 			strcpy (shows, fname);
 			*fname = '\0';
@@ -579,14 +578,25 @@ int f, n;
 		}
 
 		/* If we did not end up with a valid directory, use home. */
-		if (!*shows || !isdir (shows, NULL))
-		  strcpy(shows, (gmode&MDTREE) ? opertree : gethomedir(NULL));
-
-		s = FileBrowse(shows, fname, NULL, FB_SAVE);
-		strcat(shows, S_FILESEP);
-		strcat(shows, fname);
-		strcpy(fname, shows);
+		if (!*shows || !isdir (shows, NULL, NULL))
+		  strcpy(shows, ((gmode & MDTREE) || opertree[0])
+				   ? opertree : gethomedir(NULL));
 
+		if (s = FileBrowse(shows, NLINE, fname, NFILEN, NULL, FB_SAVE) == 1) {
+		  if (strlen(shows)+strlen(S_FILESEP)+strlen(fname) < NLINE){
+		    strcat(shows, S_FILESEP);
+		    strcat(shows, fname);
+		    strcpy(fname, shows);
+		  }
+		  else {
+		    emlwrite("Cannot write. File name too long!!",NULL);
+		    sleep(3);
+		  }
+		}
+		else if (s == -1){
+		  emlwrite("Cannot write. File name too long!!",NULL);
+		  sleep(3);
+		}
 		refresh(FALSE, 1);
 		update();
 		if(s == 1)
@@ -607,7 +617,8 @@ int f, n;
 	    if(strcmp(fname, curbp->b_fname) == 0)
 		break;
 
-	    if((s=fexist(fname, "w", &l)) == FIOSUC){ /* exists.  overwrite? */
+	    if((s=fexist(fname, "w", (off_t *)NULL)) == FIOSUC){
+						    /*exists overwrite? */
 
 		sprintf(shows, "File \"%s\" exists, OVERWRITE", fname);
 		if((s=mlyesno(shows, FALSE)) == TRUE)
@@ -623,7 +634,7 @@ int f, n;
 	}
 	emlwrite("Writing...", NULL);
 
-        if ((s=writeout(fname)) != -1) {
+        if ((s=writeout(fname, 0)) != -1) {
 	        if(!(gmode&MDTOOL)){
 		    strcpy(curbp->b_fname, fname);
 		    curbp->b_flag &= ~BFCHG;
@@ -672,7 +683,7 @@ int f, n;
         }
 
 	emlwrite("Writing...", NULL);
-        if ((s=writeout(curbp->b_fname)) != -1) {
+        if ((s=writeout(curbp->b_fname, 0)) != -1) {
                 curbp->b_flag &= ~BFCHG;
                 wp = wheadp;                    /* Update mode lines.   */
                 while (wp != NULL) {
@@ -698,18 +709,24 @@ int f, n;
  * a macro for this. Most of the grief is error
  * checking of some sort.
  *
+ * If the argument readonly is set, the file is created with user read
+ * and write permission only if it doesn't already exist. Note that the
+ * word readonly is misleading. It is user r/w permission only.
+ *
  * CHANGES: 1 Aug 91: returns number of lines written or -1 on error, MSS
  */
-writeout(fn)
+writeout(fn, readonly)
 char    *fn;
+int     readonly;
 {
         register int    s;
+        register int    t;
         register LINE   *lp;
         register int    nline;
 	char     line[80];
 
-        if (!ffelbowroom(fn) || (s=ffwopen(fn)) != FIOSUC)
-                return (-1);			/* Open writes message. */
+        if (!((s = ffwopen(fn, readonly)) == FIOSUC && ffelbowroom()))
+	  return (-1);				/* Open writes message. */
 
         lp = lforw(curbp->b_linep);             /* First line.          */
         nline = 0;                              /* Number of lines.     */
@@ -719,13 +736,13 @@ char    *fn;
                 ++nline;
                 lp = lforw(lp);
         }
-        if (s == FIOSUC) {                      /* No write error.      */
-                s = ffclose();
-        } else                                  /* Ignore close error   */
-                ffclose();                      /* if a write error.    */
-        if (s != FIOSUC)                        /* Some sort of error.  */
-                return (-1);
-        return (nline);
+
+	t = ffclose();			/* ffclose complains if error */
+
+	if(s == FIOSUC)
+	  s = t;				/* report worst case */
+
+	return ((s == FIOSUC) ? nline : -1);
 }
 
 
@@ -733,42 +750,45 @@ char    *fn;
  * writetmp - write a temporary file for message text, mindful of 
  *	      access restrictions and included text.  If n is true, include
  *	      lines that indicated included message text, otw forget them
+ *            If dir is non-null, put the temp file in that directory.
  */
-char *writetmp(f, n)
-int f, n;
+char *writetmp(n, dir)
+int   n;
+char *dir;
 {
         static   char	fn[NFILEN];
         register int    s;
+        register int    t;
         register LINE   *lp;
         register int    nline;
 
-	tmpname(fn);
-	
-        if ((s=ffwopen(fn)) != FIOSUC)          /* Open writes message. */
-                return(NULL);
+        tmpname(dir, fn);
 
-#ifdef	MODE_READONLY
-	chmod(fn, MODE_READONLY);		/* fix access rights */
-#endif
+        if ((s=ffwopen(fn, TRUE)) != FIOSUC)	/* Open writes message. */
+          return(NULL);
 
         lp = lforw(curbp->b_linep);             /* First line.          */
         nline = 0;                              /* Number of lines.     */
         while (lp != curbp->b_linep) {
 	    if(n || (!n && lp->l_text[0].c != '>'))
-                if ((s=ffputline(&lp->l_text[0], llength(lp))) != FIOSUC)
-                        break;
-                ++nline;
-                lp = lforw(lp);
+	      if ((s=ffputline(&lp->l_text[0], llength(lp))) != FIOSUC)
+		break;
+
+	    ++nline;
+	    lp = lforw(lp);
         }
-        if (s == FIOSUC) {                      /* No write error.      */
-                s = ffclose();
-        } else                                  /* Ignore close error   */
-                ffclose();                      /* if a write error.    */
-        if (s != FIOSUC){                       /* Some sort of error.  */
-	        unlink(fn);
-                return(NULL);
+
+	t = ffclose();			/* ffclose complains if error */
+
+	if(s == FIOSUC)
+	  s = t;				/* remember worst case */
+
+	if (s != FIOSUC){                       /* Some sort of error.  */
+	    unlink(fn);
+	    return(NULL);
 	}
-        return(fn);
+
+	return(fn);
 }
 
 
@@ -780,89 +800,60 @@ int f, n;
 ifile(fname)
 char    fname[];
 {
-        register LINE   *lp0;
-        register LINE   *lp1;
-        register LINE   *lp2;
-        register int    i;
-        register BUFFER *bp;
-        register int    s;
-        register int    nbytes;
-        register int    nline;
-	int		lflag;		/* any lines longer than allowed? */
-        char            line[NLINE];
-        char     dbuf[128];
-        register char    *dbufp;
-	CELL            ac;
+        char line[NLINE], *linep;
+	long nline;
+	int  s, done, newline;
 
-        bp = curbp;                             /* Cheap.               */
-        bp->b_flag |= BFCHG;			/* we have changed	*/
-	bp->b_flag &= ~BFTEMP;			/* and are not temporary*/
-	bp->b_linecnt = -1;			/* must be recalculated */
-	ac.a = 0;
         if ((s=ffropen(fname)) != FIOSUC){      /* Hard file open.      */
-		fioperr(s,fname);
-                return(FALSE);
+	    fioperr(s, fname);
+	    return(FALSE);
 	}
 
+	gotobol(FALSE, 1);
+
+        curbp->b_flag |= BFCHG;			/* we have changed	*/
+	curbp->b_flag &= ~BFTEMP;		/* and are not temporary*/
+	curbp->b_linecnt = -1;			/* must be recalculated */
+
         emlwrite("Inserting %s.", fname);
+	done = newline = 0;
+	nline = 0L;
+	while(!done)
+	  if((s = ffgetline(line, NLINE, 1)) == FIOEOF){
+	      if(llength(curwp->w_dotp) > curwp->w_doto)
+		lnewline();
+	      else
+		forwchar(FALSE, 1);
 
-	/* back up a line and save the mark here */
-	curwp->w_dotp = lback(curwp->w_dotp);
-	curwp->w_doto = 0;
-	curwp->w_imarkp = curwp->w_dotp;
-	curwp->w_imarko = 0;
+	      sprintf(line,"Inserted %d line%s", nline, (nline>1) ? "s" : "");
+	      emlwrite(line, NULL);
+	      break;
+	  }
+	  else{
+	      if(newline){
+		  lnewline();
+		  newline = 0;
+	      }
 
-        nline = 0;
-	lflag = FALSE;
-        while ((s=ffgetline(line, NLINE)) == FIOSUC || s == FIOLNG) {
-		if (s == FIOLNG)
-			lflag = TRUE;
-                nbytes = strlen(line);
-                if ((lp1=lalloc(nbytes)) == NULL) {
-                        s = FIOERR;             /* Keep message on the  */
-                        break;                  /* display.             */
-                }
-		lp0 = curwp->w_dotp;	/* line previous to insert */
-		lp2 = lp0->l_fp;	/* line after insert */
+	      switch(s){
+		case FIOSUC :			/* copy line into buf */
+		  nline++;
+		  newline = 1;
 
-		/* re-link new line between lp0 and lp2 */
-		lp2->l_bp = lp1;
-		lp0->l_fp = lp1;
-		lp1->l_bp = lp0;
-		lp1->l_fp = lp2;
+		case FIOLNG :
+		  for(linep = line; *linep; linep++)
+		    linsert(1, (unsigned char) *linep);
+
+		  break;
+
+		default :
+		  done++;
+	      }
+	  }
 
-		/* and advance and write out the current line */
-		curwp->w_dotp = lp1;
-                for (i=0; i<nbytes; ++i){
-		    ac.c = line[i];
-		    lputc(lp1, i, ac);
-		}
-                ++nline;
-        }
         ffclose();                              /* Ignore errors.       */
-	curwp->w_imarkp = lforw(curwp->w_imarkp);
-        if (s == FIOEOF) {                      /* Don't zap message!   */
-	        sprintf(dbuf,"Inserted %d line%s",nline,(nline>1) ? "s" : "");
-		emlwrite(dbuf, NULL);
-        }
-	if (lflag) {
-		sprintf(dbuf,"Inserted %d line%s, Long lines wrapped.",
-			nline, (nline>1) ? "s" : "");
-		emlwrite(dbuf, NULL);
-        }
-out:
-	/* advance to the next line and mark the window for changes */
-	curwp->w_flag |= WFHARD;
 
-	/* copy window parameters back to the buffer structure */
-	curbp->b_dotp = curwp->w_dotp;
-	curbp->b_doto = curwp->w_doto;
-	curbp->b_markp = curwp->w_imarkp;
-	curbp->b_marko = curwp->w_imarko;
-
-        if (s == FIOERR)                        /* False if error.      */
-                return (FALSE);
-        return (TRUE);
+	return(s != FIOERR);
 }
 
 
@@ -874,7 +865,7 @@ int pico_fncomplete(dir, fn, len)
 char *dir, *fn;
 int   len;
 {
-    char *p, *dlist, tmp[NFILEN], dtmp[NFILEN];
+    char *p, *dlist, tmp[NLINE], dtmp[NLINE];
     int   n, i, match = -1;
 #ifdef	DOS
 #define	FILECMP(x, y)	(toupper((unsigned char)(x))\
@@ -884,7 +875,7 @@ int   len;
 #endif
 
     strcpy(dtmp, dir);
-    pfnexpand(dir = dtmp, NFILEN);
+    pfnexpand(dir = dtmp, NLINE);
     if(*fn && (dlist = p = getfnames(dir, fn, &n, NULL))){
 	memset(tmp, 0, sizeof(tmp));
 	while(n--){			/* any names in it */
@@ -915,10 +906,18 @@ int   len;
 	strncpy(fn, tmp, len);
 	fn[len] = '\0';
 	if(match == 1){
+	  if ((strlen(dir)+strlen(S_FILESEP)+strlen(fn)) < len){
 	    strcat(dir, S_FILESEP);
 	    strcat(dir, fn);
-	    if(isdir(dir, NULL))
+	    if(isdir(dir, NULL, NULL))
 	      strcat(fn, S_FILESEP);
+	  }
+	  else{
+	    emlwrite("File name too BIG!!",0);
+	    sleep(3);
+	    *fn = '\0';
+	  }
+
 	}
     }
 
@@ -933,5 +932,12 @@ int   len;
 in_oper_tree(f)
 char *f;
 {
-    return(!strncmp(opertree, f, strlen(opertree)));
+    int end = strlen(opertree);
+
+    return(!strncmp(opertree, f, end)
+	    && (opertree[end-1] == '/'
+		|| opertree[end-1] == '\\'
+		|| f[end] == '\0'
+		|| f[end] == '/'
+		|| f[end] == '\\'));
 }
