@@ -16,11 +16,21 @@
  *		input prompts.
  * 2.4.3	Added printing capability.
  *
+ *
+ * Pine and Pico are registered trademarks of the University of Washington.
+ * No commercial use of these trademarks may be made without prior written
+ * permission of the University of Washington.
+ * 
+ * Pine, Pico, and Pilot software and its included text are Copyright
+ * 1989-1998 by the University of Washington.
+ * 
+ * The full text of our legal notices is contained in the file called
+ * CPYRIGHT, included with this distribution.
  */
 
 
-#define VER_MAJOR 2
-#define VER_MINOR 5
+#define VER_MAJOR 3
+#define VER_MINOR 4
 extern char datestamp[];
 
 
@@ -52,3 +62,13 @@ mswin_compilation_date()
 {
     return(datestamp);
 }
+
+
+/*
+ * Return special remarks...
+ */
+char *
+mswin_compilation_remarks()
+{
+    return("");
+}
