@@ -1,5 +1,5 @@
 /*
- * $Id: pico.h,v 4.75 1996/07/09 22:49:32 mikes Exp $
+ * $Id: pico.h,v 4.106 1999/01/29 19:59:05 hubert Exp $
  *
  * Program:	pico.h - definitions for Pine's composer library
  *
@@ -20,7 +20,7 @@
  * permission of the University of Washington.
  * 
  * Pine, Pico, and Pilot software and its included text are Copyright
- * 1989-1996 by the University of Washington.
+ * 1989-1998 by the University of Washington.
  * 
  * The full text of our legal notices is contained in the file called
  * CPYRIGHT, included with this distribution.
@@ -63,6 +63,12 @@ struct hdr_line {
         struct  hdr_line        *prev;
 };
  
+#if defined(DOS) || defined(HELPFILE)
+#define	HELP_T	short
+#else
+#define	HELP_T	char **
+#endif
+
  
 /* 
  *  This structure controls the header line items on the screen.  An
@@ -74,11 +80,7 @@ struct hdr_line {
 struct headerentry {
         char     *prompt;
 	char	 *name;
-#if defined(DOS) || defined(HELPFILE)
-	short	  help;
-#else
-	char	**help;
-#endif
+	HELP_T	  help;
         int	  prlen;
         int	  maxlen;
         char    **realaddr;
@@ -87,6 +89,7 @@ struct headerentry {
 				     /* entry builder's 4th arg affects     */
         char   *(*selector)();       /* Browser for possible values         */
         char     *key_label;         /* Key label for key to call browser   */
+        char   *(*fileedit)();       /* Editor for file named in header     */
         unsigned  display_it:1;	     /* field is to be displayed by default */
         unsigned  break_on_comma:1;  /* Field breaks on commas              */
         unsigned  is_attach:1;       /* Special case field for attachments  */
@@ -96,6 +99,7 @@ struct headerentry {
         unsigned  sticky:1;          /* Can't change this via affected_entry*/
         unsigned  dirty:1;           /* We've changed this entry            */
 	unsigned  start_here:1;	     /* begin composer on first on lit      */
+	unsigned  blank:1;	     /* blank line separator                */
 #ifdef	KS_OSDATAVAR
 	KS_OSDATAVAR		     /* Port-Specific keymenu data */
 #endif
@@ -105,14 +109,27 @@ struct headerentry {
 
 
 /*
- * Structure to pass as arg to builders
+ * Structure to pass as arg to builders.
+ *
+ *    me -- A pointer to the bldr_private data for the entry currently
+ *            being edited.
+ *  tptr -- A malloc'd copy of the displayed text for the affected_entry
+ *            pointed to by the  aff argument.
+ *   aff -- A pointer to the bldr_private data for the affected_entry (the
+ *            entry that this entry affects).
+ *  next -- The next affected_entry in the list. For example, the Lcc entry
+ *            affects the To entry which affects the Fcc entry.
  */
 typedef struct bld_arg {
-	char           *tptr;	/* pointer to malloc'd text 		     */
-	void          **xtra;	/* builder keeps track of private info here  */
-	struct bld_arg *next;	/* next one in list if more than one affected*/
+    void          **me;
+    char           *tptr;
+    void          **aff;
+    struct bld_arg *next;
 } BUILDER_ARG;
 
+#define BUILDER_SCREEN_MANGLED 0x1
+#define BUILDER_MESSAGE_DISPLAYED 0x2
+
 
 /*
  * structure to keep track of header display
@@ -158,18 +175,22 @@ typedef struct pico_struct {
     void  *msgtext;			/* ptrs to malloc'd arrays of char */
     char  *pine_anchor;			/* ptr to pine anchor line */
     char  *pine_version;		/* string containing Pine's version */
-    char  *alt_ed;			/* name of alternate editor or NULL */
-    char  *alt_spell;			/* Checker to use other than "spell" */
     char  *oper_dir;			/* Operating dir (confine to tree) */
     char  *quote_str;			/* prepended to lines of quoted text */
+    char  *exit_label;			/* Label for ^X in keymenu */
+    char  *ctrlr_label;			/* Label for ^R in keymenu */
+    char  *alt_spell;			/* Checker to use other than "spell" */
+    char **alt_ed;			/* name of alternate editor or NULL */
     int    fillcolumn;			/* where to wrap */
     int    menu_rows;			/* number of rows in menu (0 or 2) */
+    long   edit_offset;			/* offset into hdr line or body */
     PATMT *attachments;			/* linked list of attachments */
     long   pine_flags;			/* entry mode flags */
-    void  (*helper)();			/* Pine's help function  */
+    int   (*helper)();			/* Pine's help function  */
     int   (*showmsg)();			/* Pine's display_message */
-    void  (*keybinit)();		/* Pine's keyboard initializer  */
-    int   (*raw_io)();			/* Pine's Raw() */
+    int   (*suspend)();			/* Pine's suspend */
+    void  (*keybinput)();		/* Pine's keyboard input indicator */
+    int   (*tty_fix)();			/* Let Pine fix tty state */
     long  (*newmail)();			/* Pine's report_new_mail */
     long  (*msgntext)();		/* callback to get msg n's text */
     int   (*upload)();			/* callback to rcv uplaoded text */
@@ -177,23 +198,59 @@ typedef struct pico_struct {
     char *(*exittest)();		/* callback to verify exit request */
     char *(*canceltest)();		/* callback to verify cancel request */
     int   (*mimetype)();		/* callback to display mime type */
-    void  (*expander)();		/* callback to expand address lists */
+    int   (*expander)();		/* callback to expand address lists */
     void  (*resize)();			/* callback handling screen resize */
-#if defined(DOS) || defined(HELPFILE)
-    short search_help;
-    short ins_help;
-    short composer_help;
-    short browse_help;
-#else
-    char  **search_help;
-    char  **ins_help;
-    char  **composer_help;
-    char  **browse_help;
-#endif
+    void  (*winch_cleanup)();		/* callback handling screen resize */
+    int    arm_winch_cleanup;		/* do the winch_cleanup if resized */
+    HELP_T search_help;
+    HELP_T ins_help;
+    HELP_T ins_m_help;
+    HELP_T composer_help;
+    HELP_T browse_help;
+    HELP_T attach_help;
     struct headerentry *headents;
 } PICO;
 
 
+/*
+ * Used to save and restore global pico variables that are destroyed by
+ * calling pico a second time. This happens when pico calls a selector
+ * in the HeaderEditor and that selector calls pico again.
+ */
+typedef struct _save_stuff {
+    int                 vtrow,
+	                vtcol,
+	                lbound;
+    VIDEO             **vscreen,
+	              **pscreen;	/* save pointers */
+    struct on_display   ods;		/* save whole struct */
+    short               delim_ps,
+	                invert_ps;
+    int                 pico_all_done;
+    jmp_buf             finstate;
+    char               *pico_anchor;	/* save pointer */
+    PICO               *Pmaster;	/* save pointer */
+    int                 fillcol;
+    char               *pat;		/* save array */
+    int                 ComposerTopLine,
+                        ComposerEditing;
+    long                gmode;
+    char               *alt_speller;	/* save pointer */
+    int                 currow,
+	                curcol,
+	                thisflag,
+	                lastflag,
+	                curgoal;
+    char               *opertree;	/* save array */
+    WINDOW             *curwp;		/* save pointer */
+    WINDOW             *wheadp;		/* save pointer */
+    BUFFER             *curbp;		/* save pointer */
+    BUFFER             *bheadp;		/* save pointer */
+    int                 km_popped;
+    int                 mrow;
+} VARS_TO_SAVE;
+
+
 #ifdef	MOUSE
 /*
  * Mouse buttons.
@@ -233,11 +290,7 @@ typedef struct mouse_struct {
 
 
 
-#ifdef	ANSI
-typedef	unsigned long (*mousehandler_t)(int, int, int, int, int);
-#else
-typedef	unsigned long (*mousehandler_t)();
-#endif
+typedef	unsigned long (*mousehandler_t) PROTO((int, int, int, int, int));
 
 typedef struct point {
     unsigned	r:8;		/* row value				*/
@@ -273,18 +326,19 @@ typedef struct  KBSTREE {
  * Protos for functions used to manage keyboard escape sequences
  * NOTE: these may ot actually get defined under some OS's (ie, DOS, WIN)
  */
-extern	int	kbseq();
-extern	void	kpinsert();
-extern	void	kbdestroy();
+extern	int	kbseq PROTO((int (*)(), int (*)(), void (*)(), int *));
+extern	void	kbdestroy PROTO((KBESC_T *));
 
 
 /*
  * various flags that they may passed to PICO
  */
-#define	P_DELRUBS	0x40000000	/* map ^H to forwdel		   */
-#define	P_LOCALLF	0x20000000	/* use local vs. NVT EOL	   */
-#define	P_BODY		0x10000000	/* start composer in body	   */
-#define	P_HEADEND	0x08000000	/* start composer at end of header */
+#define P_HICTRL	0x80000000	/* overwrite mode		*/
+#define	P_CHKPTNOW	0x40000000	/* do the checkpoint on entry      */
+#define	P_DELRUBS	0x20000000	/* map ^H to forwdel		   */
+#define	P_LOCALLF	0x10000000	/* use local vs. NVT EOL	   */
+#define	P_BODY		0x08000000	/* start composer in body	   */
+#define	P_HEADEND	0x04000000	/* start composer at end of header */
 #define	P_VIEW		MDVIEW		/* read-only			   */
 #define	P_FKEYS		MDFKEY		/* run in function key mode 	   */
 #define	P_SECURE	MDSCUR		/* run in restricted (demo) mode   */
@@ -299,8 +353,9 @@ extern	void	kbdestroy();
 #define	P_SHOCUR	MDSHOCUR	/* cursor follows hilite in browser*/
 #define	P_HIBITIGN	MDHBTIGN	/* ignore chars with hi bit set    */
 #define	P_DOTFILES	MDDOTSOK	/* browser displays dot files	   */
-#define	P_ABOOK		MDHDRONLY	/* called as address book editor   */
+#define	P_NOBODY	MDHDRONLY	/* Operate only on given headers   */
 #define	P_ALLOW_GOTO	MDGOTO		/* support "Goto" in file browser  */
+#define P_REPLACE       MDREPLACE       /* allow "Replace" in "Where is"   */
 
 /*
  * definitions for various PICO modes 
@@ -309,34 +364,35 @@ extern	void	kbdestroy();
 #define	MDSPELL		0x00000002	/* spell error parcing		*/
 #define	MDEXACT		0x00000004	/* Exact matching for searches	*/
 #define	MDVIEW		0x00000008	/* read-only buffer		*/
-#define MDOVER		0x00000010	/* overwrite mode		*/
-#define MDFKEY		0x00000020	/* function key  mode		*/
-#define MDSCUR		0x00000040	/* secure (for demo) mode	*/
-#define MDSSPD		0x00000080	/* suspendable mode		*/
-#define MDADVN		0x00000100	/* Pico's advanced mode		*/
-#define MDTOOL		0x00000200	/* "tool" mode (quick exit)	*/
-#define MDBRONLY	0x00000400	/* indicates standalone browser	*/
-#define MDCURDIR	0x00000800	/* use current dir for lookups	*/
-#define MDALTNOW	0x00001000	/* enter alt ed sans hesitation */
-#define	MDSPWN		0x00002000	/* spawn subshell for suspend	*/
-#define	MDCMPLT		0x00004000	/* enable file name completion  */
-#define	MDDTKILL	0x00008000	/* kill from dot to eol		*/
-#define	MDSHOCUR	0x00010000	/* cursor follows hilite in browser */
-#define	MDHBTIGN	0x00020000	/* ignore chars with hi bit set */
-#define	MDDOTSOK	0x00040000	/* browser displays dot files   */
-#define	MDEXTFB		0x00080000	/* stand alone File browser     */
-#define	MDTREE		0x00100000	/* confine to a subtree         */
-#define	MDMOUSE		0x00200000	/* allow mouse (part. in xterm) */
-#define	MDONECOL	0x00400000	/* single column browser        */
-#define	MDHDRONLY	0x00800000	/* header editing exclusively   */
-#define	MDGOTO		0x01000000	/* support "Goto" in file browser */
+#define MDFKEY		0x00000010	/* function key  mode		*/
+#define MDSCUR		0x00000020	/* secure (for demo) mode	*/
+#define MDSSPD		0x00000040	/* suspendable mode		*/
+#define MDADVN		0x00000080	/* Pico's advanced mode		*/
+#define MDTOOL		0x00000100	/* "tool" mode (quick exit)	*/
+#define MDBRONLY	0x00000200	/* indicates standalone browser	*/
+#define MDCURDIR	0x00000400	/* use current dir for lookups	*/
+#define MDALTNOW	0x00000800	/* enter alt ed sans hesitation */
+#define	MDSPWN		0x00001000	/* spawn subshell for suspend	*/
+#define	MDCMPLT		0x00002000	/* enable file name completion  */
+#define	MDDTKILL	0x00004000	/* kill from dot to eol		*/
+#define	MDSHOCUR	0x00008000	/* cursor follows hilite in browser */
+#define	MDHBTIGN	0x00010000	/* ignore chars with hi bit set */
+#define	MDDOTSOK	0x00020000	/* browser displays dot files   */
+#define	MDEXTFB		0x00040000	/* stand alone File browser     */
+#define	MDTREE		0x00080000	/* confine to a subtree         */
+#define	MDMOUSE		0x00100000	/* allow mouse (part. in xterm) */
+#define	MDONECOL	0x00200000	/* single column browser        */
+#define	MDHDRONLY	0x00400000	/* header editing exclusively   */
+#define	MDGOTO		0x00800000	/* support "Goto" in file browser */
+#define MDREPLACE       0x01000000      /* allow "Replace" in "Where is"*/
+#define MDTCAPWINS      0x02000000      /* Termcap overrides defaults   */
 
 /*
  * Main defs 
  */
 #ifdef	maindef
 PICO	*Pmaster = NULL;		/* composer specific stuff */
-char	*version = "2.9";		/* PICO version number */
+char	*version = "3.5";		/* PICO version number */
 
 #else
 extern	PICO *Pmaster;			/* composer specific stuff */
@@ -359,36 +415,51 @@ extern	char *version;			/* pico version! */
 #define	MESSDELAY	25
 #define	NMMESSDELAY	60
 #ifndef CHKPTDELAY
-#define	CHKPTDELAY	200
+#define	CHKPTDELAY	100
 #endif
 
 
 /*
  * defs for keypad and function keys...
  */
-#define K_PAD_UP		0x0811
-#define K_PAD_DOWN		0x0812
-#define K_PAD_RIGHT		0x0813
-#define K_PAD_LEFT		0x0814
-#define K_PAD_PREVPAGE		0x0815
-#define K_PAD_NEXTPAGE		0x0816
-#define K_PAD_HOME		0x0817
-#define K_PAD_END		0x0818
-#define K_PAD_DELETE		0x0819
-#define BADESC			0x0820
-#define K_MOUSE			0x0821
-#define K_SCROLLUPLINE		0x0822
-#define K_SCROLLDOWNLINE	0x0823
-#define K_SCROLLTO		0x0824
-#define K_XTERM_MOUSE		0x0825
-#define K_DOUBLE_ESC		0x0826
-#define K_SWALLOW_TIL_Z		0x0827
-#define K_SWALLOW_UP		0x0828
-#define K_SWALLOW_DOWN		0x0829
-#define K_SWALLOW_RIGHT		0x0830
-#define K_SWALLOW_LEFT		0x0831
-#define K_KERMIT		0x0832
-#define NODATA			0x08FF
+#define KEY_UP		0x0811
+#define KEY_DOWN	0x0812
+#define KEY_RIGHT	0x0813
+#define KEY_LEFT	0x0814
+#define KEY_PGUP	0x0815
+#define KEY_PGDN	0x0816
+#define KEY_HOME	0x0817
+#define KEY_END		0x0818
+#define KEY_DEL		0x0819
+#define BADESC		0x0820
+#define KEY_MOUSE	0x0821	/* Fake key to indicate mouse event. */
+#define KEY_SCRLUPL	0x0822
+#define KEY_SCRLDNL	0x0823
+#define KEY_SCRLTO	0x0824
+#define KEY_XTERM_MOUSE	0x0825
+#define KEY_DOUBLE_ESC	0x0826
+#define KEY_SWALLOW_Z	0x0827
+#define KEY_SWAL_UP	0x0828	/* These four have to be in the same order */
+#define KEY_SWAL_DOWN	0x0829	/* as KEY_UP, KEY_DOWN, ...                */
+#define KEY_SWAL_RIGHT	0x0830
+#define KEY_SWAL_LEFT	0x0831
+#define KEY_KERMIT	0x0832
+#define KEY_JUNK	0x0840
+#define KEY_RESIZE	0x0841  /* Fake key to cause resize. */
+#define KEY_MENU_FLAG	0x1000
+#define KEY_MASK	0x13FF
+
+/*
+ * Don't think we are using the * fact that this is zero anywhere,
+ * but just in case we'll leave it.
+ */
+#define NO_OP_COMMAND	0x0	/* no-op for short timeouts      */
+#define NO_OP_IDLE	0x0842	/* no-op for >25 second timeouts */
+#define READY_TO_READ	0x0843
+#define BAIL_OUT	0x0844
+#define PANIC_NOW	0x0845
+#define READ_INTR	0x0846
+#define NODATA		0x08FF
  
 /*
  * defines for function keys
@@ -406,66 +477,100 @@ extern	char *version;			/* pico version! */
 #define F11     0x100B                  /* Functin key eleven           */
 #define F12     0x100C                  /* Functin key twelve           */
 
+/* 1st tier pine function keys */
+#define PF1	F1
+#define PF2	F2
+#define PF3	F3
+#define PF4	F4
+#define PF5	F5
+#define PF6	F6
+#define PF7	F7
+#define PF8	F8
+#define PF9	F9
+#define PF10	F10
+#define PF11	F11
+#define PF12	F12
+
+#define PF2OPF(x)	(x + 0x10)
+#define PF2OOPF(x)	(x + 0x20)
+#define PF2OOOPF(x)	(x + 0x30)
+
+/* 2nd tier pine function keys */
+#define OPF1	PF2OPF(PF1)
+#define OPF2	PF2OPF(PF2)
+#define OPF3	PF2OPF(PF3)
+#define OPF4	PF2OPF(PF4)
+#define OPF5	PF2OPF(PF5)
+#define OPF6	PF2OPF(PF6)
+#define OPF7	PF2OPF(PF7)
+#define OPF8	PF2OPF(PF8)
+#define OPF9	PF2OPF(PF9)
+#define OPF10	PF2OPF(PF10)
+#define OPF11	PF2OPF(PF11)
+#define OPF12	PF2OPF(PF12)
+
+/* 3rd tier pine function keys */
+#define OOPF1	PF2OOPF(PF1)
+#define OOPF2	PF2OOPF(PF2)
+#define OOPF3	PF2OOPF(PF3)
+#define OOPF4	PF2OOPF(PF4)
+#define OOPF5	PF2OOPF(PF5)
+#define OOPF6	PF2OOPF(PF6)
+#define OOPF7	PF2OOPF(PF7)
+#define OOPF8	PF2OOPF(PF8)
+#define OOPF9	PF2OOPF(PF9)
+#define OOPF10	PF2OOPF(PF10)
+#define OOPF11	PF2OOPF(PF11)
+#define OOPF12	PF2OOPF(PF12)
+
+/* 4th tier pine function keys */
+#define OOOPF1	PF2OOOPF(PF1)
+#define OOOPF2	PF2OOOPF(PF2)
+#define OOOPF3	PF2OOOPF(PF3)
+#define OOOPF4	PF2OOOPF(PF4)
+#define OOOPF5	PF2OOOPF(PF5)
+#define OOOPF6	PF2OOOPF(PF6)
+#define OOOPF7	PF2OOOPF(PF7)
+#define OOOPF8	PF2OOOPF(PF8)
+#define OOOPF9	PF2OOOPF(PF9)
+#define OOOPF10	PF2OOOPF(PF10)
+#define OOOPF11	PF2OOOPF(PF11)
+#define OOOPF12	PF2OOOPF(PF12)
+
 
 /*
  * useful function definitions
  */
-#ifdef	ANSI
-int   pico(PICO *);
-int   pico_file_browse(PICO *, char *, char *, char *, int);
-void *pico_get(void);
-void  pico_give(void *);
-int   pico_readc(void *, unsigned char *);
-int   pico_writec(void *, int);
-int   pico_puts(void *, char *);
-int   pico_seek(void *, long, int);
-int   pico_replace(void *, char *);
-int   pico_fncomplete(char *, char *, int);
+int   pico PROTO((PICO *));
+int   pico_file_browse PROTO((PICO *, char *, int, char *, int, char *, int));
+void *pico_get PROTO((void));
+void  pico_give PROTO((void *));
+int   pico_readc PROTO((void *, unsigned char *));
+int   pico_writec PROTO((void *, int));
+int   pico_puts PROTO((void *, char *));
+int   pico_seek PROTO((void *, long, int));
+int   pico_replace PROTO((void *, char *));
+int   pico_fncomplete PROTO((char *, char *, int));
 #if defined(DOS) || defined(OS2)
-int   pico_nfsetcolor(char *);
-int   pico_nbsetcolor(char *);
-int   pico_rfsetcolor(char *);
-int   pico_rbsetcolor(char *);
+int   pico_nfsetcolor PROTO((char *));
+int   pico_nbsetcolor PROTO((char *));
+int   pico_rfsetcolor PROTO((char *));
+int   pico_rbsetcolor PROTO((char *));
 #endif
 #ifdef	MOUSE
-int   init_mouse();
-void  end_mouse();
-int   mouseexist();
-int   register_mfunc(mousehandler_t, int, int, int, int);
-void  clear_mfunc(mousehandler_t);
-unsigned long mouse_in_content(int, int, int, int, int);
-unsigned long mouse_in_pico(int, int, int, int, int);
-void  mouse_get_last(mousehandler_t *, MOUSEPRESS *);
-int   checkmouse(unsigned *, int, int, int);
-void  invert_label(int, MENUITEM *);
-void  mouseon();
-void  mouseoff();
+int   init_mouse PROTO((void));
+void  end_mouse PROTO((void));
+int   mouseexist PROTO((void));
+int   register_mfunc PROTO((mousehandler_t, int, int, int, int));
+void  clear_mfunc PROTO((mousehandler_t));
+unsigned long mouse_in_content PROTO((int, int, int, int, int));
+unsigned long mouse_in_pico PROTO((int, int, int, int, int));
+void  mouse_get_last PROTO((mousehandler_t *, MOUSEPRESS *));
+int   checkmouse PROTO((unsigned *, int, int, int));
+void  invert_label PROTO((int, MENUITEM *));
+void  mouseon PROTO((void));
+void  mouseoff PROTO((void));
 #endif	/* MOUSE */
 
-#else
-int   pico();
-int   pico_file_browse();
-void *pico_get();
-void  pico_give();
-int   pico_readc();
-int   pico_writec();
-int   pico_puts();
-int   pico_seek();
-int   pico_replace();
-int   pico_fncomplete();
-#ifdef	MOUSE
-int   init_mouse();
-void  end_mouse();
-int   mouseexist();
-int   register_mfunc();
-void  clear_mfunc();
-unsigned long mouse_in_content();
-void  mouse_get_last();
-int   checkmouse();
-void  invert_label();
-void  mouseon();
-void  mouseoff();
-#endif	/* MOUSE */
-#endif
 
 #endif	/* PICO_H */
