@@ -1,5 +1,5 @@
 #if	!defined(lint) && !defined(DOS)
-static char rcsid[] = "$Id: word.c,v 4.23 1996/05/15 00:38:41 mikes Exp $";
+static char rcsid[] = "$Id: word.c,v 4.30 1998/04/02 20:48:47 mikes Exp $";
 #endif
 /*
  * Program:	Word at a time routines
@@ -21,7 +21,7 @@ static char rcsid[] = "$Id: word.c,v 4.23 1996/05/15 00:38:41 mikes Exp $";
  * permission of the University of Washington.
  * 
  * Pine, Pico, and Pilot software and its included text are Copyright
- * 1989-1996 by the University of Washington.
+ * 1989-1998 by the University of Washington.
  * 
  * The full text of our legal notices is contained in the file called
  * CPYRIGHT, included with this distribution.
@@ -33,12 +33,7 @@ static char rcsid[] = "$Id: word.c,v 4.23 1996/05/15 00:38:41 mikes Exp $";
  * paragraph mode commands, they are likely to be put in this file.
  */
 
-#include        <stdio.h>
-#include	"osdep.h"
-#include        "pico.h"
-#include        "estruct.h"
-#include        <ctype.h>
-#include	"edef.h"
+#include	"headers.h"
 
 
 /* Word wrap on n-spaces. Back-over whatever precedes the point on the current
@@ -128,6 +123,7 @@ wrapword()
  * move beyond the buffers.
  */
 backword(f, n)
+    int f, n;
 {
         if (n < 0)
                 return (forwword(f, -n));
@@ -151,6 +147,7 @@ backword(f, n)
  * is done by "forwchar". Error if you try and move beyond the buffer's end.
  */
 forwword(f, n)
+    int f, n;
 {
         if (n < 0)
                 return (backword(f, -n));
@@ -330,7 +327,7 @@ delfword(f, n)
         }
         curwp->w_dotp = dotp;
         curwp->w_doto = doto;
-        return (ldelete(size, TRUE));
+        return (ldelete(size, kinsert));
 }
 
 /*
@@ -363,7 +360,7 @@ delbword(f, n)
         }
         if (forwchar(FALSE, 1) == FALSE)
                 return (FALSE);
-        return (ldelete(size, TRUE));
+        return (ldelete(size, kinsert));
 }
 #endif	/* MAYBELATER */
 
@@ -401,258 +398,156 @@ fillpara(f, n)	/* Fill the current paragraph according to the current
 int f, n;	/* deFault flag and Numeric argument */
 
 {
-    register int c;			/* current char durring scan	*/
-    register int wordlen;		/* length of current word	*/
-    register int clength;		/* position on line during fill	*/
-    register int i;			/* index during word copy	*/
-    register int newlength;		/* tentative new line length	*/
-    register int eopflag;		/* Are we at the End-Of-Paragraph? */
-    register int firstflag;		/* first word? (needs no space)	*/
-    register LINE *eopline;		/* pointer to line just past EOP */
-    register int dotflag;		/* was the last char a period?	*/
-    char wbuf[NSTRING];			/* buffer for current word	*/
-    int  first_word_of_paragraph;
-    int  end_of_input_line;
-    char *quote = NULL;
-    int   quote_len;
-    LINE *bqline, *eqline, *qline, *lp;
+    int	    i, j, c, qlen, word[NSTRING], same_word,
+	    spaces, word_len, line_len, line_last;
+    char   *qstr;
+    LINE   *eopline;
+    REGION  region;
 
     if(curbp->b_mode&MDVIEW){		/* don't allow this command if	*/
 	return(rdonly());		/* we are in read only mode	*/
     }
     else if (fillcol == 0) {		/* no fill column set */
-	mlwrite("No fill column set");
+	mlwrite("No fill column set", NULL);
 	return(FALSE);
     }
-
-    qline = curwp->w_dotp;		/* remember where we started */
+    else if(curwp->w_dotp == curbp->b_linep) /* don't wrap! */
+      return(FALSE);
 
     /* record the pointer to the line just past the EOP */
     if(gotoeop(FALSE, 1) == FALSE)
       return(FALSE);
 
-    eopline = lforw(curwp->w_dotp);
+    eopline = curwp->w_dotp;		/* first line of para */
 
     /* and back to the beginning of the paragraph */
     gotobop(FALSE, 1);
 
-    /* if we're the composer, and we were given a quoting string,
-     * and it starts the line we're on, do some special magic...
-     */
-    if(Pmaster && Pmaster->quote_str
-       && (quote_len = i = strlen(Pmaster->quote_str))
-       && i <= llength(curwp->w_dotp)){
-	while(--i >= 0 && Pmaster->quote_str[i] == lgetc(curwp->w_dotp, i).c)
-	  ;
-
-	if(i < 0){				/* bingo! */
-#ifdef	MAYBELATER
-	    switch(mlyesno("Format quoted text", TRUE)){
-	      case FALSE :
-		gotoeop(FALSE, 1);
-		forwchar(FALSE, 1);
-	      case ABORT :
-		emlwrite("Text not Justified", NULL);
-		return(TRUE);
-
-	      default :
-		emlwrite("Justifying quoted text", NULL);
-	    }
-#endif
-
-	    /*
-	     * adjust fillcol and bracket the quoted text with special
-	     * blank lines
-	     */
-	    quote = Pmaster->quote_str;
-	    fillcol -= quote_len;
-
-	    i = curwp->w_doto;
-	    gotobol(FALSE, 1);
-	    lnewline();				/* insert special blank line */
-	    backchar(FALSE, 1);			/* back onto it */
-	    bqline = curwp->w_dotp;		/* mark start of quoted text */
-	    curwp->w_dotp = eopline;
-	    gotoeol(FALSE, 1);
-	    lnewline();
-	    eqline = curwp->w_dotp;		/* mark end of quoted text */
-	    curwp->w_dotp = lforw(bqline);	/* restore dot */
-	    curwp->w_doto = i;
-	}
-    }
+    /* determine if we're justifying quoted text or not */
+    qstr = (Pmaster && Pmaster->quote_str
+	      && quote_match(Pmaster->quote_str, curwp->w_dotp))
+	     ? Pmaster->quote_str : NULL;
+    qlen = qstr ? strlen(qstr) : 0;
 
     /* let yank() know that it may be restoring a paragraph */
     thisflag |= CFFILL;
-    
-    if(Pmaster == NULL)
+
+    if(!Pmaster)
       sgarbk = TRUE;
     
     curwp->w_flag |= WFMODE;
 
-    /* save text to be justified to the cut buffer */
+    /* cut the paragraph into our fill buffer */
     fdelete();
-    for(lp = curwp->w_dotp;
-	lp != curbp->b_linep && lp != eopline;
-	lp = lforw(lp)){
-	for(i=(lp == curwp->w_dotp) ? curwp->w_doto : 0; i < llength(lp); i++)
-	  finsert(lgetc(lp, i).c);
+    curwp->w_doto = 0;
+    getregion(&region, eopline, llength(eopline));
+    if(!ldelete(region.r_size, finsert))
+      return(FALSE);
 
-	finsert('\n');
+    /* Now insert it back wrapped */
+    spaces = word_len = line_len = same_word = 0;
+
+    /* Beginning with leading quoting... */
+    if(qstr){
+	while(qstr[line_len])
+	  linsert(1, qstr[line_len++]);
+
+	line_last = ' ';			/* no word-flush space! */
     }
 
-    if(quote){
-	/* remove quoting from whole shebang */
-	curwp->w_dotp = lforw(bqline);
-	curwp->w_doto = 0;
-	do
-	  if(quote_match(quote, curwp->w_dotp))
-	    forwdel(FALSE, (quote_len > llength(curwp->w_dotp))
-			     ? llength(curwp->w_dotp) : quote_len);
-	while((curwp->w_dotp = lforw(curwp->w_dotp)) != lback(eqline));
-
-	/* got to the line we started on */
-	curwp->w_dotp = qline;
-	curwp->w_doto = 0;
-	gotoeop(FALSE, 1);
-	eopline = lforw(curwp->w_dotp);	/* eop within quoted text */
-	/* make sure new eop isn't beyond special token line */
-	for(lp = curwp->w_dotp;
-	    lp != curbp->b_linep && lp != lback(eqline);
-	    lp = lforw(lp))
-	  ;
-
-	/* no quoted text to justify, just clean up and return */
-	if(lp == curbp->b_linep){
-	    fillcol += quote_len;	/* restore fillcol */
-	    curwp->w_dotp = bqline;	/* blast leading special blank line */
-	    gotobol(FALSE, 1);
-	    forwdel(FALSE, 1);
-	    do{				/* restore quoting */
-		curwp->w_doto = 0;
-		for(i = 0; i < quote_len; i++)
-		  linsert(1, quote[i]);
-	    }
-	    while((curwp->w_dotp = lforw(curwp->w_dotp)) != lback(eqline));
-	    curwp->w_dotp = lback(eqline);
-	    gotoeol(FALSE, 1);
-	    forwdel(FALSE, 1);		/* blast trailing special blank line */
-	    gotobol(FALSE, 1);
-	    fdelete();			/* blast pre-justify text */
-	    return(TRUE);
-	}
-
-	curwp->w_dotp = lback(eopline);
-	curwp->w_doto = 0;
-	gotobop(FALSE, 1);
+    /* ...and leading white space */
+    for(i = qlen; (c = fremove(i)) == ' ' || c == TAB; i++){
+	linsert(1, line_last = c);
+	line_len += ((c == TAB) ? (~line_len & 0x07) + 1 : 1);
     }
-    
-    /* initialize various info */
-    clength = curwp->w_doto;
-    if (clength && curwp->w_dotp->l_text[0].c == TAB)
-      clength = 8;
-    wordlen = 0;
-    dotflag = FALSE;
-    
-    /* scan through lines, filling words */
-    firstflag = TRUE;
-    eopflag = FALSE;
-    first_word_of_paragraph = TRUE;
-    
-    while (!eopflag) {
-	/* get the next character in the paragraph */
-	if (curwp->w_doto == llength(curwp->w_dotp)) {
-	    c = ' ';
-	    end_of_input_line = TRUE;
-	    if (lforw(curwp->w_dotp) == eopline)
-	      eopflag = TRUE;
-	}
-	else{
-	    c = lgetc(curwp->w_dotp, curwp->w_doto).c;
-	    end_of_input_line = FALSE;
-	}
 
-	/* and then delete it */
-	ldelete(1L, FALSE);
+    /* then digest the rest... */
+    while((c = fremove(i++)) > 0){
+	switch(c){
+	  case '\n' :
+	    i += qlen;				/* skip next quote string */
 
-	/* if not a separator, just add it in */
-	if (c != ' ' && c != TAB) {
-	    /* 
-	     * don't want to limit ourselves to only '.'
-	     */
-	    dotflag = (int)strchr(".?!:;\")", c); /* dot ? */
-	    if (wordlen < NSTRING - 1)
-	      wbuf[wordlen++] = c;
-	} else if (wordlen) {
-	    /* at a word break with a word waiting */
-	    /* calculate tentitive new length with word added */
-	    newlength = clength + 1 + wordlen;
-	    if (newlength <= fillcol) {
-		/* add word to current line */
-		if (!firstflag) {
-		    linsert(1, ' ');	/* the space */
-		    ++clength;
+	  case TAB :
+	  case ' ' :
+	    spaces++;
+	    same_word = 0;
+	    break;
+
+	  default :
+	    if(spaces){				/* flush word? */
+		if(line_len
+		   && line_len + word_len + qlen + 1 > fillcol
+		   && (line_len = fpnewline(qstr)))
+		  line_last = ' ';	/* no word-flush space! */
+
+		if(word_len){			/* word to write? */
+		    if(line_len && !isspace((unsigned char) line_last)){
+			linsert(1, ' ');	/* need padding? */
+			line_len++;
+		    }
+
+		    line_len += word_len;
+		    for(j = 0; j < word_len; j++)
+		      linsert(1, line_last = word[j]);
+
+		    if(spaces > 1 && strchr(".?!:;\")", line_last)){
+			linsert(2, line_last = ' ');
+			line_len += 2;
+		    }
+
+		    word_len = 0;
 		}
-		firstflag = FALSE;
-	    } else {
-		/* start a new line */
-		if(!first_word_of_paragraph)
-		  lnewline();
-		clength = 0;
+
+		spaces = 0;
 	    }
 
-	    first_word_of_paragraph = FALSE;
+	    if(word_len + 1 >= NSTRING){
+		/* Magic!  Fake that we output a wrapped word */
+		if(line_len && !same_word++)
+		  line_len = fpnewline(qstr);
 
-	    /* and add the word in in either case */
-	    for (i=0; i<wordlen; i++) {
-		linsert(1, wbuf[i]);
-		++clength;
+		line_len += word_len;
+		for(j = 0; j < word_len; j++)
+		  linsert(1, word[j]);
+
+		word_len  = 0;
+		line_last = ' ';
 	    }
 
-	    /*  Strategy:  Handle 3 cases:
-	     *     1. if . at end of line put extra space after it
-	     *     2. if . and only 1 space, leave only one space 
-	     *     3. if . and more than 1 space, leave 2 spaces
-	     *
-	     *  So, we know the current c is a space. if then 
-	     *  is no next c or the next c is a ' ' then we 
-	     *  need to insert a space else don't do it.
-	     */
-	    if(dotflag && (end_of_input_line
-			   || (' ' == lgetc(curwp->w_dotp, curwp->w_doto).c))){
-		linsert(1, ' ');
-		++clength;
-	    }
-
-	    wordlen = 0;
+	    word[word_len++] = c;
+	    break;
 	}
     }
-    
-    /* and add a last newline for the end of our new paragraph */
-    lnewline();
-    
-    if(quote){				/* replace quoting strings */
-	fillcol += quote_len;
-	/* remember the current dot. note: ldelete at eob special, handle it */
-	eopline = (lforw(curwp->w_dotp) == curbp->b_linep)
-		     ? curbp->b_linep : curwp->w_dotp;
-	curwp->w_dotp = bqline;
-	forwdel(FALSE, 1);		/* blast special blank line */
-	do{
-	    curwp->w_doto = 0;		/* fill quote strings back in */
-	    for(i = 0; i < quote_len; i++)
-	      linsert(1, quote[i]);
-	}
-	while((curwp->w_dotp = lforw(curwp->w_dotp)) != lback(eqline));
 
-	curwp->w_dotp = lback(eqline);	/* blast ending special blank line */
-	gotobol(FALSE, 1);
-	forwdel(FALSE, 1);
+    if(word_len){
+	if(line_len + word_len + qlen + 1 > fillcol)
+	  (void) fpnewline(qstr);
+	else if(line_len && !isspace((unsigned char) line_last))
+	  linsert(1, ' ');
 
-	/*
-	 * Restore cursor to eop w/in quote.  Note special handling 
-	 * when end of quoted text and end of buffer are the same...
-	 */
-	curwp->w_dotp = eopline;
-	curwp->w_doto = (quote_len > llength(curwp->w_dotp)) ? 0 : quote_len;
+	for(j = 0; j < word_len; j++)
+	  linsert(1, word[j]);
     }
+
+    /* Leave cursor on first char of first line after paragraph */
+    curwp->w_dotp = lforw(curwp->w_dotp);
+    curwp->w_doto = 0;
+
+    return(TRUE);
+}
+
+
+/*
+ * fpnewline - output a fill paragraph newline mindful of quote string
+ */
+fpnewline(quote)
+char *quote;
+{
+    int len;
+
+    lnewline();
+    for(len = 0; quote && *quote; quote++, len++)
+      linsert(1, *quote);
+
+    return(len);
 }
