@@ -1,5 +1,5 @@
 /*----------------------------------------------------------------------
-  $Id: headers.h,v 4.20 1996/03/15 07:13:42 hubert Exp $
+  $Id: headers.h,v 4.33 1998/11/16 20:14:44 hubert Exp $
 
             T H E    P I N E    M A I L   S Y S T E M
 
@@ -20,7 +20,7 @@
    permission of the University of Washington.
 
    Pine, Pico, and Pilot software and its included text are Copyright
-   1989-1996 by the University of Washington.
+   1989-1998 by the University of Washington.
 
    The full text of our legal notices is contained in the file called
    CPYRIGHT, included with this distribution.
@@ -56,36 +56,68 @@
  System specific includes and defines are in os.h, the source for which
 is os-xxx.h. (Don't edit osdep.h; edit os-xxx.h instead.)
  ----*/
-#include <stdio.h>
+#include "../pico/headers.h"
+
+
+#include "../c-client/mail.h"
 
 #include "os.h"
 
-/* These includes provide c-client prototypes and such */
-#include "../c-client/mail.h"
-#include "../c-client/osdep.h"
 #include "../c-client/rfc822.h"
-#include "../c-client/imap2.h"
 #include "../c-client/misc.h"
 
-/* These includes are all ANSI, and OK with all other compilers (so far) */
-#include <ctype.h>
-#include <errno.h>
-#include <setjmp.h>
+#ifdef  ENABLE_LDAP
 
-#include "../pico/pico.h"
-
-#ifdef	ANSI
-#define	PROTO(args)	args
+#include <lber.h>
+#include <ldap.h>
+/*
+ * If LDAP_OPT_ON is defined in ldap.h, then this is the newer version of the
+ * API (as opposed to the RFC1823 version). The newer version at the time
+ * this was written was only described in draft-ietf-asid-ldap-c-api-00.txt.
+ * Hmm. Seems mozilla.org defined LDAP_OPT_ON even in 1.1, so we'll require
+ * that and LDAP_VERSION3.
+ */
+#ifndef LDAPAPI
+#if	defined(LDAP_VERSION3)
+#define	LDAPAPI	20
 #else
-#define	PROTO(args)	()
+#if	defined(LDAP_OPT_ON)
+#define	LDAPAPI 11
+#else
+#define	LDAPAPI 10
 #endif
+#endif
+#ifndef LDAP_OPT_ON
+#define LDAP_OPT_ON ((void *)1)
+#endif
+#ifndef LDAP_OPT_OFF
+#define LDAP_OPT_OFF ((void *)0)
+#endif
+#ifndef LDAP_OPT_SIZELIMIT
+#define LDAP_OPT_SIZELIMIT 1134  /* we're hacking now! */
+#endif
+#ifndef LDAP_OPT_TIMELIMIT
+#define LDAP_OPT_TIMELIMIT 1135
+#endif
+#ifndef LDAP_OPT_PROTOCOL_VERSION
+#define LDAP_OPT_PROTOCOL_VERSION 1136
+#endif
+#endif
+
+#ifndef LDAP_MSG_ONE
+#define LDAP_MSG_ONE (0x00)
+#define LDAP_MSG_ALL (0x01)
+#define LDAP_MSG_RECEIVED (0x02)
+#endif
+
+#endif  /* ENABLE_LDAP */
+
+#include "helptext.h"
 
 #include "pine.h"
 
 #include "context.h"
 
-#include "helptext.h"
-
 
 
 /*----------------------------------------------------------------------
@@ -96,8 +128,6 @@ extern struct pine *ps_global;
 
 extern char	   *pine_version;	/* pointer to version string	     */
 
-extern int          timeout;		/* referenced in pico, set in pine.c */
-
 extern char         tmp_20k_buf[];
 
 #ifdef DEBUG
@@ -105,4 +135,4 @@ extern FILE        *debugfile;		/* file for debug output	  */
 extern int          debug;		/* debugging level or none (zero) */
 #endif
 
-#endif /* _HEADERS_INCLUDEDS */
+#endif /* _HEADERS_INCLUDED */
