@@ -1,5 +1,5 @@
 #if !defined(lint) && !defined(DOS)
-static char rcsid[] = "$Id: imap.c,v 4.79 1996/03/15 07:13:42 hubert Exp $";
+static char rcsid[] = "$Id: imap.c,v 4.133 1998/12/29 20:08:48 hubert Exp $";
 #endif
 /*----------------------------------------------------------------------
 
@@ -22,7 +22,7 @@ static char rcsid[] = "$Id: imap.c,v 4.79 1996/03/15 07:13:42 hubert Exp $";
    permission of the University of Washington.
 
    Pine, Pico, and Pilot software and its included text are Copyright
-   1989-1996 by the University of Washington.
+   1989-1998 by the University of Washington.
 
    The full text of our legal notices is contained in the file called
    CPYRIGHT, included with this distribution.
@@ -60,27 +60,38 @@ static char rcsid[] = "$Id: imap.c,v 4.79 1996/03/15 07:13:42 hubert Exp $";
  * we've already tried talking to this host before.
  * 
  */
-typedef struct mmlogin_s {
+typedef struct _mmlogin_s {
     char	   *user,
-		   *host,
 		   *passwd;
-    struct mmlogin_s *next;
+    STRLIST_S	   *hosts;
+    struct _mmlogin_s *next;
 } MMLOGIN_S;
 
 
+typedef	struct _se_app_s {
+    char *folder;
+    long  flags;
+} SE_APP_S;
+
+
 /*
  * Internal prototypes
  */
-int   imap_get_passwd PROTO((MMLOGIN_S *, char *, char *, char *));
-void  imap_set_passwd PROTO((MMLOGIN_S **, char *, char *, char *));
-#if defined(DOS) || defined(OS2)
+long  imap_seq_exec PROTO((MAILSTREAM *, char *,
+			   long (*) PROTO((MAILSTREAM *, long, void *)),
+			   void *));
+long  imap_seq_exec_append PROTO((MAILSTREAM *, long, void *));
+int   imap_get_passwd PROTO((MMLOGIN_S *, char *, char *, STRLIST_S *));
+void  imap_set_passwd PROTO((MMLOGIN_S **, char *, char *, STRLIST_S *));
+int   imap_same_host PROTO((STRLIST_S *, STRLIST_S *));
+#ifdef	PASSFILE
 char  xlate_in PROTO((int));
 char  xlate_out PROTO((char));
 char *passfile_name PROTO((char *, char *));
 int   read_passfile PROTO((char *, MMLOGIN_S **));
 void  write_passfile PROTO((char *, MMLOGIN_S *));
-int   get_passfile_passwd PROTO((char *, char *, char *, char *));
-void  set_passfile_passwd PROTO((char *, char *, char *, char *));
+int   get_passfile_passwd PROTO((char *, char *, char *, STRLIST_S *));
+void  set_passfile_passwd PROTO((char *, char *, char *, STRLIST_S *));
 #endif
 
 
@@ -90,20 +101,10 @@ void  set_passfile_passwd PROTO((char *, char *, char *, char *));
  * and to allow mm_searched to return number of matches.
  */
 MAILSTREAM *mm_search_stream;
-long	    mm_search_count	= 0L;
-
-
-/*
- * Exported globals used to report folder FIND/LIST responses.
- */
-void       *find_folder_list   = NULL;
-MAILSTREAM *find_folder_stream = NULL;
-long	    find_folder_count  = 0L;
-int	    find_folder_inbox    = 0;
-#ifdef NEWBB
-void	   *newbb_folder_list;
-#endif
+long	    mm_search_count  = 0L;
+MAILSTATUS  mm_status_result;
 
+MM_LIST_S  *mm_list_info;
 
 /*
  * Local global to hook cached list of host/user/passwd triples to.
@@ -123,15 +124,11 @@ void
 mm_dlog(string)
     char *string;
 {
+#ifdef	_WINDOWS
+    mswin_imaptelemetry(string);
+#endif
 #ifdef	DEBUG
-    time_t      now;
-    struct tm  *tm_now;
-
-    now = time((time_t *)0);
-    tm_now = localtime(&now);
-    dprint(0, (debugfile, "IMAP DEBUG %2.2d:%2.2d:%2.2d %d/%d: %s\n",
-	       tm_now->tm_hour, tm_now->tm_min, tm_now->tm_sec,
-	       tm_now->tm_mon+1, tm_now->tm_mday, string));
+    dprint(0, (debugfile, "IMAP DEBUG %s: %s\n", debug_time(1), string));
 #endif
 }
 
@@ -161,7 +158,7 @@ mm_log(string, errflg)
     now = time((time_t *)0);
     tm_now = localtime(&now);
 
-    dprint((errflg == ERROR ? 1 : 2),
+    dprint(ps_global->debug_imap ? 0 : (errflg == ERROR ? 1 : 2),
 	   (debugfile,
 	    "IMAP %2.2d:%2.2d:%2.2d %d/%d mm_log %s: %s\n",
 	    tm_now->tm_hour, tm_now->tm_min, tm_now->tm_sec, tm_now->tm_mon+1,
@@ -193,8 +190,8 @@ mm_log(string, errflg)
     message[sizeof(message) - 1] = '\0';
     occurance = srchstr(message, "mailbox");
     while(occurance) {
-	if(!*(occurance+7) || isspace(*(occurance+8))){
-	    was_capitalized = isupper(*occurance);
+	if(!*(occurance+7) || isspace((unsigned char)*(occurance+8))){
+	    was_capitalized = isupper((unsigned char)*occurance);
 	    rplstr(occurance, 7, (errflg == PARSE ? "address" : "folder"));
 	    if(was_capitalized)
 	      *occurance = (errflg == PARSE ? 'A' : 'F');
@@ -205,19 +202,20 @@ mm_log(string, errflg)
         occurance = srchstr(occurance, "mailbox");
     }
 
-    if(errflg == PARSE || ps_global->noshow_error) 
-      strcpy(ps_global->c_client_error, message);
-
-    if(ps_global->noshow_error ||
-		       (ps_global->noshow_warn && errflg == WARN) ||
-                       (errflg != ERROR && errflg != WARN))
-      return; /* Only care about errors; don't print when asked not to */
-
     if(errflg == ERROR)
       ps_global->mm_log_error = 1;
 
+    if(errflg == PARSE || (errflg == ERROR && ps_global->noshow_error))
+      strcpy(ps_global->c_client_error, message);
+
+    if(ps_global->noshow_error
+       || (ps_global->noshow_warn && errflg == WARN)
+       || !(errflg == ERROR || errflg == WARN))
+      return; /* Only care about errors; don't print when asked not to */
+
     /*---- Display the message ------*/
-    q_status_message(SM_ORDER | SM_DING, 3, 5, message);
+    q_status_message((errflg == ERROR) ? (SM_ORDER | SM_DING) : SM_ORDER,
+		     3, 5, message);
     strcpy(ps_global->last_error, message);
 }
 
@@ -242,19 +240,27 @@ mm_notify(stream, string, errflag)
     long        errflag;
 {
     /* be sure to log the message... */
-    dprint(1, (debugfile, "IMAP mm_notify %s : %s : %s\n",
+#ifdef DEBUG
+    if(ps_global->debug_imap)
+      dprint(0,
+	     (debugfile, "IMAP mm_notify %s : %s (%s) : %s\n",
                (!errflag) ? "NIL" : 
 		 (errflag == ERROR) ? "error" :
 		   (errflag == WARN) ? "warning" :
 		     (errflag == BYE) ? "bye" : "unknown",
 	       (stream && stream->mailbox) ? stream->mailbox : "-no folder-",
+	       (stream && stream == ps_global->inbox_stream) ? "inboxstream" :
+		 (stream && stream == ps_global->mail_stream) ? "mailstream" :
+		   (stream) ? "abookstream?" : "nostream",
 	       string));
+#endif
 
     sprintf(ps_global->last_error, "%.50s : %.*s",
 	    (stream && stream->mailbox) ? stream->mailbox : "-no folder-",
 	    min(MAX_SCREEN_COLS, 500-70),
 	    string);
-    ps_global->last_error[ps_global->ttyo->screen_cols] = '\0';
+    ps_global->last_error[ps_global->ttyo ? ps_global->ttyo->screen_cols : 499]
+									= '\0';
 
     /*
      * Then either set special bits in the pine struct or
@@ -278,11 +284,21 @@ mm_notify(stream, string, errflag)
     else if(!strncmp(string, "[TRYCREATE]", 11))
       ps_global->try_to_create = 1;
     else if(!strncmp(string, "[ALERT]", 7))
-      q_status_message2(SM_MODAL, 3, 3, "%s : %s",
+      q_status_message2(SM_MODAL, 3, 3, "Alert received while accessing \"%s\":  %s",
 			(stream && stream->mailbox)
 			  ? stream->mailbox : "-no folder-",
-			string + 7);
-    else if(!strncmp(string, "[READ-ONLY]", 11))
+			rfc1522_decode((unsigned char *)tmp_20k_buf, string, NULL));
+    else if(!strncmp(string, "[UNSEEN ", 8)){
+	char *p;
+	long  n = 0;
+
+	for(p = string + 8; isdigit(*p); p++)
+	  n = (n * 10) + (*p - '0');
+
+	ps_global->first_unseen = n;
+    }
+    else if(!strncmp(string, "[READ-ONLY]", 11)
+	    && !(stream && stream->mailbox && IS_NEWS(stream)))
       q_status_message2(SM_ORDER | SM_DING, 3, 3, "%s : %s",
 			(stream && stream->mailbox)
 			  ? stream->mailbox : "-no folder-",
@@ -314,35 +330,35 @@ mm_notify(stream, string, errflag)
  ----*/
 void
 mm_exists(stream, number)
-    MAILSTREAM *stream;
-    long number;
+    MAILSTREAM    *stream;
+    unsigned long  number;
 {
     long new_this_call;
 
-    dprint(3, (debugfile, "=== mm_exists(%ld,%s) called ===\n", number,
+#ifdef DEBUG
+    if(ps_global->debug_imap > 1)
+      dprint(0,
+	   (debugfile, "=== mm_exists(%lu,%s) called ===\n", number,
      !stream ? "(no stream)" : !stream->mailbox ? "(null)" : stream->mailbox));
+#endif
 
     if(stream == ps_global->mail_stream){
-	if(mn_get_total(ps_global->msgmap) != number){
+	if(mn_get_total(ps_global->msgmap) != (long) number){
 	    ps_global->mail_box_changed = 1;
-	    ps_global->mangled_header   = 1;
+	    ps_global->mangled_header	= 1;
 	}
 
-        if(mn_get_total(ps_global->msgmap) < number){
-	    new_this_call = number - mn_get_total(ps_global->msgmap);
+        if(mn_get_total(ps_global->msgmap) < (long) number){
+	    new_this_call = (long) number - mn_get_total(ps_global->msgmap);
 	    ps_global->new_mail_count += new_this_call;
-	    if(mn_get_sort(ps_global->msgmap) != SortArrival 
-	       || mn_get_revsort(ps_global->msgmap))
-	      clear_index_cache();
-
 	    mn_add_raw(ps_global->msgmap, new_this_call);
 	}
     } else if(stream == ps_global->inbox_stream) {
-	if(mn_get_total(ps_global->inbox_msgmap) != number)
+	if(mn_get_total(ps_global->inbox_msgmap) != (long) number)
 	  ps_global->inbox_changed = 1;
 
-        if(mn_get_total(ps_global->inbox_msgmap) < number){
-	    new_this_call = number - mn_get_total(ps_global->inbox_msgmap);
+        if(mn_get_total(ps_global->inbox_msgmap) < (long) number){
+	    new_this_call = (long) number - mn_get_total(ps_global->inbox_msgmap);
 	    ps_global->inbox_new_mail_count += new_this_call;
 	    mn_add_raw(ps_global->inbox_msgmap, new_this_call);
 	}
@@ -365,34 +381,88 @@ necessary.
   ----*/
 void
 mm_expunged(stream, number)
-    MAILSTREAM *stream;
-    long        number;
+    MAILSTREAM    *stream;
+    unsigned long  number;
 {
-    MSGNO_S *msgs;
+    MESSAGECACHE *mc;
 
-    dprint(3, (debugfile, "mm_expunged called %s %ld\n",
-     !stream  ? "(no stream)" : !stream->mailbox ? "(null)" : stream->mailbox, 
-	       number));
+#ifdef DEBUG
+    if(ps_global->debug_imap > 1)
+      dprint(0,
+	   (debugfile, "mm_expunged(%s,%lu)\n",
+	       stream
+		? (stream->mailbox
+		    ? stream->mailbox
+		    : "(no stream)")
+		: "(null)", number));
+#endif
 
     /*
      * If we ever deal with more than two streams, this'll break
      */
     if(stream == ps_global->mail_stream){
-	long i = mn_raw2m(ps_global->msgmap, number);
-	if(i){
-	    while(i <= mn_get_total(ps_global->msgmap))	/* flush invalid */
-	      clear_index_cache_ent(i++);		/* cache entries! */
+	long i = mn_raw2m(ps_global->msgmap, (long) number);
 
-	    mn_flush_raw(ps_global->msgmap, number);
+	if(i){
+	    /*
+	     * Keep on top of our special flag counts.
+	     * 
+	     * NOTE: This is allowed since mail_expunged releases
+	     * data for this message after the callback.
+	     */
+	    if(mc = mail_elt(stream, number)){
+		if(mc->spare)
+		  ps_global->msgmap->flagged_hid--;
+
+		if(mc->spare2)
+		  ps_global->msgmap->flagged_exld--;
+
+		if(mc->spare3)
+		  ps_global->msgmap->flagged_tmp--;
+
+		if(mc->sparep)
+		  msgno_free_exceptions((PARTEX_S **) &mc->sparep);
+	    }
+
+	    /* flush invalid cache entries */
+	    while(i <= mn_get_total(ps_global->msgmap))
+	      clear_index_cache_ent(i++);
+
+	    /* expunged something we're viewing? */
+	    if(!ps_global->expunge_in_progress
+	       && (mn_is_cur(ps_global->msgmap,
+			     mn_raw2m(ps_global->msgmap, (long) number))
+		   && (ps_global->prev_screen == mail_view_screen
+		       || ps_global->prev_screen == attachment_screen))){
+		ps_global->next_screen = mail_index_screen;
+		q_status_message(SM_ORDER | SM_DING , 3, 3,
+				 "Message you were viewing is gone!");
+	    }
+
+	    mn_flush_raw(ps_global->msgmap, (long) number);
 	    ps_global->mail_box_changed = 1;
-	    ps_global->mangled_header   = 1;
+	    ps_global->mangled_header = 1;
 	    ps_global->expunge_count++;
 	}
     }
     else if(stream == ps_global->inbox_stream){
-	long i = mn_raw2m(ps_global->msgmap, number);
+	long i = mn_raw2m(ps_global->inbox_msgmap, (long) number);
 	if(i){
-	    mn_flush_raw(ps_global->inbox_msgmap, number);
+	    if(mc = mail_elt(stream, number)){
+		if(mc->spare)
+		  ps_global->inbox_msgmap->flagged_hid--;
+
+		if(mc->spare2)
+		  ps_global->inbox_msgmap->flagged_exld--;
+
+		if(mc->spare3)
+		  ps_global->inbox_msgmap->flagged_tmp--;
+
+		if(mc->sparep)
+		  msgno_free_exceptions((PARTEX_S **) &mc->sparep);
+	    }
+
+	    mn_flush_raw(ps_global->inbox_msgmap, (long) number);
 	    ps_global->inbox_changed = 1;
 	    ps_global->inbox_expunge_count++;
 	}
@@ -410,9 +480,10 @@ mm_expunged(stream, number)
   ----*/
 void
 mm_searched(stream, number)
-    MAILSTREAM *stream;
-    long        number;
+    MAILSTREAM    *stream;
+    unsigned long  number;
 {
+    mail_elt(stream, number)->searched = 1;
     if(stream == mm_search_stream)
       mm_search_count++;
 }
@@ -422,7 +493,7 @@ mm_searched(stream, number)
 /*----------------------------------------------------------------------
       Get login and password from user for IMAP login
   
-  Args:  host   -- The host name the user is trying to log in to 
+  Args:  mb -- The mail box property struct
          user   -- Buffer to return the user name in 
          passwd -- Buffer to return the passwd in
          trial  -- The trial number or number of attempts to login
@@ -430,63 +501,101 @@ mm_searched(stream, number)
  Result: username and password passed back to imap
   ----*/
 void
-mm_login(host, user, passwd, trial)
-    char *host;
-    char *user;
-    char *passwd;
-    long  trial;
+mm_login(mb, user, pwd, trial)
+    NETMBX *mb;
+    char   *user;
+    char   *pwd;
+    long    trial;
 {
     static char junk[] = {'?', '\0'};
-    char      prompt[80], *p, *named_user = NULL;
+    char      prompt[MAX_SCREEN_COLS], *last, *host;
+    STRLIST_S hostlist[2];
     HelpType  help ;
-    int       rc, q_line;
+    int       i, j, goal, ugoal, rc, q_line, flags;
 
     q_line = -FOOTER_ROWS(ps_global);			/* 3 from bottom */
 
     if(ps_global->anonymous) {
         /*------ Anonymous login mode --------*/
-        if(trial >= 1) {
-            user[0]   = '\0';
-            passwd[0] = '\0';
-        } else {
+        if(trial < 1) {
             strcpy(user, "anonymous");
-            sprintf(passwd, "%s@%s", ps_global->VAR_USER_ID,
+            sprintf(pwd, "%s@%s", ps_global->VAR_USER_ID,
 		    ps_global->hostname);
-        }
+	}
+	else
+	  user[0] = pwd[0] = '\0';
+
         return;
     }
 
+    /*
+     * set up host list for sybil servers...
+     */
+    hostlist[0].name = mb->host;
+    if(mb->orighost[0] && strucmp(mb->host, mb->orighost)){
+	hostlist[0].next = &hostlist[1];
+	hostlist[1].name = mb->orighost;
+	hostlist[1].next = NULL;
+    }
+    else
+      hostlist[0].next = NULL;
+
     /*
      * Initialize user name with either /user= value in the stream
      * being logged into (accessed via the GET_USERNAMEBUF c-client
      * parameter) or the user name we're running under...
      */
-    named_user = (char *) mail_parameters(NULL, GET_USERNAMEBUF, NULL);
-    if(trial == 0L)
-      strcpy(user, (named_user && *named_user)
-		     ? named_user : ps_global->VAR_USER_ID);
+    if(trial == 0L){
+	strcpy(user, (*mb->user) ? mb->user : ps_global->VAR_USER_ID);
 
-    /*
-     * try last working password associated with this host.
-     */
-    if(trial == 0L && imap_get_passwd(mm_login_list, passwd, user, host))
-      return;
+	/* try last working password associated with this host. */
+	if(imap_get_passwd(mm_login_list, pwd, user, hostlist))
+	  return;
 
-#if defined(DOS) || defined(OS2)
-    /* check to see if there's a password left over from last session */
-    if(trial == 0L && get_passfile_passwd(ps_global->pinerc,passwd,user,host)){
-	imap_set_passwd(&mm_login_list, passwd, user, host);
-	return;
-    }
+#ifdef	PASSFILE
+	/* check to see if there's a password left over from last session */
+	if(get_passfile_passwd(ps_global->pinerc, pwd, user, &hostlist[0])){
+	    imap_set_passwd(&mm_login_list, pwd, user, &hostlist[0]);
+	    return;
+	}
 #endif
+    }
+
 
     ps_global->mangled_footer = 1;
-    if(!(named_user && *named_user)){
+    if(!*mb->user){
 	help = NO_HELP;
-	sprintf(prompt, "HOST: %s  ENTER LOGIN NAME: ", host);
+
+	/* Dress up long hostnames */
+	strcpy(prompt, "HOST: ");
+	/* leave space for "HOST", "ENTER NAME", and 15 chars for input name */
+	goal = ps_global->ttyo->screen_cols - (6 + 20
+				    + min(15, ps_global->ttyo->screen_cols/5));
+	last = "  ENTER LOGIN NAME: ";
+	if(goal < 9){
+	    last = " LOGIN: ";
+	    if((goal += 13) < 9){
+		last += 1;
+		goal = 0;
+	    }
+	}
+
+	if(goal){
+	    for(i = 6, j = 0; prompt[i] = mb->host[j]; i++, j++)
+	      if(i == goal && mb->host[goal+1]){
+		  strcpy(&prompt[i-3], "...");
+		  break;
+	      }
+	}
+	else
+	  i = 0;
+
+	strcpy(&prompt[i], last);
+
 	while(1) {
-	    rc = optionally_enter(user, q_line, 0, MAILTMPLEN - 1, 1, 0,
-				  prompt, NULL, help, 0);
+	    flags = OE_APPEND_CURRENT;
+	    rc = optionally_enter(user, q_line, 0, MAILTMPLEN - 1,
+				  prompt, NULL, help, &flags);
 	    if(rc == 3) {
 		help = help == NO_HELP ? h_oe_login : NO_HELP;
 		continue;
@@ -497,17 +606,65 @@ mm_login(host, user, passwd, trial)
 
 	if(rc == 1 || !user[0]) {
 	    user[0]   = '\0';
-	    passwd[0] = '\0';
+	    pwd[0] = '\0';
 	    return;
 	}
     }
+    else
+      strcpy(user, mb->user);
 
     help = NO_HELP;
-    sprintf(prompt, "HOST: %s  USER: %s  ENTER PASSWORD: ", host, user);
-    *passwd = '\0';
+
+    /* Dress up long host/user names */
+    /* leave space for "HOST", "USER" "ENTER PWD", 12 for user 6 for pwd */
+    goal  = strlen(mb->host);
+    ugoal = strlen(user);
+    if((i = ps_global->ttyo->screen_cols - (6 + 8 + 18 + 6)) < 14){
+	goal = 0;		/* no host! */
+	if((i = ps_global->ttyo->screen_cols - (6 + 18 + 6)) <= 6){
+	    ugoal = 0;		/* no user! */
+	    if((i = ps_global->ttyo->screen_cols - (18 + 6)) <= 0)
+	      i = 0;
+	}
+	else{
+	    ugoal = i;		/* whatever's left */
+	    i     = 0;
+	}
+    }
+    else
+      while(goal + ugoal > i)
+	if(goal > ugoal)
+	  goal--;
+	else
+	  ugoal--;
+
+    if(goal){
+	strcpy(prompt, "HOST: ");
+	for(i = 6, j = 0; prompt[i] = mb->host[j]; i++, j++)
+	  if(i == goal && mb->host[goal+1]){
+	      strcpy(&prompt[i-3], "...");
+	      break;
+	  }
+    }
+    else
+      i = 0;
+
+    if(ugoal){
+	strcpy(&prompt[i], &"  USER: "[i ? 0 : 2]);
+	for(i += strlen(&prompt[i]), j = 0; prompt[i] = user[j]; i++, j++)
+	  if(j == ugoal && user[ugoal+1]){
+	      strcpy(&prompt[i-3], "...");
+	      break;
+	  }
+    }
+
+    strcpy(&prompt[i], &"  ENTER PASSWORD: "[i ? 0 : 8]);
+
+    *pwd = '\0';
     while(1) {
-        rc = optionally_enter(passwd, q_line, 0, MAILTMPLEN - 1, 0,
-			      1, prompt, NULL, help, 0);
+	flags = OE_PASSWD;
+        rc = optionally_enter(pwd, q_line, 0, MAILTMPLEN - 1,
+			      prompt, NULL, help, &flags);
         if(rc == 3) {
             help = help == NO_HELP ? h_oe_passwd : NO_HELP;
             continue;
@@ -516,24 +673,22 @@ mm_login(host, user, passwd, trial)
           break;
     }
 
-    if(rc == 1 || !passwd[0]) {
-	if(!named_user){
-	    strcpy(user,junk);
-	    strcpy(passwd, junk);
-	}
-	else{
-	    user[0]   = '\0';
-	    passwd[0] = '\0';
+    if(rc == 1 || !pwd[0]) {
+	if(!*mb->user){
+	    strcpy(user, junk);
+	    strcpy(pwd, junk);
 	}
+	else
+	  user[0] = pwd[0] = '\0';
 
         return;
     }
 
     /* remember the password for next time */
-    imap_set_passwd(&mm_login_list, passwd, user, host);
-#if defined(DOS) || defined(OS2)
+    imap_set_passwd(&mm_login_list, pwd, user, hostlist);
+#ifdef	PASSFILE
     /* if requested, remember it on disk for next session */
-    set_passfile_passwd(ps_global->pinerc, passwd, user, host);
+    set_passfile_passwd(ps_global->pinerc, pwd, user, hostlist);
 #endif
 }
 
@@ -622,7 +777,7 @@ mm_diskerror (stream, errcode, serious)
 	removing_leading_white_space(ps_global->last_error);
 	for(i = 4, p = ps_global->last_error; *p && i < DE_LINE; ){
 	    for(s = NULL, q = p; *q && q - p < DE_COLS - 16; q++)
-	      if(isspace(*q))
+	      if(isspace((unsigned char)*q))
 		s = q;
 
 	    if(*q && s)
@@ -634,7 +789,7 @@ mm_diskerror (stream, errcode, serious)
 	    if(*(p = q)){
 		ClearLine(++i);
 		PutLine0(i, 0, "        ");
-		while(*p && isspace(*p))
+		while(*p && isspace((unsigned char)*p))
 		  p++;
 	    }
 	    else{
@@ -666,8 +821,8 @@ mm_diskerror (stream, errcode, serious)
 
 		filename[0] = '\0';
 		build_path(full_filename, ps_global->home_dir, filename);
-		file_lister("DISK ERROR", full_filename, filename, FALSE,
-			    FB_SAVE);
+		file_lister("DISK ERROR", full_filename, MAXPATH+1, 
+                             filename, MAXPATH+1, FALSE, FB_SAVE);
 	    }
 
 	    break;
@@ -699,75 +854,6 @@ mm_diskerror (stream, errcode, serious)
 }
 
 
-
-/*----------------------------------------------------------------------
-      Receive list of folders from c-client/imap.
- 
- Puts an entry into the specified list of mail folders associated with
- the global context (aka collection). 
- ----*/
-   
-void
-context_mailbox(name)
-    char *name;
-{
-    dprint(4, (debugfile, "====== context_mailbox: (%s)\n", name));
-
-    /*
-     * "Inbox" is filtered out here, since pine only supports one true
-     * inbox...
-     */
-    if((!ps_global->show_dot_names && *name == '.')
-       || (!find_folder_inbox && !strucmp(name,"inbox")))
-       return;
-
-    find_folder_count++;
-    if(find_folder_list){
-	FOLDER_S *new_f     = new_folder(name);
-	new_f->prefix[0]    = '\0';
-	new_f->msg_count    = 0;
-	new_f->unread_count = 0;
-	folder_insert(-1, new_f, find_folder_list);
-    }
-}
-
-
-
-/*----------------------------------------------------------------------
-     Receive list of bulliten boards from c-client/imap
-
- Puts an entry into the specified list of bulletin boards associated with
- the global context (aka collection). 
- ----*/
-void
-context_bboard(name)
-    char *name;
-{
-    dprint(4, (debugfile, "====== context_bboard: (%s)\n", name));
-
-#ifdef NEWBB
-    /*----
-
-       Hack to handle mm_bboard calls that result from the 
-       nntp_find_new_bboards call to find the newly created news groups
-     ----*/
-    if(find_folder_list == newbb_folder_list) {
-        mark_folder_as_new(name);
-        return;
-    }
-#endif
-
-    if(!ps_global->show_dot_names && *name == '.')
-       return;
-
-    find_folder_count++;
-    if(find_folder_list){
-	FOLDER_S *new_f = new_folder(name);
-	folder_insert(folder_total(find_folder_list), new_f, find_folder_list);
-    }
-}
-
-
 void
 mm_fatal(message)
     char *message;
@@ -778,26 +864,119 @@ mm_fatal(message)
 
 void
 mm_flags(stream,number)
-    MAILSTREAM *stream;
-    long number;
+    MAILSTREAM    *stream;
+    unsigned long  number;
 {
-    long i;
-
     /*
      * The idea here is to clean up any data pine might have cached
      * that has anything to do with the indicated message number.
      * At the momment, this amounts only to cached index lines, but
      * watch out for future changes...
      */
-    if(stream != ps_global->mail_stream)
-      return;			/* only worry about displayed msgs */
+    if(stream == ps_global->mail_stream){
+	long i;
 
-    /* in case number is current, fix titlebar */
-    ps_global->mangled_header = 1;
+	/* then clear index entry */
+	if(i = mn_raw2m(ps_global->msgmap, (long) number))
+	  clear_index_cache_ent(i);
 
-    /* then clear index entry */
-    if(i = mn_raw2m(ps_global->msgmap, number))
-      clear_index_cache_ent(i);
+	/* in case number is current, fix titlebar */
+	if(mn_is_cur(ps_global->msgmap, i))
+	   ps_global->mangled_header = 1;
+
+	check_point_change();
+    }
+}
+
+
+
+
+/*
+ *
+ */
+void
+mm_status(stream, mailbox, status)
+    MAILSTREAM *stream;
+    char       *mailbox;
+    MAILSTATUS *status;
+{
+    mm_status_result = *status;
+
+#ifdef DEBUG
+    if(ps_global->debug_imap < 3){
+	dprint(0, (debugfile, " Mailbox %s",mailbox));
+	if (status->flags & SA_MESSAGES)
+	  dprint(0, (debugfile, ", %lu messages",status->messages));
+
+	if (status->flags & SA_RECENT)
+	  dprint(0, (debugfile, ", %lu recent",status->recent));
+
+	if (status->flags & SA_UNSEEN)
+	  dprint(0, (debugfile, ", %lu unseen",status->unseen));
+
+	if (status->flags & SA_UIDVALIDITY)
+	  dprint(0, (debugfile, ", %lu UID validity", status->uidvalidity));
+
+	if (status->flags & SA_UIDNEXT)
+	  dprint(0, (debugfile, ", %lu next UID",status->uidnext));
+
+	dprint(0, (debugfile, "\n"));
+    }
+#endif
+}
+
+
+/*
+ *
+ */
+void
+mm_list(stream, delimiter, mailbox, attributes)
+    MAILSTREAM *stream;
+    int		delimiter;
+    char       *mailbox;
+    long	attributes;
+{
+#ifdef DEBUG
+    if(ps_global->debug_imap > 2)
+      dprint(0,
+              (debugfile, "mm_list \"%s\": delim: '%c', %s%s%s%s\n",
+	       mailbox, delimiter ? delimiter : 'X',
+	       (attributes & LATT_NOINFERIORS) ? ", no inferiors" : "",
+	       (attributes & LATT_NOSELECT) ? ", no select" : "",
+	       (attributes & LATT_MARKED) ? ", marked" : "",
+	       (attributes & LATT_UNMARKED) ? ", unmarked" : ""));
+#endif
+
+    if(!mm_list_info->stream || stream == mm_list_info->stream)
+      (*mm_list_info->filter)(stream, mailbox, delimiter,
+			      attributes, mm_list_info->data);
+}
+
+
+/*
+ *
+ */
+void
+mm_lsub(stream, delimiter, mailbox, attributes)
+  MAILSTREAM *stream;
+  int	      delimiter;
+  char	     *mailbox;
+  long	      attributes;
+{
+#ifdef DEBUG
+    if(ps_global->debug_imap > 2)
+      dprint(0,
+              (debugfile, "LSUB \"%s\": delim: '%c', %s%s%s%s\n",
+	       mailbox, delimiter ? delimiter : 'X',
+	       (attributes & LATT_NOINFERIORS) ? ", no inferiors" : "",
+	       (attributes & LATT_NOSELECT) ? ", no select" : "",
+	       (attributes & LATT_MARKED) ? ", marked" : "",
+	       (attributes & LATT_UNMARKED) ? ", unmarked" : ""));
+#endif
+
+    if(!mm_list_info->stream || stream == mm_list_info->stream)
+      (*mm_list_info->filter)(stream, mailbox, delimiter,
+			      attributes, mm_list_info->data);
 }
 
 
@@ -811,20 +990,28 @@ mm_flags(stream,number)
  * pine_tcptimeout - C-client callback to handle tcp-related timeouts.
  */
 long
-pine_tcptimeout(t)
-    long t;
+pine_tcptimeout(elapsed, sincelast)
+    long elapsed, sincelast;
 {
     long rv = 1L;			/* keep trying by default */
-    int  ch;
+    int	 ch;
+
+#ifdef	DEBUG
+    dprint(1, (debugfile, "tcptimeout: waited %s seconds\n",
+	       long2string(elapsed)));
+    fflush(debugfile);
+#endif
 
     if(ps_global->noshow_timeout)
       return(rv);
 
+    suspend_busy_alarm();
+    
     /*
      * Prompt after a minute (since by then things are probably really bad)
      * A prompt timeout means "keep trying"...
      */
-    if(t >= TO_BAIL_THRESHOLD){
+    if(elapsed >= TO_BAIL_THRESHOLD){
 	int clear_inverse;
 
 	ClearLine(ps_global->ttyo->screen_rows - FOOTER_ROWS(ps_global));
@@ -833,7 +1020,7 @@ pine_tcptimeout(t)
 
 	PutLine1(ps_global->ttyo->screen_rows - FOOTER_ROWS(ps_global), 0,
        "\007Waited %s seconds for server reply.  Break connection to server? ",
-	   long2string(t));		/* if not DOS, add fflush() */
+	   long2string(elapsed));
 	CleartoEOLN();
 	fflush(stdout);
 	flush_input();
@@ -850,11 +1037,192 @@ pine_tcptimeout(t)
     if(rv == 1L){			/* just warn 'em something's up */
 	q_status_message1(SM_ORDER, 0, 0,
 		  "Waited %s seconds for server reply.  Still Waiting...",
-		  long2string(t));
+		  long2string(elapsed));
 	flush_status_messages(0);	/* make sure it's seen */
     }
 
     mark_status_dirty();		/* make sure it get's cleared */
+
+    resume_busy_alarm((rv == 1) ? 3 : 0);
+
+    return(rv);
+}
+
+
+char *
+imap_referral(stream, ref, code)
+    MAILSTREAM *stream;
+    char       *ref;
+    long	code;
+{
+    char *buf = NULL;
+
+    if(ref && !struncmp(ref, "imap://", 7)){
+	char *folder;
+	long  uid_val, uid;
+	int rv;
+
+	rv = url_imap_folder(ref, &folder, &uid, &uid_val, NULL, 1);
+	switch(rv & URL_IMAP_MASK){
+	    /*
+	     * Only permit URLs specifying a simple "imessagelist" 
+	     */
+	  case URL_IMAP_IMESSAGELIST :
+	    buf = cpystr(folder);
+
+	  case URL_IMAP_IMESSAGEPART :
+	  case URL_IMAP_IMAILBOXLIST :
+	    fs_give((void **) &folder);
+
+	  default :
+	    break;
+	}
+    }
+
+    return(buf);
+}
+
+
+long
+imap_proxycopy(stream, sequence, mailbox, flags)
+    MAILSTREAM *stream;
+    char       *sequence, *mailbox;
+    long	flags;
+{
+    SE_APP_S args;
+
+    args.folder = mailbox;
+    args.flags  = flags;
+
+    return(imap_seq_exec(stream, sequence, imap_seq_exec_append, &args));
+}
+
+
+long
+imap_seq_exec(stream, sequence, func, args)
+    MAILSTREAM	*stream;
+    char	*sequence;
+    long       (*func) PROTO((MAILSTREAM *, long, void *));
+    void	*args;
+{
+    unsigned long i,j,x;
+
+    while (*sequence) {		/* while there is something to parse */
+	if (*sequence == '*') {	/* maximum message */
+	    if(!(i = stream->nmsgs)){
+		mm_log ("No messages, so no maximum message number",ERROR);
+		return(0L);
+	    }
+
+	    sequence++;		/* skip past * */
+	}
+	else if (!(i = strtoul ((const char *) sequence,&sequence,10))
+		 || (i > stream->nmsgs)){
+	    mm_log ("Sequence invalid",ERROR);
+	    return(0L);
+	}
+
+	switch (*sequence) {	/* see what the delimiter is */
+	  case ':':			/* sequence range */
+	    if (*++sequence == '*') {	/* maximum message */
+		if (stream->nmsgs) j = stream->nmsgs;
+		else {
+		    mm_log ("No messages, so no maximum message number",ERROR);
+		    return NIL;
+		}
+
+		sequence++;		/* skip past * */
+	    }
+				/* parse end of range */
+	    else if (!(j = strtoul ((const char *) sequence,&sequence,10)) ||
+		     (j > stream->nmsgs)) {
+		mm_log ("Sequence range invalid",ERROR);
+		return NIL;
+	    }
+	    if (*sequence && *sequence++ != ',') {
+		mm_log ("Sequence range syntax error",ERROR);
+		return NIL;
+	    }
+
+	    if (i > j) {		/* swap the range if backwards */
+		x = i; i = j; j = x;
+	    }
+				/* mark each item in the sequence */
+	    while (i <= j)
+	      if(!(*func)(stream, j--, args)){
+		  mail_elt (stream,j--)->sequence = T;
+		  return(0L);
+	      }
+
+	    break;
+	  case ',':			/* single message */
+	    ++sequence;		/* skip the delimiter, fall into end case */
+	  case '\0':		/* end of sequence, mark this message */
+	    if(!(*func)(stream, i, args)){
+		mail_elt (stream,i)->sequence = T;
+		return(0L);
+	    }
+	    break;
+	  default:			/* anything else is a syntax error! */
+	    mm_log ("Sequence syntax error",ERROR);
+	    return NIL;
+	}
+    }
+
+    return T;			/* successfully parsed sequence */
+}
+
+
+long
+imap_seq_exec_append(stream, msgno, args)
+    MAILSTREAM *stream;
+    long	msgno;
+    void       *args;
+{
+    char	 *save_folder, flags[64], date[64];
+    CONTEXT_S    *cntxt = NULL;
+    int		  our_stream = 0;
+    long	  rv = 0L;
+    MAILSTREAM   *save_stream;
+    SE_APP_S	 *sa = (SE_APP_S *) args;
+    MESSAGECACHE *mc;
+    STORE_S      *so;
+
+    save_folder = (strucmp(sa->folder, ps_global->inbox_name) == 0)
+		    ? ps_global->VAR_INBOX_PATH : sa->folder;
+
+    save_stream = save_msg_stream(cntxt, save_folder, &our_stream);
+
+    if(so = so_get(CharStar, NULL, WRITE_ACCESS)){
+	/* store flags before the fetch so UNSEEN bit isn't flipped */
+	mc = mail_elt(stream, msgno);
+	flag_string(mc, F_ANS|F_FLAG|F_SEEN, flags);
+	if(mc->day)
+	  mail_date(date, mc);
+	else
+	  *date = '\0';
+
+	rv = save_fetch_append(stream, msgno, NULL,
+			       save_stream, save_folder, NULL,
+			       mc->rfc822_size, flags, date, so);
+	if(rv < 0 || ps_global->expunge_count){
+	    cmd_cancelled("Attached message Save");
+	    rv = 0L;
+	}
+	/* else whatever broke in save_fetch_append shoulda bitched */
+
+	so_give(&so);
+    }
+    else{
+	dprint(1, (debugfile, "Can't allocate store for save: %s\n",
+		   error_description(errno)));
+	q_status_message(SM_ORDER | SM_DING, 3, 4,
+			 "Problem creating space for message text.");
+    }
+
+    if(our_stream)
+      mail_close(save_stream);
+
     return(rv);
 }
 
@@ -872,11 +1240,13 @@ cached_user_name(host)
     char *host;
 {
     MMLOGIN_S *l;
+    STRLIST_S *h;
 
     if((l = mm_login_list) && host)
       do
-	if(!strucmp(host, l->host))
-	  return(l->user);
+	for(h = l->hosts; h; h = h->next)
+	  if(!strucmp(host, h->name))
+	    return(l->user);
       while(l = l->next);
 
     return(NULL);
@@ -884,50 +1254,87 @@ cached_user_name(host)
 
 
 int
-imap_get_passwd(l, passwd, user, host)
-    MMLOGIN_S *l;
-    char *passwd, *user, *host;
+imap_same_host(hl1, hl2)
+    STRLIST_S *hl1, *hl2;
 {
-    while(l){
-	/* host names and, if present, user name, must match */
-	if(!strucmp(host, l->host) && (!*user || !strcmp(user, l->user))){
-	    strcpy(user, l->user);
-	    strcpy(passwd, l->passwd);
-	    break;
-	}
-	else
-	  l = l->next;
-    }
+    STRLIST_S *lp;
 
-    return(l != NULL);
+    for( ; hl1; hl1 = hl1->next)
+      for(lp = hl2; lp; lp = lp->next)
+      if(!strucmp(hl1->name, lp->name))
+	return(TRUE);
+
+    return(FALSE);
+}
+
+
+int
+imap_get_passwd(m_list, passwd, user, hostlist)
+    MMLOGIN_S *m_list;
+    char      *passwd, *user;
+    STRLIST_S *hostlist;
+{
+    MMLOGIN_S *l;
+    STRLIST_S *h_list;
+    char      *h;
+    
+    for(l = m_list; l; l = l->next)
+      /* host name and user name must match */
+      if(imap_same_host(l->hosts, hostlist)
+	 && (!*user || !strcmp(user, l->user))){
+	  strcpy(user, l->user);
+	  strcpy(passwd, l->passwd);
+	  return(TRUE);
+      }
+
+    return(FALSE);
 }
 
 
 
 void
-imap_set_passwd(l, passwd, user, host)
+imap_set_passwd(l, passwd, user, hostlist)
     MMLOGIN_S **l;
-    char *passwd, *user, *host;
+    char       *passwd, *user;
+    STRLIST_S  *hostlist;
 {
-    while(*l){
-	if(!strucmp(host, (*l)->host)){
-	    fs_give((void **)&(*l)->user);
+    STRLIST_S *h_list, **listp;
+
+    for(; *l; l = &(*l)->next)
+      if(imap_same_host((*l)->hosts, hostlist)
+	 && !strcmp(user, (*l)->user))
+	if(strcmp(passwd, (*l)->passwd)){
 	    fs_give((void **)&(*l)->passwd);
 	    break;
 	}
 	else
-	  l = &(*l)->next;
-    }
+	  return;
 
     if(!*l){
 	*l = (MMLOGIN_S *)fs_get(sizeof(MMLOGIN_S));
 	memset(*l, 0, sizeof(MMLOGIN_S));
     }
 
-    (*l)->user   = cpystr(user);
+    if((*l)->passwd)
+      fs_give((void **) &(*l)->passwd);
+
     (*l)->passwd = cpystr(passwd);
-    if(!(*l)->host)
-      (*l)->host = cpystr(host);
+
+    if(!(*l)->user)
+      (*l)->user = cpystr(user);
+
+    for( ; hostlist; hostlist = hostlist->next){
+	for(listp = &(*l)->hosts;
+	    *listp && strucmp((*listp)->name, hostlist->name);
+	    listp = &(*listp)->next)
+	  ;
+
+	if(!*listp){
+	    *listp = (STRLIST_S *) fs_get(sizeof(STRLIST_S));
+	    (*listp)->name = cpystr(hostlist->name);
+	    (*listp)->next = NULL;
+	}
+    }
 }
 
 
@@ -936,27 +1343,27 @@ void
 imap_flush_passwd_cache()
 {
     MMLOGIN_S *l;
+
     while(l = mm_login_list){
 	mm_login_list = mm_login_list->next;
 	if(l->user)
-	  fs_give((void **)&l->user);
+	  fs_give((void **) &l->user);
 
-	if(l->host)
-	  fs_give((void **)&l->host);
+	free_strlist(&l->hosts);
 
 	if(l->passwd)
-	  fs_give((void **)&l->passwd);
+	  fs_give((void **) &l->passwd);
 
-	fs_give((void **)&l);
+	fs_give((void **) &l);
     }
 }
 
 
+#ifdef	PASSFILE
 /* 
- * DOS-specific functions to support caching username/passwd/host
+ * Specific functions to support caching username/passwd/host
  * triples on disk for use from one session to the next...
  */
-#if defined(DOS) || defined(OS2)
 
 #define	FIRSTCH		0x20
 #define	LASTCH		0x7e
@@ -1016,24 +1423,41 @@ char *
 passfile_name(pinerc, path)
     char *pinerc, *path;
 {
-    char *p = NULL;
-    int   i;
-    FILE *fp;
-#define	PASSFILE	"pine.pwd"
+    struct stat  sbuf;
+    char	*p = NULL;
+    int		 i, j;
 
     if(!path || !pinerc || pinerc[0] == '\0')
       return(NULL);
 
-    if((p = strrchr(pinerc, '\\')) == NULL)
-      p = strchr(pinerc, ':');
+    if(p = last_cmpnt(pinerc))
+      for(i = 0; pinerc < p; pinerc++, i++)
+	path[i] = *pinerc;
+    else
+      i = 0;
 
-    if(p){
-	strncpy(path, pinerc, i = (p - pinerc) + 1);
-	path[i] = '\0';
+    for(j = 0; path[i] = PASSFILE[j]; i++, j++)
+      ;
+
+#if	defined(DOS) || defined(OS2)
+    return((stat(path, &sbuf) == 0
+	    && ((sbuf.st_mode & S_IFMT) == S_IFREG))
+	     ? path : NULL);
+#else
+    /* First, make sure it's ours and not sym-linked */
+    if(lstat(path, &sbuf) == 0
+       && ((sbuf.st_mode & S_IFMT) == S_IFREG)
+       && sbuf.st_uid == getuid()){
+	/* if too liberal permissions, fix them */
+	if((sbuf.st_mode & 077) != 0)
+	  if(chmod(path, sbuf.st_mode & ~077) != 0)
+	    return(NULL);
+	
+	return(path);
     }
-
-    strcat(path, PASSFILE);
-    return(path);
+    else
+      return(NULL);
+#endif
 }
 
 
@@ -1069,8 +1493,13 @@ read_passfile(pinerc, l)
 	      tmp[i++] = '\0';			/* tie off data */
 	}
 
-	if(ui[0] && ui[1] && ui[2])		/* valid field? */
-	  imap_set_passwd(l, ui[0], ui[1], ui[2]);
+	if(ui[0] && ui[1] && ui[2]){		/* valid field? */
+	    STRLIST_S hostlist[1];
+
+	    hostlist[0].name = ui[2];
+	    hostlist[0].next = NULL;
+	    imap_set_passwd(l, ui[0], ui[1], hostlist);
+	}
     }
 
     fclose(fp);
@@ -1094,7 +1523,7 @@ write_passfile(pinerc, l)
 
     for(n = 0; l; l = l->next, n++){
 	/*** do any necessary ENcryption here ***/
-	sprintf(tmp, "%s\t%s\t%s\n", l->passwd, l->user, l->host);
+	sprintf(tmp, "%s\t%s\t%s\n", l->passwd, l->user, l->hosts->name);
 	xlate_key = n;
 	for(i = 0; tmp[i]; i++)
 	  tmp[i] = xlate_in(tmp[i]);
@@ -1113,11 +1542,12 @@ write_passfile(pinerc, l)
  *            as the pinerc with the name defined above.
  */
 int
-get_passfile_passwd(pinerc, passwd, user, host)
-    char *pinerc, *passwd, *user, *host;
+get_passfile_passwd(pinerc, passwd, user, hostlist)
+    char      *pinerc, *passwd, *user;
+    STRLIST_S *hostlist;
 {
     return((passfile_cache || read_passfile(pinerc, &passfile_cache))
-	     ? imap_get_passwd(passfile_cache, passwd, user, host) : 0);
+	     ? imap_get_passwd(passfile_cache, passwd, user, hostlist) : 0);
 }
 
 
@@ -1128,14 +1558,15 @@ get_passfile_passwd(pinerc, passwd, user, host)
  *            as the pinerc with the name defined above.
  */
 void
-set_passfile_passwd(pinerc, passwd, user, host)
-    char *pinerc, *passwd, *user, *host;
+set_passfile_passwd(pinerc, passwd, user, hostlist)
+    char      *pinerc, *passwd, *user;
+    STRLIST_S *hostlist;
 {
     if((passfile_cache || read_passfile(pinerc, &passfile_cache))
        && want_to("Preserve password on DISK for next login", 'y', 'x',
-		  NO_HELP, 0, 0) == 'y'){
-	imap_set_passwd(&passfile_cache, passwd, user, host);
+		  NO_HELP, WT_NORM) == 'y'){
+	imap_set_passwd(&passfile_cache, passwd, user, hostlist);
 	write_passfile(pinerc, passfile_cache);
     }
 }
-#endif	/* DOS */
+#endif	/* PASSFILE */
