@@ -1,5 +1,5 @@
 #if !defined(lint) && !defined(DOS)
-static char rcsid[] = "$Id: newmail.c,v 4.68 1996/07/09 22:20:17 mikes Exp $";
+static char rcsid[] = "$Id: newmail.c,v 4.88 1999/01/27 20:38:12 hubert Exp $";
 #endif
 /*----------------------------------------------------------------------
 
@@ -22,7 +22,7 @@ static char rcsid[] = "$Id: newmail.c,v 4.68 1996/07/09 22:20:17 mikes Exp $";
    permission of the University of Washington.
 
    Pine, Pico, and Pilot software and its included text are Copyright
-   1989-1996 by the University of Washington.
+   1989-1999 by the University of Washington.
 
    The full text of our legal notices is contained in the file called
    CPYRIGHT, included with this distribution.
@@ -52,7 +52,8 @@ static char rcsid[] = "$Id: newmail.c,v 4.68 1996/07/09 22:20:17 mikes Exp $";
 /*
  * Internal prototypes
  */
-void new_mail_mess PROTO((MAILSTREAM *, char *, long, long, int));
+int  check_point PROTO((CheckPointTime, int));
+void new_mail_mess PROTO((MAILSTREAM *, char *, long, long));
 void fixup_flags PROTO((MAILSTREAM *, MSGNO_S *, long));
 
 
@@ -64,7 +65,7 @@ static long mailbox_mail_since_command = 0L,
  
   Input:  force       -- flag indicating we should check for new mail no matter
           time_for_check_point -- 0: GoodTime, 1: BadTime, 2: VeryBadTime
-	  do_status_message -- flag telling us to q a new mail status message
+	  flags -- whether to q a new mail status message or defer the sort
 
   Result: returns -1 if there was no new mail. Otherwise returns the
           sorted message number of the smallest message number that
@@ -83,55 +84,60 @@ static long mailbox_mail_since_command = 0L,
  
  --- */
 long
-new_mail(force, time_for_check_point, do_status_msg)
-    int force, time_for_check_point, do_status_msg;
+new_mail(force, time_for_check_point, flags)
+    int force, time_for_check_point, flags;
 {
     static time_t last_check = 0;
     static time_t last_check_point_call = 0;
     time_t        now;
-    long          rv;
+    long          rv = 0;
     MAILSTREAM   *stream;
     register struct pine *pine_state;
     int           checknow = 0;
-    int           act_on_deferred = 0;
-    int           deferring_sort;
 
     dprint(9, (debugfile, "new mail called (%d %d %d)\n",
-               force, time_for_check_point, do_status_msg));
+               force, time_for_check_point, flags));
     pine_state = ps_global;  /*  this gets called out of the composer which
                               *  doesn't know about pine_state
                               */
     now = time(0);
 
-    deferring_sort = pine_state->sort_is_deferred;	/* for this time */
-    pine_state->sort_is_deferred = 0;			/* for next time */
-    if(!deferring_sort && pine_state->unsorted_newmail)
-      act_on_deferred++;
+    if(time_for_check_point == 0)
+      adrbk_maintenance();
+
+    if(!force && pine_state->unsorted_newmail)
+      force = !(flags & NM_DEFER_SORT);
 
     /*
      * only check every 15 seconds, unless we're compelled to
      */
     if(!(stream = pine_state->mail_stream)
-       || (timeout == 0 && !force)
+       || (timeo == 0 && !force)
        || (now-last_check_point_call <= 15 && time_for_check_point != 0
-	   && !pine_state->mail_box_changed && !force
-	   && !act_on_deferred))
+	   && !pine_state->mail_box_changed && !force))
       return(-1);
-    else if(force || now-last_check >= timeout-2){ /* 2: check each timeout */
+    else if(force || now-last_check >= timeo-2){ /* 2: check each timeout */
 	checknow++;
         last_check = now;
     }
 
     last_check_point_call = now;
 
-    if(!check_point(time_for_check_point == 0 ? GoodTime:
-	      time_for_check_point == 1 ? BadTime : VeryBadTime) && checknow){
-	if(F_ON(F_SHOW_DELAY_CUE, ps_global) && !ps_global->in_init_seq){
+    if(!check_point((time_for_check_point == 0)
+		      ? GoodTime
+		      : (time_for_check_point == 1)
+			  ? BadTime : VeryBadTime, flags)
+       && checknow){
+	if((flags & NM_STATUS_MSG) && F_ON(F_SHOW_DELAY_CUE, ps_global)
+	   && !ps_global->in_init_seq){
 	    StartInverse();
 	    PutLine0(0, 1, "*");	/* Show something to indicate delay */
 	    MoveCursor(ps_global->ttyo->screen_rows -FOOTER_ROWS(ps_global),0);
 	    fflush(stdout);
 	}
+#ifdef _WINDOWS
+	mswin_setcursor (MSWIN_CURSOR_BUSY);
+#endif
 
 #ifdef	DEBUG
 	now = time(0);
@@ -146,21 +152,30 @@ new_mail(force, time_for_check_point, do_status_msg)
 	now = time(0);
 	dprint(7, (debugfile, "Ping complete: %s\n", ctime(&now)));
 #endif
-	if(F_ON(F_SHOW_DELAY_CUE, ps_global) && !ps_global->in_init_seq){
+	if((flags & NM_STATUS_MSG)
+	   && F_ON(F_SHOW_DELAY_CUE, ps_global)
+	   && !ps_global->in_init_seq){
 	    PutLine0(0, 1, " ");
 	    EndInverse();
 	    fflush(stdout);
 	}
+#ifdef _WINDOWS
+	mswin_setcursor (MSWIN_CURSOR_ARROW);
+#endif
     }
 
     if(checknow && pine_state->inbox_stream 
        && stream != pine_state->inbox_stream){
-	if(F_ON(F_SHOW_DELAY_CUE, ps_global)){
+	if((flags & NM_STATUS_MSG)
+	   && F_ON(F_SHOW_DELAY_CUE, ps_global)){
 	    StartInverse();
 	    PutLine0(0, 1, "*"); 	/* Show something to indicate delay */
 	    MoveCursor(ps_global->ttyo->screen_rows -FOOTER_ROWS(ps_global),0);
 	    fflush(stdout);
 	}
+#ifdef _WINDOWS
+	mswin_setcursor (MSWIN_CURSOR_BUSY);
+#endif
 
 #ifdef	DEBUG
 	now = time(0);
@@ -173,18 +188,22 @@ new_mail(force, time_for_check_point, do_status_msg)
 	now = time(0);
 	dprint(7, (debugfile, "Ping complete: %s\n", ctime(&now)));
 #endif
-	if(F_ON(F_SHOW_DELAY_CUE, ps_global)){
+	if((flags & NM_STATUS_MSG)
+	   && F_ON(F_SHOW_DELAY_CUE, ps_global)){
 	    PutLine0(0, 1, " ");
 	    EndInverse();
 	    fflush(stdout);
 	}
+#ifdef _WINDOWS
+	mswin_setcursor (MSWIN_CURSOR_ARROW);
+#endif
     }
 
 
     /*-------------------------------------------------------------
        Mail box state changed, could be additions or deletions.
       -------------------------------------------------------------*/
-    if(pine_state->mail_box_changed || act_on_deferred) {
+    if(pine_state->mail_box_changed || pine_state->unsorted_newmail) {
         dprint(7, (debugfile,
         "New mail, %s,  new_mail_count:%d  expunge count:%d,  max_msgno:%d\n",
                    pine_state->mail_stream == pine_state->inbox_stream ?
@@ -201,32 +220,25 @@ new_mail(force, time_for_check_point, do_status_msg)
 	 * Lastly, worry about sorting if we got something new, otherwise
 	 * it was taken care of inside mm_expunge...
 	 */
-	if((pine_state->new_mail_count > 0L || pine_state->unsorted_newmail)
-	   && !deferring_sort
-	   && !any_lflagged(pine_state->msgmap, MN_HIDE)){
-	    /*
-	     * When new mail arrived, mm_exists cleared the cache.
-	     * However, if we deferred the sort at that time and are
-	     * sorting now, we need to clear the cache again.
-	     */
-	    if(pine_state->unsorted_newmail)
-	      clear_index_cache();
-
-	    sort_current_folder(1, mn_get_sort(pine_state->msgmap),
-				mn_get_revsort(pine_state->msgmap));
-	}
+	if((pine_state->new_mail_count > 0L
+	    || pine_state->unsorted_newmail)
+	   && !((flags & NM_DEFER_SORT)
+		|| any_lflagged(pine_state->msgmap, MN_HIDE)))
+	  refresh_sort(pine_state->msgmap, (flags & NM_STATUS_MSG));
 	else if(pine_state->new_mail_count > 0L)
 	  pine_state->unsorted_newmail = 1;
 
-        if(ps_global->new_mail_count > 0 || mailbox_mail_since_command) {
-            mailbox_mail_since_command += ps_global->new_mail_count;
+        if(pine_state->new_mail_count > 0) {
+            mailbox_mail_since_command += pine_state->new_mail_count;
+	    rv                         += pine_state->new_mail_count;
+	    ps_global->new_mail_count   = 0L;
 
-            new_mail_mess(pine_state->mail_stream,
-                      pine_state->mail_stream == pine_state->inbox_stream ?
-                      NULL :  pine_state->cur_folder,
-                      mailbox_mail_since_command,
-		      mn_get_total(pine_state->msgmap),
-		      do_status_msg);
+	    if(flags & NM_STATUS_MSG)
+	      new_mail_mess(pine_state->mail_stream,
+			 (pine_state->mail_stream == pine_state->inbox_stream)
+			    ? NULL :  pine_state->cur_folder,
+			 mailbox_mail_since_command,
+			 mn_get_total(pine_state->msgmap));
         }
     }
 
@@ -243,31 +255,24 @@ new_mail(force, time_for_check_point, do_status_msg)
 	fixup_flags(pine_state->inbox_stream, pine_state->inbox_msgmap,
 		    pine_state->inbox_new_mail_count);
 
-        if(pine_state->inbox_new_mail_count > 0 || inbox_mail_since_command) {
-            inbox_mail_since_command       += ps_global->inbox_new_mail_count;
+        if(pine_state->inbox_new_mail_count > 0) {
+            inbox_mail_since_command       += pine_state->inbox_new_mail_count;
+	    rv                             += pine_state->inbox_new_mail_count;
             ps_global->inbox_new_mail_count = 0;
-            new_mail_mess(pine_state->inbox_stream,NULL,
-                          inbox_mail_since_command,
-			  mn_get_total(pine_state->inbox_msgmap),
-			  do_status_msg);
+
+	    if(flags & NM_STATUS_MSG)
+	      new_mail_mess(pine_state->inbox_stream, NULL,
+			    inbox_mail_since_command,
+			    mn_get_total(pine_state->inbox_msgmap));
         }
     }
 
-    rv = mailbox_mail_since_command
-	 + inbox_mail_since_command + pine_state->expunge_count;
-    if(do_status_msg){
+    rv += pine_state->expunge_count;
+    if(flags & NM_STATUS_MSG){
 	pine_state->inbox_changed    = 0;
 	pine_state->mail_box_changed = 0;
     }
 
-    ps_global->new_mail_count    = 0L;
-
-    if(pine_state->shown_newmail
-       && !(mailbox_mail_since_command || inbox_mail_since_command)){
-	icon_text(NULL);
-	pine_state->shown_newmail = 0;
-    }
-
     dprint(6, (debugfile, "******** new mail returning %ld  ********\n", 
 	       rv ? rv : -1));
     return(rv ? rv : -1);
@@ -287,15 +292,15 @@ status_message code will fit what it can on the screen and truncation on
 the right is about what we want which is what will happen.
   ----*/
 void
-new_mail_mess(stream, folder, number, max_num, do_status_msg)
+new_mail_mess(stream, folder, number, max_num)
      MAILSTREAM *stream;
      long        number, max_num;
      char       *folder;
-     int         do_status_msg;
 
 {
     ENVELOPE	*e;
-    char	 subject[200], from[2*MAX_SCREEN_COLS], intro[50 + MAXFOLDER];
+    char	 subject[200], from[2*MAX_SCREEN_COLS],
+		 intro[MAX_SCREEN_COLS+1];
     static char *carray[] = { "regarding",
 				"concerning",
 				"about",
@@ -310,18 +315,14 @@ new_mail_mess(stream, folder, number, max_num, do_status_msg)
 				"subject:"
     };
 
-#ifdef _WINDOWS
-    mswin_newmailicon ();
-#endif
-
-    if(!do_status_msg)
-      return;
+    e = mail_fetchstructure(stream, max_num, NULL);
 
     if(!folder) {
         if(number > 1)
           sprintf(intro, "%ld new messages!", number);
         else
-          strcpy(intro, "New mail!");
+	  sprintf(intro, "New mail%s!",
+		  (e && address_is_us(e->to, ps_global)) ? " to you" : "");
     }
     else {
         if(number > 1)
@@ -330,7 +331,7 @@ new_mail_mess(stream, folder, number, max_num, do_status_msg)
           sprintf(intro, "Mail saved to folder \"%s\"", folder);
     }
 
-    if((e = mail_fetchstructure(stream, max_num, NULL)) && e->from){
+    if(e && e->from){
 	sprintf(from, " %srom ", (number > 1L) ? "Most recent f" : "F");
 	if(e->from->personal)
 	  istrncpy(from + ((number > 1L) ? 18 : 6),
@@ -338,8 +339,10 @@ new_mail_mess(stream, folder, number, max_num, do_status_msg)
 					   e->from->personal, NULL),
 		   ps_global->ttyo->screen_cols);
 	else
-	  sprintf(from + ((number > 1L) ? 18 : 6), "%s@%s", 
-		  e->from->mailbox, e->from->host);
+	  sprintf(from + ((number > 1L) ? 18 : 6), "%s%s%s", 
+		  e->from->mailbox,
+		  e->from->host ? "@" : "",
+		  e->from->host ? e->from->host : "");
     }
     else
       from[0] = '\0';
@@ -365,7 +368,6 @@ new_mail_mess(stream, folder, number, max_num, do_status_msg)
 
     sprintf(tmp_20k_buf, "%s%s", intro, from);
     icon_text(tmp_20k_buf);
-    ps_global->shown_newmail = 1;
 }
 
 
@@ -453,8 +455,9 @@ static long first_status_change = 0;  /* time of 1st change since last chk_pt*/
 static long last_status_change	= 0;  /* time of last change                 */
 static long check_count_ave	= 10 * 10;
 
-check_point(timing)
-     CheckPointTime timing;
+check_point(timing, flags)
+    CheckPointTime timing;
+    int		   flags;
 {
     int     freq, tm;
     long    adj_cca;
@@ -475,21 +478,18 @@ check_point(timing)
     freq = CHECK_POINT_FREQ * (timing==GoodTime ? 1 : timing==BadTime ? 3 : 4);
     tm   = CHECK_POINT_TIME * (timing==GoodTime ? 1 : timing==BadTime ? 2 : 3);
 
-    dprint(9, (debugfile, "freq: %d  count: %d\n", freq, check_count));
-    dprint(9, (debugfile, "tm: %d  elapsed: %d\n", tm,
-               time(0) - first_status_change));
-
     if(!last_status_change)
         last_status_change = time(0);
 
     tmp = 10*(time(0)-last_status_change);
-    adj_cca = (tmp > check_count_ave) ?
-	  min((check_count_ave + tmp)/2, 2*check_count_ave) : check_count_ave;
+    adj_cca = (check_count_ave > tmp || check_count_ave > 200)
+		? check_count_ave
+		: min((check_count_ave + tmp)/2, 200);
 
     dprint(9, (debugfile, "freq %d tm %d changes %d since_1st_change %d\n",
 	       freq, tm, check_count, time(0)-first_status_change));
     dprint(9, (debugfile, "since_status_chg %d chk_cnt_ave %d (tenths)\n",
-	       time(0)-last_status_change, check_count_ave));
+	       tmp, check_count_ave));
     dprint(9, (debugfile, "adj_chk_cnt_ave %d (tenths)\n", adj_cca));
     dprint(9, (debugfile, "Check:if changes(%d)xadj_cca(%d) >= freq(%d)x200\n",
 	       check_count, adj_cca, freq));
@@ -506,12 +506,15 @@ check_point(timing)
                      "Doing checkpoint: %s  Since 1st status change: %d\n",
                      ctime(&now), now - first_status_change));
 #endif
-	if(F_ON(F_SHOW_DELAY_CUE, ps_global)){
+	if((flags & NM_STATUS_MSG) && F_ON(F_SHOW_DELAY_CUE, ps_global)){
 	    StartInverse();
 	    PutLine0(0, 0, "**");	/* Show something indicate delay*/
 	    MoveCursor(ps_global->ttyo->screen_rows -FOOTER_ROWS(ps_global),0);
 	    fflush(stdout);
 	}
+#ifdef _WINDOWS
+	mswin_setcursor (MSWIN_CURSOR_BUSY);
+#endif
 
         mail_check(ps_global->mail_stream);
 					/* Causes mail file to be written */
@@ -521,11 +524,14 @@ check_point(timing)
 #endif
         check_count = 0;
         first_status_change = time(0);
-	if(F_ON(F_SHOW_DELAY_CUE, ps_global)){
+	if((flags & NM_STATUS_MSG) && F_ON(F_SHOW_DELAY_CUE, ps_global)){
 	    PutLine0(0, 0, "  ");
 	    EndInverse();
 	    fflush(stdout);
 	}
+#ifdef _WINDOWS
+	mswin_setcursor (MSWIN_CURSOR_ARROW);
+#endif
 
         return(1);
     } else {
@@ -581,6 +587,13 @@ void
 zero_new_mail_count()
 {
     dprint(9, (debugfile, "New_mail_count zeroed\n"));
+
+    /*
+     * Decide if likewise need to clean up the new mail icon...
+     */
+    if(mailbox_mail_since_command || inbox_mail_since_command)
+      icon_text(NULL);
+
     mailbox_mail_since_command = 0L;
     inbox_mail_since_command   = 0L;
 }
