@@ -1,6 +1,6 @@
 #line 2 "osdep/os-os2.h"
 /*----------------------------------------------------------------------
-  $Id: os-os2.h,v 4.7 1996/06/05 17:12:38 mikes Exp $
+  $Id: os-os2.h,v 4.15 1998/11/24 01:56:27 mikes Exp $
 
             T H E    P I N E    M A I L   S Y S T E M
 
@@ -21,7 +21,7 @@
    permission of the University of Washington.
 
    Pine, Pico, and Pilot software and its included text are Copyright
-   1989-1996 by the University of Washington.
+   1989-1998 by the University of Washington.
 
    The full text of our legal notices is contained in the file called
    CPYRIGHT, included with this distribution.
@@ -41,11 +41,13 @@
 #ifndef _OS_INCLUDED
 #define _OS_INCLUDED
 
+#include "../c-client/osdep.h"
+
 
 /*----------------------------------------------------------------------
 
    This first section has some constants that you may want to change
-   for your configuration.  This is the DOS version of the os.h file.
+   for your configuration.  This is the OS/2 version of the os.h file.
    Further down in the file are os-dependent things that need to be set up
    correctly for each os.  They aren't interesting, they just have to be
    right.  There are also a few constants down there that may be of
@@ -54,22 +56,28 @@
  ----*/
 
 /*----------------------------------------------------------------------
- Define this if you want the disk quota to be checked on startup.
- Of course, this only makes sense if your system has quotas.  If it doesn't,
- there should be a dummy disk_quota() routine in os-xxx.c so that defining
- this won't be harmful anyway.
+   Define this if you want the disk quota to be checked on startup.
+   Of course, this only makes sense if your system has quotas.  If it doesn't,
+   there should be a dummy disk_quota() routine in os-xxx.c so that defining
+   this won't be harmful anyway.
  ----*/
 /* #define USE_QUOTAS */
 
 
 
 /*----------------------------------------------------------------------
-   Define this if you want to allow the users to change their From header
-   line when they send out mail.  The users will still have to configure
-   either default-composer-hdrs or customized-hdrs to get at the From
-   header, even if this is set.
+   Define this if you want to be sure to not allow users to change their
+   From header line when they send out mail. Even if you don't define
+   this the default is to not allow From to be changed. The user would have
+   to edit (by hand) "allow-changing-from" into their .pinerc feature list
+   or the sys-admin would have to include it in the default feature-list.
+   Even after that, the user still has to include it in either their
+   default-composer-hdrs or customized-hdrs to get at it. Instead of
+   defining NEVER_ALLOW_CHANGING_FROM, an easier way of preventing From
+   changing is to put the feature "no-allow-changing-from" in the
+   /usr/local/lib/pine.conf.fixed file.
  ----*/
-/* #define ALLOW_CHANGING_FROM  /* comment out to not allow changing From */
+/* #define NEVER_ALLOW_CHANGING_FROM */
 
 
 
@@ -81,8 +89,8 @@
 
 
 /*----------------------------------------------------------------------
-    Turn this on if it's OK to encode sent message text containing
-  "From " at the beginning of any line.
+    Turn this on to trigger QP encoding of sent message text if it contains
+  "From " at the beginning of a line or "." on a line by itself.
  ----*/
 /* #define ENCODE_FROMS */
 
@@ -99,9 +107,11 @@
 /*----------------------------------------------------------------------
     Timeouts (seconds)
  ----*/
-#define DF_MAILCHECK      "150" /* How often to check for new mail. There's 
-                                   some expense in doing this so it shouldn't
-                                   be done too frequently.  */
+#define DF_MAILCHECK      "150" /* How often to check for new mail, by
+				   default.  There's some expense in doing
+				   this so it shouldn't be done too
+				   frequently.  (Can be set in config
+				   file now.)  */
 
 /*----------------------------------------------------------------------
     Check pointing (seconds)
@@ -125,7 +135,7 @@
  In scrolling through text, the number of lines from the previous
  screen to overlap when showing the next screen.  Usually set to two.
  ----*/
-#define	DF_OVERLAP	    "2"
+#define	DF_OVERLAP	"2"
 
 
 
@@ -148,19 +158,20 @@
 
 
 /*----- System-wide config file ----------------------------------------*/
-#define SYSTEM_PINERC "pinerc"
+#define SYSTEM_PINERC             "pinerc"
 
 
 
 /*----- Used only in DOS -----------------------------------------------*/
 #define DF_PINEDIR      "\\pine"
-#ifdef	HELPFILE
-#undef	HELPFILE
+#ifdef HELPFILE
+#undef HELPFILE
 #define HELPFILE        "pine.hlp"
 #define HELPINDEX       "pine.ndx"
 #endif
 
-  
+
+
 /*----------------------------------------------------------------------
    The default folder names and folder directories (some for backwards
    compatibility).  Think hard before changing any of these.
@@ -171,7 +182,7 @@
 #define POSTPONED_MSGS            "postpond"
 #define INTERRUPTED_MAIL          "intruptd"
 #define DEADLETTER                "deadletr"
-#define	DF_FOLDER_EXTENSION	  "mtx"
+#define DF_FOLDER_EXTENSION       "mtx"
 #define DF_MAIL_DIRECTORY         "mail"
 #define INBOX_NAME                "INBOX"
 #define DF_SIGNATURE_FILE         "pine.sig"
@@ -183,6 +194,7 @@
 #define DF_SAVE_BY_SENDER         "no"
 #define DF_SORT_KEY               "arrival"
 #define DF_AB_SORT_RULE           "fullname-with-lists-last"
+#define DF_FLD_SORT_RULE          "alphabetical"
 #define DF_SAVED_MSG_NAME_RULE    "default-folder"
 #define DF_FCC_RULE               "default-fcc"
 #define DF_STANDARD_PRINTER       "PRN"
@@ -197,6 +209,7 @@
 #define DF_LOCAL_FULLNAME         "Local Support"
 #define DF_LOCAL_ADDRESS          "postmaster"
 #define DF_KBLOCK_PASSWD_COUNT    "1"
+#define DF_REMOTE_ABOOK_HISTORY   "3"
 
 /*----------------------------------------------------------------------
    The default printer when pine starts up for the first time with no printer
@@ -205,185 +218,39 @@
 
 
 
-/*----------------------------------------------------------------------
-
-   OS dependencies, OS/2 version.  See also the os-os2.c files.
-   The following stuff may need to be changed for a new port, but once
-   the port is done it won't change.  Further down in the file are a few
-   constants that you may want to configure differently than they
-   are configured, but probably not.
-
- ----*/
-
-
-
-/*----------------- Are we ANSI? ---------------------------------------*/
-#define ANSI
-
-/*------ If our compiler doesn't understand type void ------------------*/
-/* #define void char */
-
-/*-------- Standard ANSI functions usually defined in stdlib.h ---------*/
-#include <stdlib.h>
-/*---- Declare getenv() if not already declared ------------------------*/
-/* char *getenv(); */
-
-
-/*------- Standard string functions ------------------------------------*/
-#include <string.h>
-/*---- Declare these if not already declared ---------------------------*/
-/* char *strcpy(), *strncpy(); */
-
-
-/*------- Include types.h if you have it -------------------------------*/
-#include <sys/types.h>
-
-
-/*------- Some systems need this ---------------------------------------*/
-/* #include <stdarg.h> */
-
-
-#include <io.h>
-#include <sys/stat.h>
-#include <sys/socket.h>
-#include <netinet/in.h>
-#include <netdb.h>
-#include <memory.h>
-#include <fcntl.h>
-#include <sys/timeb.h>
-
-#undef ERROR
-
-
-/*---- Declare sys_errlist() if not already declared -------------------*/
-/* extern char *sys_errlist[]; */
-
-
-
-/*----------------- locale.h -------------------------------------------*/
-/* #include <locale.h>  /* To make matching and sorting work right */
-
-
-
-/*----------------- time.h ---------------------------------------------*/
-#include <time.h>
-/* plain time.h isn't enough on some systems */
-#include <sys/time.h>  /* For struct timeval usually in time.h */ 
-
-
-/*--------------- signal.h ---------------------------------------------*/
-#include <signal.h>
-/* #include <sys/signal.h> */
-
-#undef	SIGCHLD				/* clear unsupp'd signal def */
-
-#define SigType void
-/* #define SigType int */
-
-
-/* #define POSIX_SIGNALS */
-/* #define SYSV_SIGNALS    /* use System-V signal semantics (ttyin.c) */
-
-
-/*-------------- A couple typedef's for integer sizes ------------------*/
-typedef unsigned long usign32_t;
-typedef unsigned short usign16_t;
-
-
-/*-------------- qsort argument type -----------------------------------*/
-#define QSType void
-/* #define QSType char */
-
-
-
-/*------ how help text is referenced (always char ** on Unix) ----------*/
-#ifdef HELPFILE
-#define HelpType short
-#define NO_HELP  (-1)
-#else
-#define HelpType char **
-#define NO_HELP (char **)NULL
-#endif
-
-
-
-/*--------- System V terminal driver -----------------------------------*/
-/* #define HAVE_TERMIO   this is for pure System V */
-/* #define HAVE_TERMIOS  this is an alternative    */
-/*- Use the poll() system call instead of select() */
-/* #define USE_POLL */
-/*---- Use terminfo instead of termcap ---------------------------------*/
-/* #define USE_TERMINFO */
-
-
-
-/*-- What argument does wait(2) take? Define this if it is a union -----*/
-/* #define HAVE_WAIT_UNION */
-
-
-/*-------- Is window resizing available? -------------------------------*/
-/* #define RESIZING */
-	/* Actually, under windows it is, but RESIZING compiles in UNIX 
-	 * signals code for determining when the window resized.  Window's
-	 * works differently. */
-
-
-
-
-/*-------- If no vfork, use regular fork -------------------------------*/
-#define vfork fork
-
-
-
-/*----------------------------------------------------------------------
-    This is DOS specific and is used to determine if a mail message is
- too big to fit in core.  It's just a static value now, but could be
- turned into a fuller macro that takes an adaptive approach to figure
- out what can fit into available free storage.
- Examining mailboxes around here, it looks like 90-something percent of
- messages come in under 8K so that seems like an OK limit on what can
- be dealt with in core.
- ----*/
-/* #define MAX_MSG_INCORE  (8192L) */
-
-
-
-/*----------------------------------------------------------------------
-    This is DOS specific and is used to pass the results of dos_gets
- calls back to functions that had previously installed it as the
- c-client "mailgets" allocation function.  It's basically a way around
- c-client's exclusive use of malloc'd strings which is quite
- impractical if we want to deal with text objects bigger than 64K on
- DOS (assuming we even have 64 contiguous K anyway!).
- ----*/
-/* extern  FILE    *append_file; */
-
-
-
 /*----- The usual sendmail configuration for sending mail on Unix ------*/
 /* #define SENDMAIL	"/usr/lib/sendmail" */
-/* #define SENDMAILFLAGS	"-oi -oem -t"	*/
+/* #define SENDMAILFLAGS	"-bs -odb -oem"	*/
+
+
+/*----------------------------------------------------------------------
+   If no nntp-servers are defined, this program will be used to post news.
+ ----*/
+/* #define SENDNEWS	"/usr/local/bin/inews -h" */
 
 
 /*--------- Program employed by users to change their password ---------*/
-/* #define	PASSWD_PROG	"" */
+/* #define	PASSWD_PROG	"/bin/passwd" */
 
 
 /*-------------- A couple constants used to size arrays ----------------*/
-/* #include <sys/param.h>          /* Get it from param.h if available */
-/* #define MAXPATH MAXPATHLEN      /* Longest pathname we ever expect */
+/* #include <sys/param.h>   */     /* Get it from param.h if available */
+/* #define MAXPATH MAXPATHLEN */   /* Longest pathname we ever expect */
 #define MAXPATH        (256)    /* Longest pathname we ever expect */
 #define MAXFOLDER      (64)     /* Longest foldername we ever expect */  
 
 
 /*-- Max screen pine will display on. Used to define some array sizes --*/
-#define MAX_SCREEN_COLS  (170) 
-#define MAX_SCREEN_ROWS  (200) 
+#define MAX_SCREEN_COLS  (170)
+#define MAX_SCREEN_ROWS  (200)
 
 
-/*---- When no screen size can be discovered this is the size used -----*/
-#define DEFAULT_LINES_ON_TERMINAL	(25)
-#define DEFAULT_COLUMNS_ON_TERMINAL	(80)
+
+/*----------------------------------------------------------------------
+    File name used to store the user's server/id/password triple between
+ session.  It is rooted in the same directory as the PINERC.
+ ----*/
+#define	PASSFILE	"pine.pwd"
 
 
 /*----------------------------------------------------------------------
@@ -394,7 +261,6 @@ typedef unsigned short usign16_t;
 #define DEBUGFILE	"pinedebg.txt"
 #define CRASHFILE	"pinecrsh.txt"
 
-
 /*----------------------------------------------------------------------
     The number of debug files to save in the user's home diretory. The files
  are useful for figuring out what a user did when he complains that something
@@ -426,17 +292,11 @@ typedef unsigned short usign16_t;
  ----*/
 #define MAX_FULLNAME     (100) 
 #define MAX_NICKNAME      (40)
-#define MAX_ADDRESS      (200)
+#define MAX_ADDRESS      (500)
 #define MAX_NEW_LIST     (500)  /* Max addrs to be added when creating list */
 #define MAX_SEARCH       (100)  /* Longest string to search for             */
 #define MAX_ADDR_EXPN   (1000)  /* Longest expanded addr                    */
 #define MAX_ADDR_FIELD (10000)  /* Longest fully-expanded addr field        */
 
-extern int kbd_ready();
-extern int kbd_getkey();
-
-#ifndef fflush
-#define fflush os2_fflush
-#endif
 
 #endif /* _OS_INCLUDED */
