@@ -1,5 +1,5 @@
 #if !defined(lint) && !defined(DOS)
-static char rcsid[] = "$Id: signals.c,v 4.105 1996/06/05 04:47:09 mikes Exp $";
+static char rcsid[] = "$Id: signals.c,v 4.127 1998/11/12 22:48:57 hubert Exp $";
 #endif
 /*----------------------------------------------------------------------
 
@@ -22,7 +22,7 @@ static char rcsid[] = "$Id: signals.c,v 4.105 1996/06/05 04:47:09 mikes Exp $";
    permission of the University of Washington.
 
    Pine, Pico, and Pilot software and its included text are Copyright
-   1989-1996 by the University of Washington.
+   1989-1998 by the University of Washington.
 
    The full text of our legal notices is contained in the file called
    CPYRIGHT, included with this distribution.
@@ -61,12 +61,7 @@ void  fast_clean_up PROTO(());
 void  suspend_notice PROTO((char *));
 void  suspend_warning PROTO(());
 void  alarm_signal_reset PROTO(());
-
-#if defined(DOS) || defined(OS2)
-#define	SIG_PROTO(args) args
-#else
-#define	SIG_PROTO(args) ()
-#endif
+void  alarm_wrapper PROTO ((unsigned));
 
 
               /* SigType is defined in os.h and is either int or void */
@@ -77,6 +72,7 @@ static	SigType	alarm_signal SIG_PROTO((int));
 static	SigType	intr_signal SIG_PROTO((int));
 
 
+
 /*----------------------------------------------------------------------
     Install handlers for all the signals we care to catch
   ----------------------------------------------------------------------*/
@@ -274,12 +270,44 @@ hup_signal()
     end_signals(1);			/* don't catch any more signals */
     dprint(1, (debugfile, "\n\n** Received SIGHUP **\n\n\n\n"));
     fast_clean_up();
+#ifdef	_WINDOWS
+    _exit(0);				/* cleaning up can crash */
+#else
     printf("\n\nPine finished. Received hang up signal\n\n");
+#endif
 #endif	/* !DOS */
     exit(0);
 }
 
 
+/*----------------------------------------------------------------------
+      Timeout when no user input for a long, long time.
+      Treat it pretty much the same as if we got a HUP.
+      Only difference is we sometimes turns the timeout off (when composing).
+  ----------------------------------------------------------------------*/
+void
+user_input_timeout_exit(to_hours)
+    int to_hours;
+{
+    char msg[40];
+
+    dprint(1, (debugfile,
+	   "\n\n** Exiting: user input timeout (%d hours) **\n\n\n\n",
+	   to_hours));
+    sprintf(msg, "\n\nPine timed out (No user input for %d %s)\n", to_hours,
+	    to_hours > 1 ? "hours" : "hour");
+    fast_clean_up();
+    completely_done_with_adrbks();
+    end_screen(msg);
+    end_titlebar();
+    end_keymenu();
+    end_keyboard(F_ON(F_USE_FK,ps_global));
+    end_tty_driver(ps_global);
+    end_signals(0);
+    exit(0);
+}
+
+
 /*----------------------------------------------------------------------
       handle terminate signal -- SIGTERM
 
@@ -303,28 +331,32 @@ term_signal()
 Not much to do. Rely on periodic mail file check pointing.  Don't try
 cleaning up screen or flushing output since stdout is likely already
 gone.  To be safe, though, we'll at least restore the original tty mode.
+Also delete any remnant _DATAFILE_ from sending-filters.
   ----------------------------------------------------------------------*/
 void
 fast_clean_up()
 {
+    if(filter_data_file(0))
+        unlink(filter_data_file(0));
     if(ps_global->expunge_in_progress){
-	Raw(0);
+	PineRaw(0);
 	return;
     }
 
 #if !defined(DOS) && !defined(OS2)
     if(ps_global->inbox_stream != NULL && !ps_global->inbox_stream->lock){
-        if(ps_global->inbox_stream == ps_global->mail_stream)
-          ps_global->mail_stream = NULL; 
-        mail_close(ps_global->inbox_stream);
+	if(ps_global->inbox_stream == ps_global->mail_stream)
+	  ps_global->mail_stream = NULL; 
+
+	pine_close_stream(ps_global->inbox_stream);
     }
 
-    if(ps_global->mail_stream != NULL &&
-       ps_global->mail_stream != ps_global->inbox_stream &&
-       !ps_global->mail_stream->lock)
-      mail_close(ps_global->mail_stream);
+    if(ps_global->mail_stream != NULL
+       && ps_global->mail_stream != ps_global->inbox_stream
+       && !ps_global->mail_stream->lock)
+      pine_close_stream(ps_global->mail_stream);
 
-    Raw(0);
+    PineRaw(0);
 
 #endif	/* !DOS */
 #if	defined(DEBUG) && (!defined(DOS) || defined(_WINDOWS))
@@ -411,12 +443,23 @@ static winch_signal SIG_PROTO((int sig))
                ready_for_winch, winch_occured));
     clear_cursor_pos();
     init_sigwinch();
+    winch_cleanup();
+}
+#endif
+
+/*
+ * winch_cleanup - 
+ */
+void
+winch_cleanup()
+{
+#ifdef	RESIZING
     if(ready_for_winch)
       longjmp(winch_state, 1);
     else
       winch_occured = 1;
-}
 #endif
+}
 
 
 #ifdef	SIGCHLD
@@ -529,30 +572,32 @@ busy_alarm(seconds, msg, pc_func, init_msg)
 	    int space_left, slots_used;
 
 	    final_message = 1;
-	    space_left = (ps_global->ttyo ? ps_global->ttyo->screen_cols
-					  : 80) -
-					  busy_len - 2;  /* 2 is for [] */
-	    slots_used = max(0, min(space_left-3, 10));
+	    if(!status_message_remaining()){
+		space_left = (ps_global->ttyo ? ps_global->ttyo->screen_cols
+					      : 80) -
+					      busy_len - 2;  /* 2 is for [] */
+		slots_used = max(0, min(space_left-3, 10));
 
-	    if(percent_done_ptr && slots_used >= 4){
-		sprintf(progress, "%s |%*s|", busy_message, slots_used, "");
-		q_status_message(SM_ORDER, 0, 1, progress);
-	    }
-	    else{
-		dotcount++;
-		sprintf(progress, "%s%*s", busy_message,
-		    DISPLAY_CHARS_COLS + 1, "");
-		q_status_message(SM_ORDER, 0, 1, progress);
-	    }
+		if(percent_done_ptr && slots_used >= 4){
+		    sprintf(progress, "%s |%*s|", busy_message, slots_used, "");
+		    q_status_message(SM_ORDER, 0, 1, progress);
+		}
+		else{
+		    dotcount++;
+		    sprintf(progress, "%s%*s", busy_message,
+			DISPLAY_CHARS_COLS + 1, "");
+		    q_status_message(SM_ORDER, 0, 1, progress);
+		}
 
-	    /*
-	     * We use display_message so that the initial message will
-	     * be forced out only if there is not a previous message
-	     * currently being displayed that hasn't been displayed for
-	     * its min display time yet.  In that case, we don't want
-	     * to force out the initial message.
-	     */
-	    display_message('x');
+		/*
+		 * We use display_message so that the initial message will
+		 * be forced out only if there is not a previous message
+		 * currently being displayed that hasn't been displayed for
+		 * its min display time yet.  In that case, we don't want
+		 * to force out the initial message.
+		 */
+		display_message('x');
+	    }
 	}
 	
 #ifdef _WINDOWS
@@ -563,11 +608,19 @@ busy_alarm(seconds, msg, pc_func, init_msg)
 
     /* set alarm */
     if(F_OFF(F_DISABLE_ALARM, ps_global))
-      (void)alarm(seconds);
+      alarm_wrapper(seconds);
 
     return(retval);
 }
 
+void
+alarm_wrapper(seconds)
+    unsigned seconds;
+{
+    mail_parameters(NULL, SET_ALARMSAVE, (void *) (seconds ? 1L : 0L));
+    (void)alarm(seconds);
+}
+
 
 /*
  * If final_message was set when busy_alarm was called:
@@ -580,7 +633,7 @@ cancel_busy_alarm(message_pri)
 {
     dprint(9,(debugfile, "cancel_busy_alarm(%d)\n", message_pri));
 
-    (void)alarm(0);
+    alarm_wrapper(0);
 
     if(busy_alarm_outstanding){
 	int space_left, slots_used;
@@ -628,7 +681,7 @@ suspend_busy_alarm()
     dprint(9,(debugfile, "suspend_busy_alarm\n"));
 
     if(busy_alarm_outstanding)
-      alarm(0);
+      alarm_wrapper(0);
 }
 
 
@@ -636,12 +689,13 @@ suspend_busy_alarm()
  * resume_busy_alarm - continue previously installed busy_alarm.
  */
 void
-resume_busy_alarm()
+resume_busy_alarm(pause)
+    unsigned pause;
 {
     dprint(9,(debugfile, "resume_busy_alarm\n"));
 
     if(busy_alarm_outstanding && F_OFF(F_DISABLE_ALARM, ps_global))
-      (void)alarm(alarm_increment);
+      alarm_wrapper(pause ? pause : alarm_increment);
 }
 
 
@@ -743,7 +797,7 @@ alarm_signal_reset()
 #if	!defined(DOS) || defined(_WINDOWS)
     signal(SIGALRM, alarm_signal);
 #endif
-    (void)alarm(alarm_increment);
+    alarm_wrapper(alarm_increment);
 }
 
 
@@ -762,7 +816,7 @@ fake_alarm_blip()
        (long)((now=time((time_t *)0)) - last_alarm) >= (long)alarm_increment){
 	last_alarm = now;
 	/* manually execute code that an alarm would cause to execute */
-	alarm_signal(1);
+	alarm_signal SIG_PROTO((1));
     }
 }
 #endif
@@ -776,6 +830,16 @@ static SigType
 intr_signal SIG_PROTO((int sig))
 {
     ps_global->intr_pending = 1;
+
+    /*
+     * Anyone else want to know?
+     */
+    {
+	extern SORTPGM *g_sort_prog;
+
+	if(g_sort_prog)
+	  g_sort_prog->abort = 1;
+    }
 }
 
 
@@ -795,7 +859,7 @@ intr_disallow()
     if(signal(SIGINT, SIG_IGN) == SIG_IGN)	/* already off! */
       return;
 
-    pine_sigunblock(SIGINT);		/* unblock signal after longjmp */
+    our_sigunblock(SIGINT);		/* unblock signal after longjmp */
     ps_global->intr_pending = 0;
     intr_proc(0);			/* turn off interrupt char */
 }
@@ -828,6 +892,48 @@ intr_handling_off()
 }
 
 
+/*----------------------------------------------------------------------
+     Set or reset what needs to be set when coming out of pico to run
+     an alternate editor.
+
+   Args:   come_back -- If come_back is 0 we're going out of our environment
+		          to set up for an external editor.
+		        If come_back is 1 we're coming back into pine.
+  ----------------------------------------------------------------------*/
+int
+ttyfix(come_back) 
+    int come_back;
+{
+#if	defined(DEBUG) && (!defined(DOS) || defined(_WINDOWS))
+    if(debugfile)
+      fflush(debugfile);
+#endif
+
+    if(come_back){
+#ifdef OS2
+	enter_text_mode(NULL);
+#endif
+	init_screen();
+	init_tty_driver(ps_global);
+	init_keyboard(F_ON(F_USE_FK,ps_global));
+#ifdef OS2
+	dont_interrupt();
+#endif
+	fix_windsize(ps_global);
+    }
+    else{
+	EndInverse();
+	end_keyboard(F_ON(F_USE_FK,ps_global));
+	end_tty_driver(ps_global);
+	end_screen(NULL);
+#ifdef OS2
+	interrupt_ok();
+#endif
+    }
+}
+
+
+
 /*----------------------------------------------------------------------
      Suspend Pine. Reset tty and suspend. Suspend is finished when this returns
 
@@ -841,9 +947,9 @@ the ^Z and then clean up the tty driver, then kill ourself to stop, and
 pick up where we left off when execution resumes.
   ----------------------------------------------------------------------*/
 int
-do_suspend(pine) 
-    struct pine *pine;
+do_suspend() 
 {
+    struct pine *pine = ps_global;
     time_t now;
     int   result, isremote, retval;
     int   orig_cols, orig_rows;
@@ -854,8 +960,6 @@ do_suspend(pine)
 #ifdef	OS2
     static char *shell = NULL;
 #define	STD_SHELL	"CMD.EXE"
-#else
-    char *shell;
 #endif
 #endif
 
@@ -871,7 +975,8 @@ do_suspend(pine)
     }
 
 #ifdef	_WINDOWS
-    /* suspend what? */
+    mswin_minimize();
+    return(NO_OP_COMMAND);
 #else
 
     isremote = (ps_global->mail_stream && ps_global->mail_stream->mailbox
@@ -882,14 +987,9 @@ do_suspend(pine)
     now = time((time_t *)0);
     dprint(1, (debugfile, "\n\n --- %s - SUSPEND ----  %s",
 	       isremote ? "REMOTE" : "LOCAL", ctime(&now)));
-    EndInverse();
-    end_keyboard(F_ON(F_USE_FK,pine));
-    end_tty_driver(pine);
-    end_screen(NULL);
+    ttyfix(0);
+
 #if defined(DOS) || defined(OS2)
-#ifdef OS2
-    interrupt_ok();
-#endif
     suspend_notice("exit");
     if (!shell){
 	char *p;
@@ -907,7 +1007,7 @@ do_suspend(pine)
     if(F_ON(F_SUSPEND_SPAWNS, ps_global)){
 	PIPE_S *syspipe;
 
-	if(syspipe = open_system_pipe(NULL, NULL, NULL, PIPE_USER|PIPE_RESET)){
+	if(syspipe = open_system_pipe(NULL,NULL,NULL,PIPE_USER|PIPE_RESET,0)){
 	    suspend_notice("exit");
 #ifndef	SIGCHLD
 	    if(isremote)
@@ -929,16 +1029,9 @@ do_suspend(pine)
     now = time((time_t *)0);
     dprint(1, (debugfile, "\n\n ---- RETURN FROM SUSPEND ----  %s",
                ctime(&now)));
-#ifdef OS2
-    enter_text_mode(NULL);
-#endif
-    init_screen();
-    init_tty_driver(pine);
-    init_keyboard(F_ON(F_USE_FK,pine));
+    
+    ttyfix(1);
 
-#ifdef OS2
-    dont_interrupt();
-#endif
     orig_cols = pine->ttyo->screen_cols;
     orig_rows = pine->ttyo->screen_rows;
     fix_windsize(pine);
@@ -996,19 +1089,25 @@ void
 fix_windsize(pine)
     struct pine *pine;
 {
+    int old_width = pine->ttyo->screen_cols;
+
+    dprint(9, (debugfile, "fix_windsize()\n"));
     mark_keymenu_dirty();
     mark_status_dirty();
     mark_titlebar_dirty();
     clear_cursor_pos();
     get_windsize(pine->ttyo);
+
+    if(old_width != pine->ttyo->screen_cols)
+      clear_index_cache();
 }
 
 
-#if defined(DOS) || defined(OS2)
-SigType (*hold_int)(int), (*hold_term)(int), (*hold_quit)(int);
-#else
-SigType (*hold_hup)(), (*hold_int)(), (*hold_term)(), (*hold_usr2)(),
-	(*hold_quit)();
+SigType (*hold_int) SIG_PROTO((int));
+SigType (*hold_term) SIG_PROTO((int));
+#if !defined(DOS) && !defined(OS2)
+SigType (*hold_hup) SIG_PROTO((int));
+SigType (*hold_usr2) SIG_PROTO((int));
 #endif
 
 /*----------------------------------------------------------------------
@@ -1054,36 +1153,3 @@ mm_nocritical(stream)
 }
 
 
-#ifdef	POSIX_SIGNALS
-/*----------------------------------------------------------------------
-   Reset signals after critical imap code
- ----*/
-SigType
-(*posix_signal(sig_num, action))()
-    int	    sig_num;
-    SigType (*action)();
-{
-    struct sigaction new_action, old_action;
-
-    sigemptyset (&new_action.sa_mask);
-    new_action.sa_handler = action;
-#ifdef	SA_RESTART
-    new_action.sa_flags = SA_RESTART;
-#else
-    new_action.sa_flags = 0;
-#endif
-    sigaction(sig_num, &new_action, &old_action);
-    return(old_action.sa_handler);
-}
-
-int
-posix_sigunblock(mask)
-    int mask;
-{
-    sigset_t sig_mask;
-
-    sigemptyset(&sig_mask);
-    sigaddset(&sig_mask, mask);
-    return(sigprocmask(SIG_UNBLOCK, &sig_mask, NULL));
-}
-#endif /* POSIX_SIGNALS */
