@@ -1,5 +1,5 @@
 #if !defined(lint) && !defined(DOS)
-static char rcsid[] = "$Id: status.c,v 4.88 1996/05/14 20:00:17 mikes Exp $";
+static char rcsid[] = "$Id: status.c,v 4.107 1998/08/20 07:40:22 mikes Exp $";
 #endif
 /*----------------------------------------------------------------------
 
@@ -22,7 +22,7 @@ static char rcsid[] = "$Id: status.c,v 4.88 1996/05/14 20:00:17 mikes Exp $";
    permission of the University of Washington.
 
    Pine, Pico, and Pilot software and its included text are Copyright
-   1989-1996 by the University of Washington.
+   1989-1998 by the University of Washington.
 
    The full text of our legal notices is contained in the file called
    CPYRIGHT, included with this distribution.
@@ -55,7 +55,7 @@ static char rcsid[] = "$Id: status.c,v 4.88 1996/05/14 20:00:17 mikes Exp $";
  * allocated on demand, and cleared as each message is displayed.
  */
 typedef struct message {
-    char	    text[MAX_SCREEN_COLS+1];
+    char	   *text;
     unsigned	    flags:8;
     unsigned	    shown:1;
     int		    min_display_time, max_display_time;
@@ -66,15 +66,36 @@ typedef struct message {
 #define	RAD_BUT_COL	0
 
 
+/*
+ * Keymenu for Modal Message display screen
+ */
+static struct key modal_message_keys[] =
+       {NULL_MENU,
+	NULL_MENU,
+	{"Ret","Finished",{MC_EXIT,2,{ctrl('m'),ctrl('j')}},KS_NONE},
+	NULL_MENU,
+	NULL_MENU,
+	NULL_MENU,
+	PREVPAGE_MENU,
+	NEXTPAGE_MENU,
+	NULL_MENU,
+	NULL_MENU,
+	NULL_MENU,
+	NULL_MENU
+       };
+INST_KEY_MENU(modal_message_keymenu, modal_message_keys);
+
+
 /*
  * Internal prototypes
  */
 int  output_message PROTO((SMQ_T *));
 void radio_help PROTO((int, int, HelpType));
-void draw_radio_prompt PROTO((int, int, char *));
+void draw_radio_prompt PROTO((int, int, int, char *));
 void pause_for_current_message PROTO(());
 int  messages_in_queue PROTO(());
 void delay_cmd_cue PROTO((int));
+int  modal_bogus_input PROTO((int));
 
 
 
@@ -116,12 +137,25 @@ q_status_message(flags, min_time, max_time, message)
 {
     SMQ_T *new;
 
-    if((flags & SM_INFO) && message_queue)
-      return;
+    if(message_queue){
+	if(!strcmp(message_queue->text, message)){
+	    message_queue->shown = 0;
+
+	    if(message_queue->min_display_time < min_time)
+	      message_queue->min_display_time = min_time;
+
+	    if(message_queue->max_display_time < max_time)
+	      message_queue->max_display_time = max_time;
+
+	    return;
+	}
+	else if(flags & SM_INFO)
+	  return;
+    }
 
     new = (SMQ_T *)fs_get(sizeof(SMQ_T));
     memset(new, 0, sizeof(SMQ_T));
-    strncpy(new->text, message, MAX_SCREEN_COLS);
+    new->text = cpystr(message);
     new->min_display_time = min_time;
     new->max_display_time = max_time;
     new->flags            = flags;
@@ -156,10 +190,8 @@ q_status_message1(flags, min_t, max_t, s, a)
     char *s;
     void *a;
 {
-    char buf[1000];
-
-    sprintf(buf, s, a);
-    q_status_message(flags, min_t, max_t, buf);
+    sprintf(tmp_20k_buf, s, a);
+    q_status_message(flags, min_t, max_t, tmp_20k_buf);
 }
 
 
@@ -184,10 +216,8 @@ q_status_message2(flags, min_t, max_t, s, a1, a2)
     char *s;
     void *a1, *a2;
 {
-    char buf[1000];
-
-    sprintf(buf, s, a1, a2);
-    q_status_message(flags, min_t, max_t, buf);
+    sprintf(tmp_20k_buf, s, a1, a2);
+    q_status_message(flags, min_t, max_t, tmp_20k_buf);
 }
 
 
@@ -213,10 +243,8 @@ q_status_message3(flags, min_t, max_t, s, a1, a2, a3)
     char *s;
     void *a1, *a2, *a3;
 {
-    char buf[1000];
-
-    sprintf(buf, s, a1, a2, a3);
-    q_status_message(flags, min_t, max_t, buf);
+    sprintf(tmp_20k_buf, s, a1, a2, a3);
+    q_status_message(flags, min_t, max_t, tmp_20k_buf);
 }
 
 
@@ -243,10 +271,8 @@ q_status_message4(flags, min_t, max_t, s, a1, a2, a3, a4)
     char *s;
     void *a1, *a2, *a3, *a4;
 {
-    char buf[1000];
-
-    sprintf(buf, s, a1, a2, a3, a4);
-    q_status_message(flags, min_t, max_t, buf);
+    sprintf(tmp_20k_buf, s, a1, a2, a3, a4);
+    q_status_message(flags, min_t, max_t, tmp_20k_buf);
 }
 
 
@@ -276,10 +302,8 @@ q_status_message7(flags, min_t, max_t, s, a1, a2, a3, a4, a5, a6, a7)
     char *s;
     void *a1, *a2, *a3, *a4, *a5, *a6, *a7;
 {
-    char buf[1000];
-
-    sprintf(buf, s, a1, a2, a3, a4, a5, a6, a7);
-    q_status_message(flags, min_t, max_t, buf);
+    sprintf(tmp_20k_buf, s, a1, a2, a3, a4, a5, a6, a7);
+    q_status_message(flags, min_t, max_t, tmp_20k_buf);
 }
 
 
@@ -291,10 +315,8 @@ q_status_message8(flags, min_t, max_t, s, a1, a2, a3, a4, a5, a6, a7, a8)
     char *s;
     void *a1, *a2, *a3, *a4, *a5, *a6, *a7, *a8;
 {
-    char buf[1000];
-
-    sprintf(buf, s, a1, a2, a3, a4, a5, a6, a7, a8);
-    q_status_message(flags, min_t, max_t, buf);
+    sprintf(tmp_20k_buf, s, a1, a2, a3, a4, a5, a6, a7, a8);
+    q_status_message(flags, min_t, max_t, tmp_20k_buf);
 }
 
 
@@ -602,10 +624,15 @@ d_q_status_message()
 	    SMQ_T *p = message_queue;
 	    p->next->prev = p->prev;
 	    message_queue = p->prev->next = p->next;
+	    if(p->text)
+	      fs_give((void *)&p->text);
 	    fs_give((void **)&p);
 	}
-	else
-	  fs_give((void **)&message_queue);
+	else{
+	    if(message_queue->text)
+	      fs_give((void **)&message_queue->text);
+	    fs_give((void **)&message_queue);
+	}
     }
 }
 
@@ -733,7 +760,7 @@ int
 output_message(mq_entry)
     SMQ_T *mq_entry;
 {
-    int rv;
+    int rv = 0;
 
     dprint(9, (debugfile, "output_message(%s)\n", mq_entry->text));
 
@@ -741,51 +768,120 @@ output_message(mq_entry)
       Writechar(BELL, 0);			/* ring bell */
       /* flush() handled below */
 
-    rv = status_message_write(mq_entry->text, 0);
-
-    /*
-     * If we're a modal message, loop waiting for acknowlegment
-     * *and* keep the stream alive.
-     */
-    if((mq_entry->flags & SM_MODAL) && !mq_entry->shown){
-	static char *modal_msg =
-	     "* * * Read message above then hit Return to continue Pine * * *";
-	int ch;
-
-	ClearLine(ps_global->ttyo->screen_rows-2);
-	MoveCursor(ps_global->ttyo->screen_rows-2,
-		   max((ps_global->ttyo->screen_rows-sizeof(modal_msg))/2, 0));
-	Write_to_screen(modal_msg);
-	ClearLine(ps_global->ttyo->screen_rows-1);
-	mark_status_dirty();
-	mark_keymenu_dirty();
-
-	while(1){
-	    flush_input();
+    if(!(mq_entry->flags & SM_MODAL)){
+	rv = status_message_write(mq_entry->text, 0);
+    	if(ps_global->status_msg_delay){
 	    MoveCursor(ps_global->ttyo->screen_rows-FOOTER_ROWS(ps_global), 0);
-	    if((ch = read_char(600)) == NO_OP_COMMAND || ch == NO_OP_IDLE)
-	      new_mail(0, 2, 0);		/* wake up and prod server */
-	    else if(ch != ctrl('M'))
-	      Writechar(BELL, 0);		/* complain */
-	    else
-	      break;				/* done. */
+	    fflush(stdout);
+	    sleep(ps_global->status_msg_delay);
 	}
 
-	if(FOOTER_ROWS(ps_global) != 1){
-	    ClearLine(ps_global->ttyo->screen_rows-FOOTER_ROWS(ps_global));
-	    if(!ps_global->mangled_footer)
-	      redraw_keymenu();			/* else repaint next pass... */
-	}
-	else if(ps_global->redrawer)
-	  (*ps_global->redrawer)();
-    }
-    else if(!(mq_entry->flags & SM_MODAL) && ps_global->status_msg_delay){
-	MoveCursor(ps_global->ttyo->screen_rows-FOOTER_ROWS(ps_global), 0);
-	fflush(stdout);
-	sleep(ps_global->status_msg_delay);
+	mq_entry->shown = 1;
     }
+    else if (!mq_entry->shown){
+	int	  i      = 0,
+		  pad    = max(0, (ps_global->ttyo->screen_cols - 59) / 2);
+	char	 *p, *q, *s, *t;
+	SMQ_T	 *m;
+	SCROLL_S  sargs;
+	
+	/* Count the number of modal messsages and add up their lengths. */
+	for(m = mq_entry->next; m != mq_entry; m = m->next)
+	  if((m->flags & SM_MODAL) && !m->shown){
+	      i++;
+	  }
 
-    mq_entry->shown = 1;
+	sprintf(tmp_20k_buf, "%*s%s\n%*s%s\n%*s%s\n%*s%s\n%*s%s\n%*s%s\n%*s%s\n\n", 
+		/*        1         2         3         4         5         6*/
+		/*23456789012345678901234567890123456789012345678901234567890*/
+		pad, "",
+		"***********************************************************",
+		pad, "", i ? 
+		"* What follows are advisory messages.  After reading them *" :
+		"* What follows is an advisory message.  After reading it  *", 
+		
+		pad, "",
+		"* simply hit \"Return\" to continue your Pine session.      *", 
+		pad, "",
+		"*                                                         *",
+		pad, "", i ?
+		"* To review these messages later, press 'J' from the      *" :
+		"* To review this message later, press 'J' from the        *",
+		pad, "",
+		"* MAIN MENU.                                              *",
+		pad, "",
+		"***********************************************************");
+	t = tmp_20k_buf + strlen(tmp_20k_buf);	
+
+	m = mq_entry;
+	do{
+	    if((m->flags & SM_MODAL) && !m->shown){
+		int   indent;
+
+		indent = ps_global->ttyo->screen_cols > 80 
+		         ? (ps_global->ttyo->screen_cols - 80) / 3 : 0;
+		
+		if(t - tmp_20k_buf > 19000){
+		    sprintf(t, "\n%*s* * *  Running out of buffer space   * * *", indent, "");
+		    t += strlen(t);
+		    sprintf(t, "\n%*s* * * Press RETURN for more messages * * *", indent, "");
+		    break;
+		}
+		
+		add_review_message(m->text);
+		
+		if (p = strstr(m->text, "[ALERT]")){
+		    sprintf(t, "%*.*s\n", indent + p - m->text, p - m->text, m->text);
+		    t += strlen(t);
+
+		    for(p += 7; *p && isspace(*p); p++)
+		      ;
+		    indent += 8;
+		}
+		else{
+		    p = m->text;
+		}
+		
+		while(strlen(p) > ps_global->ttyo->screen_cols - 2 * indent){
+		    for(q = p + ps_global->ttyo->screen_cols - 2 * indent; 
+			q > p && !isspace(*q); q--)
+		      ;
+
+		    sprintf(t, "\n%*.*s", indent + q - p, q - p, p);
+		    t += strlen(t);
+		    p = q + 1;
+		}
+
+		sprintf(t, "\n%*s%s", indent, "", p);
+		t += strlen(t);
+		
+		if(i--){
+		    sprintf(t, "\n\n%*s\n", pad + 30, "-  -  -");
+		    t += strlen(t);
+		}
+
+		m->shown = 1;
+	    }
+	    m = m->next;
+	} while(m != mq_entry);
+
+	s = cpystr(tmp_20k_buf);
+	ClearLine(ps_global->ttyo->screen_rows - FOOTER_ROWS(ps_global));
+
+	memset(&sargs, 0, sizeof(SCROLL_S));
+	sargs.text.text	  = s;
+	sargs.text.src	  = CharStar;
+	sargs.bar.title	  = "Status Message";
+	sargs.bogus_input = modal_bogus_input;
+	sargs.no_stat_msg = 1;
+	sargs.keys.menu   = &modal_message_keymenu;
+	setbitmap(sargs.keys.bitmap);
+
+	scrolltool(&sargs);
+
+	fs_give((void **)&s);
+	ps_global->mangled_screen = 1;
+    }
     return(rv);
 }
 
@@ -824,6 +920,25 @@ delay_cmd_cue(on)
 }
 
 
+/*
+ * modal_bogus_input - used by scrolltool to complain about
+ *		       invalid user input.
+ */
+int
+modal_bogus_input(ch)
+    int ch;		
+{
+    char s[MAX_SCREEN_COLS+1];
+		
+    sprintf(s, "Command \"%s\" not allowed.  Press RETURN to continue Pine.",
+	    pretty_command(ch));
+    status_message_write(s, 0);
+    Writechar(BELL, 0);
+    return(0);
+}
+
+
+
 
 /*
  * want_to's array passed to radio_buttions...
@@ -843,16 +958,21 @@ static ESCKEY_S yorn[] = {
 			 be y or n)
          on_ctrl_C    -- Answer returned on ^C
 	 help         -- Two line help text
-	 display_help -- If true, display help without being asked
+	 flags        -- Flags to modify behavior
+			 WT_FLUSH_IN      - Discard pending input.
+			 WT_SEQ_SENSITIVE - Caller is sensitive to sequence
+			                    number changes caused by
+					    unsolicited expunges while we're
+					    viewing a message.
 
  Result: Messes up the status line,
-         returns y, n, dflt, or dflt_C
+         returns y, n, dflt, on_ctrl_C, or SEQ_EXCEPTION
   ---*/
 int
-want_to(question, dflt, on_ctrl_C, help, display_help, flush)
+want_to(question, dflt, on_ctrl_C, help, flags)
     char	*question;
     HelpType	help;
-    int		dflt, on_ctrl_C, display_help, flush;
+    int		dflt, on_ctrl_C, flags;
 {
     char *q2;
     int	  rv;
@@ -873,7 +993,7 @@ want_to(question, dflt, on_ctrl_C, help, display_help, flush)
        One problem with adding the (y/n) here is that shrinking the 
        screen while in radio_buttons() will cause it to get chopped
        off. It would be better to truncate the question passed in
-       hear and leave the full "(y/n) [x] : " on.
+       here and leave the full "(y/n) [x] : " on.
       ----*/
     q2 = fs_get(strlen(question) + 6);
     sprintf(q2, "%.*s? ", ps_global->ttyo->screen_cols - 6, question);
@@ -882,7 +1002,7 @@ want_to(question, dflt, on_ctrl_C, help, display_help, flush)
 
     rv = radio_buttons(q2,
 	(ps_global->ttyo->screen_rows > 4) ? - FOOTER_ROWS(ps_global) : -1,
-	yorn, dflt, on_ctrl_C, help, flush ? RB_FLUSH_IN : RB_NORM);
+	yorn, dflt, on_ctrl_C, help, flags);
     fs_give((void **)&q2);
 
     return(rv);
@@ -890,10 +1010,10 @@ want_to(question, dflt, on_ctrl_C, help, display_help, flush)
 
 
 int
-one_try_want_to(question, dflt, on_ctrl_C, help, display_help, flush)
+one_try_want_to(question, dflt, on_ctrl_C, help, flags)
     char      *question;
     HelpType   help;
-    int    dflt, on_ctrl_C, display_help, flush;
+    int    dflt, on_ctrl_C, flags;
 {
     char     *q2;
     int	      rv;
@@ -902,8 +1022,7 @@ one_try_want_to(question, dflt, on_ctrl_C, help, display_help, flush)
     sprintf(q2, "%.*s? ", ps_global->ttyo->screen_cols - 6, question);
     rv = radio_buttons(q2,
 	(ps_global->ttyo->screen_rows > 4) ? - FOOTER_ROWS(ps_global) : -1,
-	yorn, dflt, on_ctrl_C, help,
-        (flush ? RB_FLUSH_IN : RB_NORM) | RB_ONE_TRY);
+	yorn, dflt, on_ctrl_C, help, flags | RB_ONE_TRY);
     fs_give((void **)&q2);
 
     return(rv);
@@ -927,13 +1046,17 @@ Args --  prompt:    The prompt for the question/selection
 				 on_ctrl_C value if not answered acceptably
 				 on first try.
 		RB_NO_NEWMAIL  - Quell the usual newmail check.
+		RB_SEQ_SENSITIVE - The caller is sensitive to sequence number
+				   changes so return on_ctrl_C if an
+				   unsolicited expunge happens while we're
+				   viewing a message.
 	
 	 Note: If there are enough keys in the esc_list to need a second
 	       screen, and there is no help, then the 13th key will be
 	       put in the help position.
 
 Result -- Returns the letter pressed. Will be one of the characters in the
-          esc_list argument or one of the two deefaults.
+          esc_list argument, or dflt, or on_ctrl_C, or SEQ_EXCEPTION.
 
 This will pause for any new status message to be seen and then prompt the user.
 The prompt will be truncated to fit on the screen. Redraw and resize are
@@ -953,11 +1076,14 @@ radio_buttons(prompt, line, esc_list, dflt, on_ctrl_C, help_text, flags)
 {
     register int     ch, real_line;
     char            *q, *ds = NULL;
-    int              cursor_moved, max_label, i, start, fkey_table[12];
+    int              max_label, i, start, maxcol, fkey_table[12];
     int		     km_popped = 0;
     struct key	     rb_keys[12];
     struct key_menu  rb_keymenu;
     bitmap_t	     bitmap;
+#ifdef	_WINDOWS
+    int		     cursor_shown;
+#endif
 
 #ifdef _WINDOWS
     if (mswin_usedialog ()) {
@@ -970,20 +1096,21 @@ radio_buttons(prompt, line, esc_list, dflt, on_ctrl_C, help_text, flags)
 	memset (&button_list, 0, sizeof (MDlgButton) * 12);
 	b = 0;
 	for (i = 0; esc_list && esc_list[i].ch != -1 && i < 11; ++i) {
+	  if(esc_list[i].ch != -2){
 	    button_list[b].ch = esc_list[i].ch;
 	    button_list[b].rval = esc_list[i].rval;
 	    button_list[b].name = esc_list[i].name;
 	    button_list[b].label = esc_list[i].label;
 	    ++b;
+	  }
 	}
 	button_list[b].ch = -1;
 	
-	help = get_help_text (help_text, NULL);
+	help = get_help_text (help_text);
 
 	ret = mswin_select (prompt, button_list, dflt, on_ctrl_C, 
 			help, flags);
-	if (help != NULL) 
-	    free_help_text (help);
+	free_list_array(&help);
 	return (ret);
     }
 #endif
@@ -997,15 +1124,19 @@ radio_buttons(prompt, line, esc_list, dflt, on_ctrl_C, help_text, flags)
 
     /*---- Find longest label ----*/
     max_label = 0;
-    for(i = 0; esc_list && esc_list[i].ch != -1 && i < 12; i++){
+    for(i = 0; esc_list && esc_list[i].ch != -1 && i < 11; i++){
       if(esc_list[i].ch == -2) /* -2 means to skip this key and leave blank */
 	continue;
       max_label = max(max_label, strlen(esc_list[i].name));
     }
 
-    q = cpystr(prompt); /* So we can truncate string below if need be */
-    if(strlen(q) + max_label > ps_global->ttyo->screen_cols) 
-        q[ps_global->ttyo->screen_cols - max_label] = '\0';
+    maxcol = ps_global->ttyo->screen_cols - max_label - 1;
+
+    /*
+     * We need to be able to truncate q, so copy it in case it is
+     * a readonly string.
+     */
+    q = cpystr(prompt);
 
     /*---- Init structs for keymenu ----*/
     for(i = 0; i < 12; i++)
@@ -1037,29 +1168,41 @@ radio_buttons(prompt, line, esc_list, dflt, on_ctrl_C, help_text, flags)
 
     start = (start) ? 2 : 0;
     /*---- Show the usual possible keys ----*/
-    for(i=start; esc_list && esc_list[i-start].ch != -1 && i <= 12; i++){
+    for(i=start; esc_list && esc_list[i-start].ch != -1; i++){
 	/*
 	 * If we have an esc_list item we'd like to put in the non-existent
 	 * 13th slot, and there is no help, we put it in the help slot
 	 * instead.  We're hacking now...!
+	 *
+	 * We may also have invisible esc_list items that don't show up
+	 * on the screen.  We use this when we have two different keys
+	 * which are synonyms, like ^P and KEY_UP.  If all the slots are
+	 * already full we can still fit invisible keys off the screen to
+	 * the right.  A key is invisible if it's label is "".
 	 */
-	if(i == 12){
-	    if(help_text != NO_HELP)
-	      panic("Programming botch in radio_buttons(): too many keys");
+	if(i >= 12){
+	    if(esc_list[i-start].label[0] != '\0'){  /* visible */
+		if(i == 12){  /* special case where we put it in help slot */
+		    if(help_text != NO_HELP)
+		  panic("Programming botch in radio_buttons(): too many keys");
 
-	    if(esc_list[i-start].ch != -2)
-	      setbitn(0, bitmap); /* the help slot */
+		    if(esc_list[i-start].ch != -2)
+		      setbitn(0, bitmap); /* the help slot */
 
-	    fkey_table[0] = esc_list[i-start].ch;
-	    rb_keymenu.keys[0].name  = esc_list[i-start].name;
-	    if(esc_list[i-start].rval == dflt){
-		ds = (char *)fs_get((strlen(esc_list[i-start].label) + 3)
-				    * sizeof(char));
-		sprintf(ds, "[%s]", esc_list[i-start].label);
-		rb_keymenu.keys[0].label = ds;
+		    fkey_table[0] = esc_list[i-start].ch;
+		    rb_keymenu.keys[0].name  = esc_list[i-start].name;
+		    if(esc_list[i-start].rval == dflt){
+			ds = (char *)fs_get((strlen(esc_list[i-start].label)+3)
+					    * sizeof(char));
+			sprintf(ds, "[%s]", esc_list[i-start].label);
+			rb_keymenu.keys[0].label = ds;
+		    }
+		    else
+		      rb_keymenu.keys[0].label = esc_list[i-start].label;
+		}
+		else
+		  panic("Botch in radio_buttons(): too many keys");
 	    }
-	    else
-	      rb_keymenu.keys[0].label = esc_list[i-start].label;
 	}
 	else{
 	    if(esc_list[i-start].ch != -2)
@@ -1083,7 +1226,11 @@ radio_buttons(prompt, line, esc_list, dflt, on_ctrl_C, help_text, flags)
 
     ps_global->mangled_footer = 1;
 
-    draw_radio_prompt(real_line, RAD_BUT_COL, q);
+#ifdef	_WINDOWS
+    cursor_shown = mswin_showcursor(1);
+#endif
+
+    draw_radio_prompt(real_line, RAD_BUT_COL, maxcol, q);
 
     while(1){
         fflush(stdout);
@@ -1091,9 +1238,9 @@ radio_buttons(prompt, line, esc_list, dflt, on_ctrl_C, help_text, flags)
 	/*---- Paint the keymenu ----*/
 	EndInverse();
 	draw_keymenu(&rb_keymenu, bitmap, ps_global->ttyo->screen_cols,
-		     1 - FOOTER_ROWS(ps_global), 0, FirstMenu, 0);
+		     1 - FOOTER_ROWS(ps_global), 0, FirstMenu);
 	StartInverse();
-	MoveCursor(real_line, RAD_BUT_COL + strlen(q));
+	MoveCursor(real_line, min(RAD_BUT_COL+strlen(q), maxcol+1));
 
 	if(flags & RB_FLUSH_IN)
 	  flush_input();
@@ -1132,7 +1279,7 @@ radio_buttons(prompt, line, esc_list, dflt, on_ctrl_C, help_text, flags)
 	      if(ch == esc_list[i].ch){
 		  int len, n;
 
-		  MoveCursor(real_line, len = RAD_BUT_COL + strlen(q));
+		  MoveCursor(real_line,len=min(RAD_BUT_COL+strlen(q),maxcol+1));
 		  for(n = 0, len = ps_global->ttyo->screen_cols - len;
 		      esc_list[i].label[n] && len > 0;
 		      n++, len--)
@@ -1172,7 +1319,7 @@ radio_buttons(prompt, line, esc_list, dflt, on_ctrl_C, help_text, flags)
 		line = -3;
 		real_line = ps_global->ttyo->screen_rows + line;
 		clearfooter(ps_global);
-		draw_radio_prompt(real_line, RAD_BUT_COL, q);
+		draw_radio_prompt(real_line, RAD_BUT_COL, maxcol, q);
 		break;
 	    }
 
@@ -1184,7 +1331,7 @@ radio_buttons(prompt, line, esc_list, dflt, on_ctrl_C, help_text, flags)
 		CleartoEOLN();
 		radio_help(real_line, RAD_BUT_COL, help_text);
 		sleep(5);
-		MoveCursor(real_line, RAD_BUT_COL + strlen(q));
+		MoveCursor(real_line, min(RAD_BUT_COL+strlen(q), maxcol+1));
 	    }
 	    else
 	      Writechar(BELL, 0);
@@ -1194,13 +1341,23 @@ radio_buttons(prompt, line, esc_list, dflt, on_ctrl_C, help_text, flags)
 
           case NO_OP_COMMAND:
 	    goto newcmd;		/* misunderstood escape? */
-	    break;			/* never reached */
 
           case NO_OP_IDLE:		/* UNODIR, keep the stream alive */
 	    if(flags & RB_NO_NEWMAIL)
 	      goto newcmd;
-	    else if(new_mail(0, 2, 0) < 0)
-              break;			/* no changes, get on with life */
+
+	    i = new_mail(0, 2, NM_DEFER_SORT);
+	    if(ps_global->expunge_count && flags & RB_SEQ_SENSITIVE){
+		if(on_ctrl_C)
+		  ch = on_ctrl_C;
+		else
+		  ch = SEQ_EXCEPTION;
+
+		goto out_of_loop;
+	    }
+
+	    if(i < 0)
+	      break;		/* no changes, get on with life */
             /* Else fall into redraw to adjust displayed numbers and such */
 
 
@@ -1212,21 +1369,30 @@ radio_buttons(prompt, line, esc_list, dflt, on_ctrl_C, help_text, flags)
             redraw_titlebar();
             if(ps_global->redrawer != NULL)
               (*ps_global->redrawer)();
-            redraw_keymenu();
-            draw_radio_prompt(real_line, RAD_BUT_COL, q);
+	    if(FOOTER_ROWS(ps_global) == 3 || km_popped)
+              redraw_keymenu();
+
+	    maxcol = ps_global->ttyo->screen_cols - max_label - 1;
+	    draw_radio_prompt(real_line, RAD_BUT_COL, maxcol, q);
             break;
 
         } /* switch */
     }
 
   out_of_loop:
+
+#ifdef	_WINDOWS
+    if(!cursor_shown)
+      mswin_showcursor(0);
+#endif
+
     fs_give((void **)&q);
     if(ds)
       fs_give((void **)&ds);
 
     EndInverse();
     fflush(stdout);
-    resume_busy_alarm();
+    resume_busy_alarm(0);
     if(km_popped){
 	FOOTER_ROWS(ps_global) = 1;
 	clearfooter(ps_global);
@@ -1248,7 +1414,7 @@ radio_help(line, column, help)
     char **text;
 
 #if defined(HELPFILE)
-    text = get_help_text(help, NULL);
+    text = get_help_text(help);
 #else
     text = help;
 #endif
@@ -1268,7 +1434,7 @@ radio_help(line, column, help)
 
     StartInverse();
 #if defined(HELPFILE)
-    free_help_text(text);
+    free_list_array(&text);
 #endif
     fflush(stdout);
 }
@@ -1278,18 +1444,28 @@ radio_help(line, column, help)
    Paint the screen with the radio buttons prompt
   ----*/
 void
-draw_radio_prompt(l, c, q)
-    int       l, c;
+draw_radio_prompt(line, start_c, max_c, q)
+    int       line, start_c, max_c;
     char     *q;
 {
-    int x;
+    int x, len, save = -1;
+
+    len = (int)strlen(q);
+    if(len > max_c - start_c + 1){
+	save = q[max_c - start_c + 1];
+	q[max_c - start_c + 1] = '\0';
+	len = max_c - start_c + 1;
+    }
 
     StartInverse();
-    PutLine0(l, c, q);
-    x = c + strlen(q);
-    MoveCursor(l, x);
+    PutLine0(line, start_c, q);
+    x = start_c + len;
+    MoveCursor(line, x);
     while(x++ < ps_global->ttyo->screen_cols)
       Writechar(' ', 0);
-    MoveCursor(l, c + strlen(q));
+
+    MoveCursor(line, start_c + len);
     fflush(stdout);
+    if(save > 0)
+      q[max_c - start_c + 1] = save;
 }
